@@ -4,43 +4,36 @@ import (
 	"fmt"
 	"os"
 	"runtime/debug"
-	"runtime/pprof"
-	"runtime/trace"
 	"strings"
 	"time"
 
+	"github.com/evanw/esbuild/internal/api_helpers"
 	"github.com/evanw/esbuild/internal/logger"
 	"github.com/evanw/esbuild/pkg/cli"
 )
 
 var helpText = func(colors logger.Colors) string {
 	return `
-` + colors.Bold + `Usage:` + colors.Default + `
+` + colors.Bold + `Usage:` + colors.Reset + `
   esbuild [options] [entry points]
 
-` + colors.Bold + `Documentation:` + colors.Default + `
-  ` + colors.Underline + `https://esbuild.github.io/` + colors.Default + `
+` + colors.Bold + `Documentation:` + colors.Reset + `
+  ` + colors.Underline + `https://esbuild.github.io/` + colors.Reset + `
 
-` + colors.Bold + `Repository:` + colors.Default + `
-  ` + colors.Underline + `https://github.com/evanw/esbuild` + colors.Default + `
+` + colors.Bold + `Repository:` + colors.Reset + `
+  ` + colors.Underline + `https://github.com/evanw/esbuild` + colors.Reset + `
 
-` + colors.Bold + `Simple options:` + colors.Default + `
+` + colors.Bold + `Simple options:` + colors.Reset + `
   --bundle              Bundle all dependencies into the output files
   --define:K=V          Substitute K with V while parsing
-  --external:M          Exclude module M from the bundle
+  --external:M          Exclude module M from the bundle (can use * wildcards)
   --format=...          Output format (iife | cjs | esm, no default when not
                         bundling, otherwise default is iife when platform
                         is browser and cjs when platform is node)
-  --global-name=...     The name of the global for the IIFE format
-  --jsx-factory=...     What to use instead of React.createElement
-  --jsx-fragment=...    What to use instead of React.Fragment
   --loader:X=L          Use loader L to load file extension X, where L is
                         one of: js | jsx | ts | tsx | json | text | base64 |
                         file | dataurl | binary
-  --minify              Sets all --minify-* flags
-  --minify-whitespace   Remove whitespace
-  --minify-identifiers  Shorten identifiers
-  --minify-syntax       Use equivalent but shorter syntax
+  --minify              Minify the output (sets all --minify-* flags)
   --outdir=...          The output directory (for multiple entry points)
   --outfile=...         The output file (for one entry point)
   --platform=...        Platform target (browser | node | neutral,
@@ -48,33 +41,53 @@ var helpText = func(colors logger.Colors) string {
   --serve=...           Start a local HTTP server on this host:port for outputs
   --sourcemap           Emit a source map
   --splitting           Enable code splitting (currently only for esm)
-  --summary             Print some helpful information at the end of a build
   --target=...          Environment target (e.g. es2017, chrome58, firefox57,
                         safari11, edge16, node10, default esnext)
   --watch               Watch mode: rebuild on file system changes
 
-` + colors.Bold + `Advanced options:` + colors.Default + `
-  --banner=...              Text to be prepended to each output file
+` + colors.Bold + `Advanced options:` + colors.Reset + `
+  --allow-overwrite         Allow output files to overwrite input files
+  --asset-names=...         Path template to use for "file" loader files
+                            (default "[name]-[hash]")
+  --banner:T=...            Text to be prepended to each output file of type T
+                            where T is one of: css | js
   --charset=utf8            Do not escape UTF-8 code points
+  --chunk-names=...         Path template to use for code splitting chunks
+                            (default "[name]-[hash]")
   --color=...               Force use of color terminal escapes (true | false)
-  --error-limit=...         Maximum error count or 0 to disable (default 10)
-  --footer=...              Text to be appended to each output file
+  --entry-names=...         Path template to use for entry point output paths
+                            (default "[dir]/[name]", can also use "[hash]")
+  --footer:T=...            Text to be appended to each output file of type T
+                            where T is one of: css | js
+  --global-name=...         The name of the global for the IIFE format
   --inject:F                Import the file F into all input files and
                             automatically replace matching globals with imports
+  --jsx-factory=...         What to use for JSX instead of React.createElement
+  --jsx-fragment=...        What to use for JSX instead of React.Fragment
   --keep-names              Preserve "name" on functions and classes
-  --log-level=...           Disable logging (info | warning | error | silent,
-                            default info)
+  --legal-comments=...      Where to place license comments (none | inline |
+                            eof | linked | external, default eof when bundling
+                            and inline otherwise)
+  --log-level=...           Disable logging (verbose | debug | info | warning |
+                            error | silent, default info)
+  --log-limit=...           Maximum message count or 0 to disable (default 10)
   --main-fields=...         Override the main file order in package.json
                             (default "browser,module,main" when platform is
                             browser and "main,module" when platform is node)
   --metafile=...            Write metadata about the build to a JSON file
+  --minify-whitespace       Remove whitespace in output files
+  --minify-identifiers      Shorten identifiers in output files
+  --minify-syntax           Use equivalent but shorter syntax in output files
   --out-extension:.js=.mjs  Use a custom output extension instead of ".js"
   --outbase=...             The base path used to determine entry point output
                             paths (for multiple entry points)
+  --preserve-symlinks       Disable symlink resolution for module lookup
   --public-path=...         Set the base URL for the "file" loader
   --pure:N                  Mark the name N as a pure function for tree shaking
   --resolve-extensions=...  A comma-separated list of implicit extensions
-                            (default ".tsx,.ts,.jsx,.mjs,.cjs,.js,.css,.json")
+                            (default ".tsx,.ts,.jsx,.js,.css,.json")
+  --servedir=...            What to serve in addition to generated output files
+  --source-root=...         Sets the "sourceRoot" field in generated source maps
   --sourcefile=...          Set the source file for the source map (for stdin)
   --sourcemap=external      Do not link to the source map with a comment
   --sourcemap=inline        Emit the source map with an inline data URL
@@ -84,28 +97,37 @@ var helpText = func(colors logger.Colors) string {
   --tsconfig=...            Use this tsconfig.json file instead of other ones
   --version                 Print the current version (` + esbuildVersion + `) and exit
 
-` + colors.Bold + `Examples:` + colors.Default + `
-  ` + colors.Dim + `# Produces dist/entry_point.js and dist/entry_point.js.map` + colors.Default + `
+` + colors.Bold + `Examples:` + colors.Reset + `
+  ` + colors.Dim + `# Produces dist/entry_point.js and dist/entry_point.js.map` + colors.Reset + `
   esbuild --bundle entry_point.js --outdir=dist --minify --sourcemap
 
-  ` + colors.Dim + `# Allow JSX syntax in .js files` + colors.Default + `
+  ` + colors.Dim + `# Allow JSX syntax in .js files` + colors.Reset + `
   esbuild --bundle entry_point.js --outfile=out.js --loader:.js=jsx
 
-  ` + colors.Dim + `# Substitute the identifier RELEASE for the literal true` + colors.Default + `
+  ` + colors.Dim + `# Substitute the identifier RELEASE for the literal true` + colors.Reset + `
   esbuild example.js --outfile=out.js --define:RELEASE=true
 
-  ` + colors.Dim + `# Provide input via stdin, get output via stdout` + colors.Default + `
+  ` + colors.Dim + `# Provide input via stdin, get output via stdout` + colors.Reset + `
   esbuild --minify --loader=ts < input.ts > output.js
 
+  ` + colors.Dim + `# Automatically rebuild when input files are changed` + colors.Reset + `
+  esbuild app.ts --bundle --watch
+
+  ` + colors.Dim + `# Start a local HTTP server for everything in "www"` + colors.Reset + `
+  esbuild app.ts --bundle --servedir=www --outdir=www/js
+
 `
 }
 
 func main() {
+	logger.API = logger.CLIAPI
+
 	osArgs := os.Args[1:]
 	heapFile := ""
 	traceFile := ""
 	cpuprofileFile := ""
 	isRunningService := false
+	sendPings := false
 
 	// Do an initial scan over the argument list
 	argsEnd := 0
@@ -127,6 +149,11 @@ func main() {
 		case strings.HasPrefix(arg, "--trace="):
 			traceFile = arg[len("--trace="):]
 
+		case strings.HasPrefix(arg, "--timing"):
+			// This is a hidden flag because it's only intended for debugging esbuild
+			// itself. The output is not documented and not stable.
+			api_helpers.UseTimer = true
+
 		case strings.HasPrefix(arg, "--cpuprofile="):
 			cpuprofileFile = arg[len("--cpuprofile="):]
 
@@ -146,6 +173,9 @@ func main() {
 				os.Exit(1)
 			}
 
+		case strings.HasPrefix(arg, "--ping"):
+			sendPings = true
+
 		default:
 			// Strip any arguments that were handled above
 			osArgs[argsEnd] = arg
@@ -156,7 +186,7 @@ func main() {
 
 	// Run in service mode if requested
 	if isRunningService {
-		runService()
+		runService(sendPings)
 		return
 	}
 
@@ -172,34 +202,22 @@ func main() {
 		// To view a CPU trace, use "go tool trace [file]". Note that the trace
 		// viewer doesn't work under Windows Subsystem for Linux for some reason.
 		if traceFile != "" {
-			f, err := os.Create(traceFile)
-			if err != nil {
-				logger.PrintErrorToStderr(osArgs, fmt.Sprintf(
-					"Failed to create trace file: %s", err.Error()))
+			if done := createTraceFile(osArgs, traceFile); done == nil {
 				return
+			} else {
+				defer done()
 			}
-			defer f.Close()
-			trace.Start(f)
-			defer trace.Stop()
 		}
 
 		// To view a heap trace, use "go tool pprof [file]" and type "top". You can
 		// also drop it into https://speedscope.app and use the "left heavy" or
 		// "sandwich" view modes.
 		if heapFile != "" {
-			f, err := os.Create(heapFile)
-			if err != nil {
-				logger.PrintErrorToStderr(osArgs, fmt.Sprintf(
-					"Failed to create heap file: %s", err.Error()))
+			if done := createHeapFile(osArgs, heapFile); done == nil {
 				return
+			} else {
+				defer done()
 			}
-			defer func() {
-				if err := pprof.WriteHeapProfile(f); err != nil {
-					logger.PrintErrorToStderr(osArgs, fmt.Sprintf(
-						"Failed to write heap profile: %s", err.Error()))
-				}
-				f.Close()
-			}()
 		}
 
 		// To view a CPU profile, drop the file into https://speedscope.app.
@@ -207,15 +225,11 @@ func main() {
 		// Linux. The profiler has to be built for native Windows and run using the
 		// command prompt instead.
 		if cpuprofileFile != "" {
-			f, err := os.Create(cpuprofileFile)
-			if err != nil {
-				logger.PrintErrorToStderr(osArgs, fmt.Sprintf(
-					"Failed to create cpuprofile file: %s", err.Error()))
+			if done := createCpuprofileFile(osArgs, cpuprofileFile); done == nil {
 				return
+			} else {
+				defer done()
 			}
-			defer f.Close()
-			pprof.StartCPUProfile(f)
-			defer pprof.StopCPUProfile()
 		}
 
 		if cpuprofileFile != "" {
@@ -228,10 +242,10 @@ func main() {
 				exitCode = cli.Run(osArgs)
 			}
 		} else {
-			// Don't disable the GC if this is a long-running server process
+			// Don't disable the GC if this is a long-running process
 			isServe := false
 			for _, arg := range osArgs {
-				if arg == "--serve" || strings.HasPrefix(arg, "--serve=") {
+				if arg == "--serve" || arg == "--watch" || strings.HasPrefix(arg, "--serve=") {
 					isServe = true
 					break
 				}
