@@ -46,7 +46,9 @@ type outgoingPacket struct {
 	refCount int
 }
 
-func runService() {
+func runService(sendPings bool) {
+	logger.API = logger.JSAPI
+
 	service := serviceType{
 		callbacks:       make(map[uint32]responseCallback),
 		rebuilds:        make(map[int]rebuildCallback),
@@ -83,14 +85,16 @@ func runService() {
 	// where the host has disappeared and will never send us anything else but
 	// we incorrectly think we are still needed. In that case we will now try
 	// to write to stdout and fail, and then know that we should exit.
-	go func() {
-		for {
-			time.Sleep(1 * time.Second)
-			service.sendRequest(map[string]interface{}{
-				"command": "ping",
-			})
-		}
-	}()
+	if sendPings {
+		go func() {
+			for {
+				time.Sleep(1 * time.Second)
+				service.sendRequest(map[string]interface{}{
+					"command": "ping",
+				})
+			}
+		}()
+	}
 
 	for {
 		// Read more data from stdin
@@ -302,6 +306,11 @@ func (service *serviceType) handleIncomingPacket(bytes []byte) (result outgoingP
 				}),
 			}
 
+		case "format-msgs":
+			return outgoingPacket{
+				bytes: service.handleFormatMessagesRequest(p.id, request),
+			}
+
 		default:
 			return outgoingPacket{
 				bytes: encodePacket(packet{
@@ -343,25 +352,30 @@ func (service *serviceType) handleBuildRequest(id uint32, request map[string]int
 	key := request["key"].(int)
 	write := request["write"].(bool)
 	incremental := request["incremental"].(bool)
-	hasOnRebuild := request["hasOnRebuild"].(bool)
 	serveObj, isServe := request["serve"].(interface{})
+	entries := request["entries"].([]interface{})
 	flags := decodeStringArray(request["flags"].([]interface{}))
 
 	options, err := cli.ParseBuildOptions(flags)
 	options.AbsWorkingDir = request["absWorkingDir"].(string)
+	options.NodePaths = decodeStringArray(request["nodePaths"].([]interface{}))
+
+	for _, entry := range entries {
+		entry := entry.([]interface{})
+		key := entry[0].(string)
+		value := entry[1].(string)
+		options.EntryPointsAdvanced = append(options.EntryPointsAdvanced, api.EntryPoint{
+			OutputPath: key,
+			InputPath:  value,
+		})
+	}
 
 	// Normally when "write" is true and there is no output file/directory then
 	// the output is written to stdout instead. However, we're currently using
 	// stdout as a communication channel and writing the build output to stdout
-	// would corrupt our protocol.
-	//
-	// While we could channel this back to the host process and write it to
-	// stdout there, the public Go API we're about to call doesn't have an option
-	// for "write to stdout but don't actually write" and I don't think it should.
-	// For now let's just forbid this case because it's not even that useful.
-	if err == nil && !isServe && write && options.Outfile == "" && options.Outdir == "" {
-		err = errors.New("Either provide \"outfile\" or set \"write\" to false")
-	}
+	// would corrupt our protocol. Special-case this to channel this back to the
+	// host process and write it to stdout there.
+	writeToStdout := err == nil && !isServe && write && options.Outfile == "" && options.Outdir == ""
 
 	if err != nil {
 		return outgoingPacket{bytes: encodeErrorPacket(id, err)}
@@ -414,10 +428,16 @@ func (service *serviceType) handleBuildRequest(id uint32, request map[string]int
 		if options.Watch != nil {
 			response["watchID"] = watchID
 		}
+		if options.Metafile {
+			response["metafile"] = result.Metafile
+		}
+		if writeToStdout && len(result.OutputFiles) == 1 {
+			response["writeToStdout"] = result.OutputFiles[0].Contents
+		}
 		return response
 	}
 
-	if options.Watch != nil && hasOnRebuild {
+	if options.Watch != nil {
 		options.Watch.OnRebuild = func(result api.BuildResult) {
 			service.sendRequest(map[string]interface{}{
 				"command": "watch-rebuild",
@@ -427,7 +447,9 @@ func (service *serviceType) handleBuildRequest(id uint32, request map[string]int
 		}
 	}
 
-	options.Write = write
+	if !writeToStdout {
+		options.Write = write
+	}
 	options.Incremental = incremental
 	result := api.Build(options)
 	response := resultToResponse(result)
@@ -485,6 +507,9 @@ func (service *serviceType) handleServeRequest(id uint32, options api.BuildOptio
 	if host, ok := serve["host"]; ok {
 		serveOptions.Host = host.(string)
 	}
+	if servedir, ok := serve["servedir"]; ok {
+		serveOptions.Servedir = servedir.(string)
+	}
 	serveOptions.OnRequest = func(args api.ServeOnRequestArgs) {
 		service.sendRequest(map[string]interface{}{
 			"command": "serve-request",
@@ -597,6 +622,24 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 	goPlugins = append(goPlugins, api.Plugin{
 		Name: "JavaScript plugins",
 		Setup: func(build api.PluginBuild) {
+			build.OnStart(func() (api.OnStartResult, error) {
+				result := api.OnStartResult{}
+
+				response := service.sendRequest(map[string]interface{}{
+					"command": "start",
+					"key":     key,
+				}).(map[string]interface{})
+
+				if value, ok := response["errors"]; ok {
+					result.Errors = decodeMessages(value.([]interface{}))
+				}
+				if value, ok := response["warnings"]; ok {
+					result.Warnings = decodeMessages(value.([]interface{}))
+				}
+
+				return result, nil
+			})
+
 			build.OnResolve(api.OnResolveOptions{Filter: ".*"}, func(args api.OnResolveArgs) (api.OnResolveResult, error) {
 				var ids []interface{}
 				applyPath := logger.Path{Text: args.Path, Namespace: args.Namespace}
@@ -611,6 +654,31 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 					return result, nil
 				}
 
+				var kind string
+				switch args.Kind {
+				case api.ResolveEntryPoint:
+					kind = "entry-point"
+
+				// JS
+				case api.ResolveJSImportStatement:
+					kind = "import-statement"
+				case api.ResolveJSRequireCall:
+					kind = "require-call"
+				case api.ResolveJSDynamicImport:
+					kind = "dynamic-import"
+				case api.ResolveJSRequireResolve:
+					kind = "require-resolve"
+
+				// CSS
+				case api.ResolveCSSImportRule:
+					kind = "import-rule"
+				case api.ResolveCSSURLToken:
+					kind = "url-token"
+
+				default:
+					panic("Internal error")
+				}
+
 				response := service.sendRequest(map[string]interface{}{
 					"command":    "resolve",
 					"key":        key,
@@ -619,6 +687,7 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 					"importer":   args.Importer,
 					"namespace":  args.Namespace,
 					"resolveDir": args.ResolveDir,
+					"kind":       kind,
 					"pluginData": args.PluginData,
 				}).(map[string]interface{})
 
@@ -655,6 +724,12 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 				if value, ok := response["warnings"]; ok {
 					result.Warnings = decodeMessages(value.([]interface{}))
 				}
+				if value, ok := response["watchFiles"]; ok {
+					result.WatchFiles = decodeStringArray(value.([]interface{}))
+				}
+				if value, ok := response["watchDirs"]; ok {
+					result.WatchDirs = decodeStringArray(value.([]interface{}))
+				}
 
 				return result, nil
 			})
@@ -713,6 +788,12 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 				if value, ok := response["warnings"]; ok {
 					result.Warnings = decodeMessages(value.([]interface{}))
 				}
+				if value, ok := response["watchFiles"]; ok {
+					result.WatchFiles = decodeStringArray(value.([]interface{}))
+				}
+				if value, ok := response["watchDirs"]; ok {
+					result.WatchDirs = decodeStringArray(value.([]interface{}))
+				}
 				if value, ok := response["loader"]; ok {
 					loader, err := cli_helpers.ParseLoader(value.(string))
 					if err != nil {
@@ -792,6 +873,40 @@ func (service *serviceType) handleTransformRequest(id uint32, request map[string
 	})
 }
 
+func (service *serviceType) handleFormatMessagesRequest(id uint32, request map[string]interface{}) []byte {
+	msgs := decodeMessages(request["messages"].([]interface{}))
+
+	options := api.FormatMessagesOptions{
+		Kind: api.ErrorMessage,
+	}
+	if request["isWarning"].(bool) {
+		options.Kind = api.WarningMessage
+	}
+	if value, ok := request["color"].(bool); ok {
+		options.Color = value
+	}
+	if value, ok := request["terminalWidth"].(int); ok {
+		options.TerminalWidth = value
+	}
+
+	result := api.FormatMessages(msgs, options)
+
+	return encodePacket(packet{
+		id: id,
+		value: map[string]interface{}{
+			"messages": encodeStringArray(result),
+		},
+	})
+}
+
+func encodeStringArray(strings []string) []interface{} {
+	values := make([]interface{}, len(strings))
+	for i, value := range strings {
+		values[i] = value
+	}
+	return values
+}
+
 func decodeStringArray(values []interface{}) []string {
 	strings := make([]string, len(values))
 	for i, value := range values {
@@ -811,12 +926,39 @@ func encodeOutputFiles(outputFiles []api.OutputFile) []interface{} {
 	return values
 }
 
+func encodeLocation(loc *api.Location) interface{} {
+	if loc == nil {
+		return nil
+	}
+	return map[string]interface{}{
+		"file":       loc.File,
+		"namespace":  loc.Namespace,
+		"line":       loc.Line,
+		"column":     loc.Column,
+		"length":     loc.Length,
+		"lineText":   loc.LineText,
+		"suggestion": loc.Suggestion,
+	}
+}
+
 func encodeMessages(msgs []api.Message) []interface{} {
 	values := make([]interface{}, len(msgs))
 	for i, msg := range msgs {
-		value := make(map[string]interface{})
+		value := map[string]interface{}{
+			"pluginName": msg.PluginName,
+			"text":       msg.Text,
+			"location":   encodeLocation(msg.Location),
+		}
 		values[i] = value
-		value["text"] = msg.Text
+
+		notes := make([]interface{}, len(msg.Notes))
+		for j, note := range msg.Notes {
+			notes[j] = map[string]interface{}{
+				"text":     note.Text,
+				"location": encodeLocation(note.Location),
+			}
+		}
+		value["notes"] = notes
 
 		// Send "-1" to mean "undefined"
 		detail, ok := msg.Detail.(int)
@@ -824,79 +966,87 @@ func encodeMessages(msgs []api.Message) []interface{} {
 			detail = -1
 		}
 		value["detail"] = detail
-
-		// Some messages won't have a location
-		loc := msg.Location
-		if loc == nil {
-			value["location"] = nil
-		} else {
-			value["location"] = map[string]interface{}{
-				"file":     loc.File,
-				"line":     loc.Line,
-				"column":   loc.Column,
-				"length":   loc.Length,
-				"lineText": loc.LineText,
-			}
-		}
 	}
 	return values
 }
 
+func decodeLocation(value interface{}) *api.Location {
+	if value == nil {
+		return nil
+	}
+	loc := value.(map[string]interface{})
+	namespace := loc["namespace"].(string)
+	if namespace == "" {
+		namespace = "file"
+	}
+	return &api.Location{
+		File:       loc["file"].(string),
+		Namespace:  namespace,
+		Line:       loc["line"].(int),
+		Column:     loc["column"].(int),
+		Length:     loc["length"].(int),
+		LineText:   loc["lineText"].(string),
+		Suggestion: loc["suggestion"].(string),
+	}
+}
+
 func decodeMessages(values []interface{}) []api.Message {
 	msgs := make([]api.Message, len(values))
 	for i, value := range values {
 		obj := value.(map[string]interface{})
 		msg := api.Message{
-			Text:   obj["text"].(string),
-			Detail: obj["detail"].(int),
+			PluginName: obj["pluginName"].(string),
+			Text:       obj["text"].(string),
+			Location:   decodeLocation(obj["location"]),
+			Detail:     obj["detail"].(int),
 		}
-
-		// Some messages won't have a location
-		loc := obj["location"]
-		if loc != nil {
-			loc := loc.(map[string]interface{})
-			namespace := loc["namespace"].(string)
-			if namespace == "" {
-				namespace = "file"
-			}
-			msg.Location = &api.Location{
-				File:      loc["file"].(string),
-				Namespace: namespace,
-				Line:      loc["line"].(int),
-				Column:    loc["column"].(int),
-				Length:    loc["length"].(int),
-				LineText:  loc["lineText"].(string),
-			}
+		for _, note := range obj["notes"].([]interface{}) {
+			noteObj := note.(map[string]interface{})
+			msg.Notes = append(msg.Notes, api.Note{
+				Text:     noteObj["text"].(string),
+				Location: decodeLocation(noteObj["location"]),
+			})
 		}
-
 		msgs[i] = msg
 	}
 	return msgs
 }
 
-func decodeMessageToPrivate(obj map[string]interface{}) logger.Msg {
-	msg := logger.Msg{Data: logger.MsgData{
-		Text:       obj["text"].(string),
-		UserDetail: obj["detail"].(int),
-	}}
-
-	// Some messages won't have a location
-	loc := obj["location"]
-	if loc != nil {
-		loc := loc.(map[string]interface{})
-		namespace := loc["namespace"].(string)
-		if namespace == "" {
-			namespace = "file"
-		}
-		msg.Data.Location = &logger.MsgLocation{
-			File:      loc["file"].(string),
-			Namespace: namespace,
-			Line:      loc["line"].(int),
-			Column:    loc["column"].(int),
-			Length:    loc["length"].(int),
-			LineText:  loc["lineText"].(string),
-		}
+func decodeLocationToPrivate(value interface{}) *logger.MsgLocation {
+	if value == nil {
+		return nil
 	}
+	loc := value.(map[string]interface{})
+	namespace := loc["namespace"].(string)
+	if namespace == "" {
+		namespace = "file"
+	}
+	return &logger.MsgLocation{
+		File:       loc["file"].(string),
+		Namespace:  namespace,
+		Line:       loc["line"].(int),
+		Column:     loc["column"].(int),
+		Length:     loc["length"].(int),
+		LineText:   loc["lineText"].(string),
+		Suggestion: loc["suggestion"].(string),
+	}
+}
 
+func decodeMessageToPrivate(obj map[string]interface{}) logger.Msg {
+	msg := logger.Msg{
+		PluginName: obj["pluginName"].(string),
+		Data: logger.MsgData{
+			Text:       obj["text"].(string),
+			Location:   decodeLocationToPrivate(obj["location"]),
+			UserDetail: obj["detail"].(int),
+		},
+	}
+	for _, note := range obj["notes"].([]interface{}) {
+		noteObj := note.(map[string]interface{})
+		msg.Notes = append(msg.Notes, logger.MsgData{
+			Text:     noteObj["text"].(string),
+			Location: decodeLocationToPrivate(noteObj["location"]),
+		})
+	}
 	return msg
 }
