@@ -9,8 +9,11 @@ import "github.com/evanw/esbuild/internal/logger"
 type ImportKind uint8
 
 const (
+	// An entry point provided by the user
+	ImportEntryPoint ImportKind = iota
+
 	// An ES6 import or re-export statement
-	ImportStmt ImportKind = iota
+	ImportStmt
 
 	// A call to "require()"
 	ImportRequire
@@ -24,11 +27,11 @@ const (
 	// A CSS "@import" rule
 	ImportAt
 
+	// A CSS "@import" rule with import conditions
+	ImportAtConditional
+
 	// A CSS "url(...)" token
 	ImportURL
-
-	// An entry point provided by the user
-	ImportEntryPoint
 )
 
 func (kind ImportKind) StringForMetafile() string {
@@ -41,7 +44,7 @@ func (kind ImportKind) StringForMetafile() string {
 		return "dynamic-import"
 	case ImportRequireResolve:
 		return "require-resolve"
-	case ImportAt:
+	case ImportAt, ImportAtConditional:
 		return "import-rule"
 	case ImportURL:
 		return "url-token"
@@ -62,7 +65,7 @@ type ImportRecord struct {
 
 	// The resolved source index for an internal import (within the bundle) or
 	// nil for an external import (not included in the bundle)
-	SourceIndex *uint32
+	SourceIndex Index32
 
 	// Sometimes the parser creates an import record and decides it isn't needed.
 	// For example, TypeScript code may have import statements that later turn
@@ -74,19 +77,50 @@ type ImportRecord struct {
 	// CommonJS wrapper or not.
 	ContainsImportStar bool
 
+	// If this is true, the import contains an import for the alias "default",
+	// either via the "import x from" or "import {default as x} from" syntax.
+	ContainsDefaultAlias bool
+
 	// If true, this "export * from 'path'" statement is evaluated at run-time by
-	// calling the "__exportStar()" helper function
-	CallsRunTimeExportStarFn bool
+	// calling the "__reExport()" helper function
+	CallsRunTimeReExportFn bool
 
 	// Tell the printer to wrap this call to "require()" in "__toModule(...)"
 	WrapWithToModule bool
 
-	// True for require calls like this: "try { require() } catch {}". In this
-	// case we shouldn't generate an error if the path could not be resolved.
-	IsInsideTryBody bool
+	// True for the following cases:
+	//
+	//   try { require('x') } catch { handle }
+	//   try { await import('x') } catch { handle }
+	//   try { require.resolve('x') } catch { handle }
+	//   import('x').catch(handle)
+	//   import('x').then(_, handle)
+	//
+	// In these cases we shouldn't generate an error if the path could not be
+	// resolved.
+	HandlesImportErrors bool
 
 	// If true, this was originally written as a bare "import 'file'" statement
 	WasOriginallyBareImport bool
 
 	Kind ImportKind
 }
+
+// This stores a 32-bit index where the zero value is an invalid index. This is
+// a better alternative to storing the index as a pointer since that has the
+// same properties but takes up more space and costs an extra pointer traversal.
+type Index32 struct {
+	flippedBits uint32
+}
+
+func MakeIndex32(index uint32) Index32 {
+	return Index32{flippedBits: ^index}
+}
+
+func (i Index32) IsValid() bool {
+	return i.flippedBits != 0
+}
+
+func (i Index32) GetIndex() uint32 {
+	return ^i.flippedBits
+}
