@@ -124,7 +124,7 @@ func TestCSSFromJSMissingImport(t *testing.T) {
 			Mode:         config.ModeBundle,
 			AbsOutputDir: "/out",
 		},
-		expectedCompileLog: `entry.js: error: No matching export for import "missing"
+		expectedCompileLog: `entry.js: error: No matching export in "a.css" for import "missing"
 `,
 	})
 }
@@ -517,7 +517,91 @@ func TestCSSAtImportExtensionOrderCollisionUnsupported(t *testing.T) {
 				".css": config.LoaderCSS,
 			},
 		},
-		expectedScanLog: `entry.css: error: File could not be loaded: test.sass
+		expectedScanLog: `entry.css: error: No loader is configured for ".sass" files: test.sass
 `,
 	})
 }
+
+func TestCSSAtImportConditionsNoBundle(t *testing.T) {
+	css_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.css": `@import "./print.css" print;`,
+		},
+		entryPaths: []string{"/entry.css"},
+		options: config.Options{
+			Mode:          config.ModePassThrough,
+			AbsOutputFile: "/out.css",
+		},
+	})
+}
+
+func TestCSSAtImportConditionsBundleExternal(t *testing.T) {
+	css_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.css": `@import "https://example.com/print.css" print;`,
+		},
+		entryPaths: []string{"/entry.css"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.css",
+		},
+	})
+}
+
+func TestCSSAtImportConditionsBundle(t *testing.T) {
+	css_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.css": `@import "./print.css" print;`,
+			"/print.css": `body { color: red }`,
+		},
+		entryPaths: []string{"/entry.css"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.css",
+		},
+		expectedScanLog: `entry.css: error: Bundling with conditional "@import" rules is not currently supported
+`,
+	})
+}
+
+// This test mainly just makes sure that this scenario doesn't crash
+func TestCSSAndJavaScriptCodeSplittingIssue1064(t *testing.T) {
+	css_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/a.js": `
+				import shared from './shared.js'
+				console.log(shared() + 1)
+			`,
+			"/b.js": `
+				import shared from './shared.js'
+				console.log(shared() + 2)
+			`,
+			"/c.css": `
+				@import "./shared.css";
+				body { color: red }
+			`,
+			"/d.css": `
+				@import "./shared.css";
+				body { color: blue }
+			`,
+			"/shared.js": `
+				export default function() { return 3 }
+			`,
+			"/shared.css": `
+				body { background: black }
+			`,
+		},
+		entryPaths: []string{
+			"/a.js",
+			"/b.js",
+			"/c.css",
+			"/d.css",
+		},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			OutputFormat:  config.FormatESModule,
+			CodeSplitting: true,
+			AbsOutputDir:  "/out",
+		},
+	})
+}
