@@ -7,6 +7,7 @@ package bundler
 import (
 	"testing"
 
+	"github.com/evanw/esbuild/internal/compat"
 	"github.com/evanw/esbuild/internal/config"
 )
 
@@ -1153,6 +1154,13 @@ func TestTSLowerObjectRest2017NoBundle(t *testing.T) {
 
 				({ ...assign } = {});
 				({ obj_method({ ...x }) {} });
+
+				// Check for used return values
+				({ ...x } = x);
+				for ({ ...x } = x; 0; ) ;
+				console.log({ ...x } = x);
+				console.log({ x, ...xx } = { x });
+				console.log({ x: { ...xx } } = { x });
 			`,
 		},
 		entryPaths: []string{"/entry.ts"},
@@ -1194,6 +1202,13 @@ func TestTSLowerObjectRest2018NoBundle(t *testing.T) {
 
 				({ ...assign } = {});
 				({ obj_method({ ...x }) {} });
+
+				// Check for used return values
+				({ ...x } = x);
+				for ({ ...x } = x; 0; ) ;
+				console.log({ ...x } = x);
+				console.log({ x, ...xx } = { x });
+				console.log({ x: { ...xx } } = { x });
 			`,
 		},
 		entryPaths: []string{"/entry.ts"},
@@ -1277,3 +1292,248 @@ func TestLowerExportStarAsNameCollision(t *testing.T) {
 		},
 	})
 }
+
+func TestLowerStrictModeSyntax(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				import './for-in'
+			`,
+			"/for-in.js": `
+				if (test)
+					for (var a = b in {}) ;
+				for (var x = y in {}) ;
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			OutputFormat:  config.FormatESModule,
+			AbsOutputFile: "/out.js",
+		},
+	})
+}
+
+func TestLowerForbidStrictModeSyntax(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				import './with'
+				import './delete-1'
+				import './delete-2'
+				import './delete-3'
+			`,
+			"/with.js": `
+				with (x) y
+			`,
+			"/delete-1.js": `
+				delete x
+			`,
+			"/delete-2.js": `
+				delete (y)
+			`,
+			"/delete-3.js": `
+				delete (1 ? z : z)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			OutputFormat:  config.FormatESModule,
+			AbsOutputFile: "/out.js",
+		},
+		expectedScanLog: `delete-1.js: error: Delete of a bare identifier cannot be used with the "esm" output format due to strict mode
+delete-2.js: error: Delete of a bare identifier cannot be used with the "esm" output format due to strict mode
+with.js: error: With statements cannot be used with the "esm" output format due to strict mode
+`,
+	})
+}
+
+func TestLowerPrivateClassFieldOrder(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				class Foo {
+					#foo = 123 // This must be set before "bar" is initialized
+					bar = this.#foo
+				}
+				console.log(new Foo().bar === 123)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:                  config.ModePassThrough,
+			AbsOutputFile:         "/out.js",
+			UnsupportedJSFeatures: compat.ClassPrivateField,
+		},
+	})
+}
+
+func TestLowerPrivateClassMethodOrder(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				class Foo {
+					bar = this.#foo()
+					#foo() { return 123 } // This must be set before "bar" is initialized
+				}
+				console.log(new Foo().bar === 123)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:                  config.ModePassThrough,
+			AbsOutputFile:         "/out.js",
+			UnsupportedJSFeatures: compat.ClassPrivateMethod,
+		},
+	})
+}
+
+func TestLowerPrivateClassAccessorOrder(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				class Foo {
+					bar = this.#foo
+					get #foo() { return 123 } // This must be set before "bar" is initialized
+				}
+				console.log(new Foo().bar === 123)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:                  config.ModePassThrough,
+			AbsOutputFile:         "/out.js",
+			UnsupportedJSFeatures: compat.ClassPrivateAccessor,
+		},
+	})
+}
+
+func TestLowerPrivateClassStaticFieldOrder(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				class Foo {
+					static #foo = 123 // This must be set before "bar" is initialized
+					static bar = Foo.#foo
+				}
+				console.log(Foo.bar === 123)
+
+				class FooThis {
+					static #foo = 123 // This must be set before "bar" is initialized
+					static bar = this.#foo
+				}
+				console.log(FooThis.bar === 123)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:                  config.ModePassThrough,
+			AbsOutputFile:         "/out.js",
+			UnsupportedJSFeatures: compat.ClassPrivateStaticField,
+		},
+	})
+}
+
+func TestLowerPrivateClassStaticMethodOrder(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				class Foo {
+					static bar = Foo.#foo()
+					static #foo() { return 123 } // This must be set before "bar" is initialized
+				}
+				console.log(Foo.bar === 123)
+
+				class FooThis {
+					static bar = this.#foo()
+					static #foo() { return 123 } // This must be set before "bar" is initialized
+				}
+				console.log(FooThis.bar === 123)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:                  config.ModePassThrough,
+			AbsOutputFile:         "/out.js",
+			UnsupportedJSFeatures: compat.ClassPrivateStaticMethod,
+		},
+	})
+}
+
+func TestLowerPrivateClassStaticAccessorOrder(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				class Foo {
+					static bar = Foo.#foo
+					static get #foo() { return 123 } // This must be set before "bar" is initialized
+				}
+				console.log(Foo.bar === 123)
+
+				class FooThis {
+					static bar = this.#foo
+					static get #foo() { return 123 } // This must be set before "bar" is initialized
+				}
+				console.log(FooThis.bar === 123)
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:                  config.ModePassThrough,
+			AbsOutputFile:         "/out.js",
+			UnsupportedJSFeatures: compat.ClassPrivateStaticAccessor,
+		},
+	})
+}
+
+func TestLowerPrivateClassBrandCheckUnsupported(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				class Foo {
+					#foo
+					#bar
+					baz() {
+						return [
+							this.#foo,
+							this.#bar,
+							#foo in this,
+						]
+					}
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:                  config.ModePassThrough,
+			AbsOutputFile:         "/out.js",
+			UnsupportedJSFeatures: compat.ClassPrivateBrandCheck,
+		},
+	})
+}
+
+func TestLowerPrivateClassBrandCheckSupported(t *testing.T) {
+	lower_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				class Foo {
+					#foo
+					#bar
+					baz() {
+						return [
+							this.#foo,
+							this.#bar,
+							#foo in this,
+						]
+					}
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:          config.ModePassThrough,
+			AbsOutputFile: "/out.js",
+		},
+	})
+}
