@@ -206,8 +206,8 @@ func TestTSImportMissingES6(t *testing.T) {
 			Mode:          config.ModeBundle,
 			AbsOutputFile: "/out.js",
 		},
-		expectedCompileLog: `entry.ts: error: No matching export for import "default"
-entry.ts: error: No matching export for import "y"
+		expectedCompileLog: `entry.ts: error: No matching export in "foo.js" for import "default"
+entry.ts: error: No matching export in "foo.js" for import "y"
 `,
 	})
 }
@@ -521,7 +521,8 @@ func TestTypeScriptDecorators(t *testing.T) {
 				import {i} from './i'
 				import {j} from './j'
 				import k from './k'
-				console.log(all, all_computed, a, b, c, d, e, f, g, h, i, j, k)
+				import {fn} from './arguments'
+				console.log(all, all_computed, a, b, c, d, e, f, g, h, i, j, k, fn)
 			`,
 			"/all.ts": `
 				@x.y()
@@ -529,6 +530,7 @@ func TestTypeScriptDecorators(t *testing.T) {
 				export default class Foo {
 					@x @y mUndef
 					@x @y mDef = 1
+					constructor(@x0 @y0 arg0, @x1 @y1 arg1) {}
 					@x @y method(@x0 @y0 arg0, @x1 @y1 arg1) { return new Foo }
 					@x @y static sUndef
 					@x @y static sDef = new Foo
@@ -624,6 +626,16 @@ func TestTypeScriptDecorators(t *testing.T) {
 					foo(@x(() => 0) @y(() => 1) x) {}
 				}
 			`,
+			"/arguments.ts": `
+				function dec(x: any): any {}
+				export function fn(x: string): any {
+					class Foo {
+						@dec(arguments[0])
+						[arguments[0]]() {}
+					}
+					return Foo;
+				}
+			`,
 		},
 		entryPaths: []string{"/entry.js"},
 		options: config.Options{
@@ -653,13 +665,13 @@ func TestTSExportDefaultTypeIssue316(t *testing.T) {
 				import tone_def, { bar as tone } from './remove/type-only-namespace-exported'
 
 				export default [
-					dc,
-					dl,
-					im,
-					_in,
-					tn,
-					vn,
-					vnm,
+					dc_def, dc,
+					dl_def, dl,
+					im_def, im,
+					in_def, _in,
+					tn_def, tn,
+					vn_def, vn,
+					vnm_def, vnm,
 
 					i,
 					ie,
@@ -865,8 +877,164 @@ func TestExportTypeIssue379(t *testing.T) {
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:          config.ModeBundle,
+			Mode:                    config.ModeBundle,
+			AbsOutputFile:           "/out.js",
+			UseDefineForClassFields: config.False,
+		},
+	})
+}
+
+func TestThisInsideFunctionTS(t *testing.T) {
+	default_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				function foo(x = this) { console.log(this) }
+				const objFoo = {
+					foo(x = this) { console.log(this) }
+				}
+				class Foo {
+					x = this
+					static y = this.z
+					foo(x = this) { console.log(this) }
+					static bar(x = this) { console.log(this) }
+				}
+				new Foo(foo(objFoo))
+				if (nested) {
+					function bar(x = this) { console.log(this) }
+					const objBar = {
+						foo(x = this) { console.log(this) }
+					}
+					class Bar {
+						x = this
+						static y = this.z
+						foo(x = this) { console.log(this) }
+						static bar(x = this) { console.log(this) }
+					}
+					new Bar(bar(objBar))
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:                    config.ModeBundle,
+			AbsOutputFile:           "/out.js",
+			UseDefineForClassFields: config.False,
+		},
+	})
+}
+
+func TestThisInsideFunctionTSUseDefineForClassFields(t *testing.T) {
+	default_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				function foo(x = this) { console.log(this) }
+				const objFoo = {
+					foo(x = this) { console.log(this) }
+				}
+				class Foo {
+					x = this
+					static y = this.z
+					foo(x = this) { console.log(this) }
+					static bar(x = this) { console.log(this) }
+				}
+				new Foo(foo(objFoo))
+				if (nested) {
+					function bar(x = this) { console.log(this) }
+					const objBar = {
+						foo(x = this) { console.log(this) }
+					}
+					class Bar {
+						x = this
+						static y = this.z
+						foo(x = this) { console.log(this) }
+						static bar(x = this) { console.log(this) }
+					}
+					new Bar(bar(objBar))
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:                    config.ModeBundle,
+			AbsOutputFile:           "/out.js",
+			UseDefineForClassFields: config.True,
+		},
+	})
+}
+
+func TestThisInsideFunctionTSNoBundle(t *testing.T) {
+	default_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				function foo(x = this) { console.log(this) }
+				const objFoo = {
+					foo(x = this) { console.log(this) }
+				}
+				class Foo {
+					x = this
+					static y = this.z
+					foo(x = this) { console.log(this) }
+					static bar(x = this) { console.log(this) }
+				}
+				new Foo(foo(objFoo))
+				if (nested) {
+					function bar(x = this) { console.log(this) }
+					const objBar = {
+						foo(x = this) { console.log(this) }
+					}
+					class Bar {
+						x = this
+						static y = this.z
+						foo(x = this) { console.log(this) }
+						static bar(x = this) { console.log(this) }
+					}
+					new Bar(bar(objBar))
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModePassThrough,
 			AbsOutputFile: "/out.js",
 		},
 	})
 }
+
+func TestThisInsideFunctionTSNoBundleUseDefineForClassFields(t *testing.T) {
+	default_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				function foo(x = this) { console.log(this) }
+				const objFoo = {
+					foo(x = this) { console.log(this) }
+				}
+				class Foo {
+					x = this
+					static y = this.z
+					foo(x = this) { console.log(this) }
+					static bar(x = this) { console.log(this) }
+				}
+				new Foo(foo(objFoo))
+				if (nested) {
+					function bar(x = this) { console.log(this) }
+					const objBar = {
+						foo(x = this) { console.log(this) }
+					}
+					class Bar {
+						x = this
+						static y = this.z
+						foo(x = this) { console.log(this) }
+						static bar(x = this) { console.log(this) }
+					}
+					new Bar(bar(objBar))
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:                    config.ModePassThrough,
+			AbsOutputFile:           "/out.js",
+			UseDefineForClassFields: config.True,
+		},
+	})
+}
