@@ -3,8 +3,10 @@ package config
 import (
 	"fmt"
 	"regexp"
+	"strings"
 	"sync"
 
+	"github.com/evanw/esbuild/internal/ast"
 	"github.com/evanw/esbuild/internal/compat"
 	"github.com/evanw/esbuild/internal/js_ast"
 	"github.com/evanw/esbuild/internal/logger"
@@ -51,6 +53,20 @@ const (
 	SourceMapInlineAndExternal
 )
 
+type LegalComments uint8
+
+const (
+	LegalCommentsInline LegalComments = iota
+	LegalCommentsNone
+	LegalCommentsEndOfFile
+	LegalCommentsLinkedWithComment
+	LegalCommentsExternalWithoutComment
+)
+
+func (lc LegalComments) HasExternalFile() bool {
+	return lc == LegalCommentsLinkedWithComment || lc == LegalCommentsExternalWithoutComment
+}
+
 type Loader int
 
 const (
@@ -158,12 +174,33 @@ const (
 	ModeBundle
 )
 
+type ModuleType uint8
+
+const (
+	ModuleUnknown ModuleType = iota
+	ModuleCommonJS
+	ModuleESM
+)
+
+type MaybeBool uint8
+
+const (
+	Unspecified MaybeBool = iota
+	True
+	False
+)
+
 type Options struct {
 	Mode              Mode
+	ModuleType        ModuleType
+	PreserveSymlinks  bool
 	RemoveWhitespace  bool
 	MinifyIdentifiers bool
 	MangleSyntax      bool
 	CodeSplitting     bool
+	WatchMode         bool
+	AllowOverwrite    bool
+	LegalComments     LegalComments
 
 	// Setting this to true disables warnings about code that is very likely to
 	// be a bug. This is used to ignore issues inside "node_modules" directories.
@@ -182,7 +219,7 @@ type Options struct {
 
 	OmitRuntimeForTests     bool
 	PreserveUnusedImportsTS bool
-	UseDefineForClassFields bool
+	UseDefineForClassFields MaybeBool
 	ASCIIOnly               bool
 	KeepNames               bool
 	IgnoreDCEAnnotations    bool
@@ -195,8 +232,14 @@ type Options struct {
 	UnsupportedJSFeatures  compat.JSFeature
 	UnsupportedCSSFeatures compat.CSSFeature
 
+	// This is the original information that was used to generate the
+	// unsupported feature sets above. It's used for error messages.
+	OriginalTargetEnv string
+
 	ExtensionOrder  []string
 	MainFields      []string
+	Conditions      []string
+	AbsNodePaths    []string // The "NODE_PATH" variable from Node.js
 	ExternalModules ExternalModules
 
 	AbsOutputFile      string
@@ -212,20 +255,128 @@ type Options struct {
 	InjectAbsPaths     []string
 	InjectedDefines    []InjectedDefine
 	InjectedFiles      []InjectedFile
-	Banner             string
-	Footer             string
+
+	JSBanner  string
+	JSFooter  string
+	CSSBanner string
+	CSSFooter string
+
+	EntryPathTemplate []PathTemplate
+	ChunkPathTemplate []PathTemplate
+	AssetPathTemplate []PathTemplate
 
 	Plugins []Plugin
 
-	// If present, metadata about the bundle is written as JSON here
-	AbsMetadataFile string
+	NeedsMetafile bool
 
 	SourceMap             SourceMap
+	SourceRoot            string
 	ExcludeSourcesContent bool
 
 	Stdin *StdinInfo
 }
 
+type PathPlaceholder uint8
+
+const (
+	NoPlaceholder PathPlaceholder = iota
+
+	// The relative path from the original parent directory to the configured
+	// "outbase" directory, or to the lowest common ancestor directory
+	DirPlaceholder
+
+	// The original name of the file, or the manual chunk name, or the name of
+	// the type of output file ("entry" or "chunk" or "asset")
+	NamePlaceholder
+
+	// A hash of the contents of this file, and the contents and output paths of
+	// all dependencies (except for their hash placeholders)
+	HashPlaceholder
+)
+
+type PathTemplate struct {
+	Data        string
+	Placeholder PathPlaceholder
+}
+
+type PathPlaceholders struct {
+	Dir  *string
+	Name *string
+	Hash *string
+}
+
+func (placeholders PathPlaceholders) Get(placeholder PathPlaceholder) *string {
+	switch placeholder {
+	case DirPlaceholder:
+		return placeholders.Dir
+	case NamePlaceholder:
+		return placeholders.Name
+	case HashPlaceholder:
+		return placeholders.Hash
+	}
+	return nil
+}
+
+func TemplateToString(template []PathTemplate) string {
+	if len(template) == 1 && template[0].Placeholder == NoPlaceholder {
+		// Avoid allocations in this case
+		return template[0].Data
+	}
+	sb := strings.Builder{}
+	for _, part := range template {
+		sb.WriteString(part.Data)
+		switch part.Placeholder {
+		case DirPlaceholder:
+			sb.WriteString("[dir]")
+		case NamePlaceholder:
+			sb.WriteString("[name]")
+		case HashPlaceholder:
+			sb.WriteString("[hash]")
+		}
+	}
+	return sb.String()
+}
+
+func HasPlaceholder(template []PathTemplate, placeholder PathPlaceholder) bool {
+	for _, part := range template {
+		if part.Placeholder == placeholder {
+			return true
+		}
+	}
+	return false
+}
+
+func SubstituteTemplate(template []PathTemplate, placeholders PathPlaceholders) []PathTemplate {
+	// Don't allocate if no substitution is possible and the template is already minimal
+	shouldSubstitute := false
+	for i, part := range template {
+		if placeholders.Get(part.Placeholder) != nil || (part.Placeholder == NoPlaceholder && i+1 < len(template)) {
+			shouldSubstitute = true
+			break
+		}
+	}
+	if !shouldSubstitute {
+		return template
+	}
+
+	// Otherwise, substitute and merge as appropriate
+	result := make([]PathTemplate, 0, len(template))
+	for _, part := range template {
+		if sub := placeholders.Get(part.Placeholder); sub != nil {
+			part.Data += *sub
+			part.Placeholder = NoPlaceholder
+		}
+		if last := len(result) - 1; last >= 0 && result[last].Placeholder == NoPlaceholder {
+			last := &result[last]
+			last.Data += part.Data
+			last.Placeholder = part.Placeholder
+		} else {
+			result = append(result, part)
+		}
+	}
+	return result
+}
+
 func IsTreeShakingEnabled(mode Mode, outputFormat Format) bool {
 	return mode == ModeBundle || (mode == ModeConvertFormat && outputFormat == FormatIIFE)
 }
@@ -303,10 +454,21 @@ func PluginAppliesToPath(path logger.Path, filter *regexp.Regexp, namespace stri
 
 type Plugin struct {
 	Name      string
+	OnStart   []OnStart
 	OnResolve []OnResolve
 	OnLoad    []OnLoad
 }
 
+type OnStart struct {
+	Name     string
+	Callback func() OnStartResult
+}
+
+type OnStartResult struct {
+	Msgs        []logger.Msg
+	ThrownError error
+}
+
 type OnResolve struct {
 	Name      string
 	Filter    *regexp.Regexp
@@ -318,6 +480,7 @@ type OnResolveArgs struct {
 	Path       string
 	Importer   logger.Path
 	ResolveDir string
+	Kind       ast.ImportKind
 	PluginData interface{}
 }
 
@@ -330,6 +493,9 @@ type OnResolveResult struct {
 
 	Msgs        []logger.Msg
 	ThrownError error
+
+	AbsWatchFiles []string
+	AbsWatchDirs  []string
 }
 
 type OnLoad struct {
@@ -354,4 +520,7 @@ type OnLoadResult struct {
 
 	Msgs        []logger.Msg
 	ThrownError error
+
+	AbsWatchFiles []string
+	AbsWatchDirs  []string
 }
