@@ -29,6 +29,7 @@ type parser struct {
 type Options struct {
 	UnsupportedCSSFeatures compat.CSSFeature
 	MangleSyntax           bool
+	RemoveWhitespace       bool
 }
 
 func Parse(log logger.Log, source logger.Source, options Options) css_ast.AST {
@@ -152,18 +153,58 @@ type ruleContext struct {
 }
 
 func (p *parser) parseListOfRules(context ruleContext) []css_ast.R {
+	didWarnAboutCharset := false
+	didWarnAboutImport := false
 	rules := []css_ast.R{}
+	locs := []logger.Loc{}
+
+loop:
 	for {
 		switch p.current().Kind {
 		case css_lexer.TEndOfFile, css_lexer.TCloseBrace:
-			return rules
+			break loop
 
 		case css_lexer.TWhitespace:
 			p.advance()
 			continue
 
 		case css_lexer.TAtKeyword:
-			rules = append(rules, p.parseAtRule(atRuleContext{}))
+			first := p.current().Range
+			rule := p.parseAtRule(atRuleContext{})
+
+			// Validate structure
+			if context.isTopLevel {
+				switch rule.(type) {
+				case *css_ast.RAtCharset:
+					if !didWarnAboutCharset && len(rules) > 0 {
+						p.log.AddRangeWarningWithNotes(&p.source, first, "\"@charset\" must be the first rule in the file",
+							[]logger.MsgData{logger.RangeData(&p.source, logger.Range{Loc: locs[len(locs)-1]},
+								"This rule cannot come before a \"@charset\" rule")})
+						didWarnAboutCharset = true
+					}
+
+				case *css_ast.RAtImport:
+					if !didWarnAboutImport {
+					importLoop:
+						for i, before := range rules {
+							switch before.(type) {
+							case *css_ast.RAtCharset, *css_ast.RAtImport:
+							default:
+								p.log.AddRangeWarningWithNotes(&p.source, first, "All \"@import\" rules must come first",
+									[]logger.MsgData{logger.RangeData(&p.source, logger.Range{Loc: locs[i]},
+										"This rule cannot come before an \"@import\" rule")})
+								didWarnAboutImport = true
+								break importLoop
+							}
+						}
+					}
+				}
+			}
+
+			rules = append(rules, rule)
+			if context.isTopLevel {
+				locs = append(locs, first.Loc)
+			}
 			continue
 
 		case css_lexer.TCDO, css_lexer.TCDC:
@@ -173,12 +214,20 @@ func (p *parser) parseListOfRules(context ruleContext) []css_ast.R {
 			}
 		}
 
+		if context.isTopLevel {
+			locs = append(locs, p.current().Range.Loc)
+		}
 		if context.parseSelectors {
 			rules = append(rules, p.parseSelectorRule())
 		} else {
 			rules = append(rules, p.parseQualifiedRuleFrom(p.index, false /* isAlreadyInvalid */))
 		}
 	}
+
+	if p.options.MangleSyntax {
+		rules = removeEmptyRules(rules)
+	}
+	return rules
 }
 
 func (p *parser) parseListOfDeclarations() (list []css_ast.R) {
@@ -189,6 +238,9 @@ func (p *parser) parseListOfDeclarations() (list []css_ast.R) {
 
 		case css_lexer.TEndOfFile, css_lexer.TCloseBrace:
 			p.processDeclarations(list)
+			if p.options.MangleSyntax {
+				list = removeEmptyRules(list)
+			}
 			return
 
 		case css_lexer.TAtKeyword:
@@ -206,6 +258,32 @@ func (p *parser) parseListOfDeclarations() (list []css_ast.R) {
 	}
 }
 
+func removeEmptyRules(rules []css_ast.R) []css_ast.R {
+	end := 0
+	for _, rule := range rules {
+		switch r := rule.(type) {
+		case *css_ast.RAtKeyframes:
+			if len(r.Blocks) == 0 {
+				continue
+			}
+
+		case *css_ast.RKnownAt:
+			if len(r.Rules) == 0 {
+				continue
+			}
+
+		case *css_ast.RSelector:
+			if len(r.Rules) == 0 {
+				continue
+			}
+		}
+
+		rules[end] = rule
+		end++
+	}
+	return rules[:end]
+}
+
 func (p *parser) parseURLOrString() (string, logger.Range, bool) {
 	t := p.current()
 	switch t.Kind {
@@ -254,7 +332,45 @@ var specialAtRules = map[string]atRuleKind{
 	"font-face": atRuleDeclarations,
 	"page":      atRuleDeclarations,
 
-	"document": atRuleInheritContext,
+	// These go inside "@page": https://www.w3.org/TR/css-page-3/#syntax-page-selector
+	"bottom-center":       atRuleDeclarations,
+	"bottom-left-corner":  atRuleDeclarations,
+	"bottom-left":         atRuleDeclarations,
+	"bottom-right-corner": atRuleDeclarations,
+	"bottom-right":        atRuleDeclarations,
+	"left-bottom":         atRuleDeclarations,
+	"left-middle":         atRuleDeclarations,
+	"left-top":            atRuleDeclarations,
+	"right-bottom":        atRuleDeclarations,
+	"right-middle":        atRuleDeclarations,
+	"right-top":           atRuleDeclarations,
+	"top-center":          atRuleDeclarations,
+	"top-left-corner":     atRuleDeclarations,
+	"top-left":            atRuleDeclarations,
+	"top-right-corner":    atRuleDeclarations,
+	"top-right":           atRuleDeclarations,
+
+	// These properties are very deprecated and appear to only be useful for
+	// mobile versions of internet explorer (which may no longer exist?), but
+	// they are used by the https://ant.design/ design system so we recognize
+	// them to avoid the warning.
+	//
+	//   Documentation: https://developer.mozilla.org/en-US/docs/Web/CSS/@viewport
+	//   Discussion: https://github.com/w3c/csswg-drafts/issues/4766
+	//
+	"viewport":     atRuleDeclarations,
+	"-ms-viewport": atRuleDeclarations,
+
+	// This feature has been removed from the web because it's actively harmful.
+	// However, there is one exception where "@-moz-document url-prefix() {" is
+	// accepted by Firefox to basically be an "if Firefox" conditional rule.
+	//
+	//   Documentation: https://developer.mozilla.org/en-US/docs/Web/CSS/@document
+	//   Discussion: https://bugzilla.mozilla.org/show_bug.cgi?id=1035091
+	//
+	"document":      atRuleInheritContext,
+	"-moz-document": atRuleInheritContext,
+
 	"media":    atRuleInheritContext,
 	"scope":    atRuleInheritContext,
 	"supports": atRuleInheritContext,
@@ -289,34 +405,38 @@ func (p *parser) parseAtRule(context atRuleContext) css_ast.R {
 		}
 		p.expect(css_lexer.TString)
 
-	case "namespace":
-		kind = atRuleEmpty
-		p.eat(css_lexer.TWhitespace)
-		prefix := ""
-		if p.peek(css_lexer.TIdent) {
-			prefix = p.decoded()
-			p.advance()
-			p.eat(css_lexer.TWhitespace)
-		}
-		if path, _, ok := p.expectURLOrString(); ok {
-			p.eat(css_lexer.TWhitespace)
-			p.expect(css_lexer.TSemicolon)
-			return &css_ast.RAtNamespace{Prefix: prefix, Path: path}
-		}
-
 	case "import":
 		kind = atRuleEmpty
 		p.eat(css_lexer.TWhitespace)
 		if path, r, ok := p.expectURLOrString(); ok {
-			p.eat(css_lexer.TWhitespace)
+			importConditionsStart := p.index
+			for p.current().Kind != css_lexer.TSemicolon && p.current().Kind != css_lexer.TEndOfFile {
+				p.parseComponentValue()
+			}
+			importConditions := p.convertTokens(p.tokens[importConditionsStart:p.index])
+			kind := ast.ImportAt
+
+			// Insert or remove whitespace before the first token
+			if len(importConditions) > 0 {
+				kind = ast.ImportAtConditional
+				if p.options.RemoveWhitespace {
+					importConditions[0].Whitespace &= ^css_ast.WhitespaceBefore
+				} else {
+					importConditions[0].Whitespace |= css_ast.WhitespaceBefore
+				}
+			}
+
 			p.expect(css_lexer.TSemicolon)
 			importRecordIndex := uint32(len(p.importRecords))
 			p.importRecords = append(p.importRecords, ast.ImportRecord{
-				Kind:  ast.ImportAt,
+				Kind:  kind,
 				Path:  logger.Path{Text: path},
 				Range: r,
 			})
-			return &css_ast.RAtImport{ImportRecordIndex: importRecordIndex}
+			return &css_ast.RAtImport{
+				ImportRecordIndex: importRecordIndex,
+				ImportConditions:  importConditions,
+			}
 		}
 
 	case "keyframes", "-webkit-keyframes", "-moz-keyframes", "-ms-keyframes", "-o-keyframes":
@@ -395,10 +515,14 @@ func (p *parser) parseAtRule(context atRuleContext) css_ast.R {
 					if p.expect(css_lexer.TOpenBrace) {
 						rules := p.parseListOfDeclarations()
 						p.expect(css_lexer.TCloseBrace)
-						blocks = append(blocks, css_ast.KeyframeBlock{
-							Selectors: selectors,
-							Rules:     rules,
-						})
+
+						// "@keyframes { from {} to { color: red } }" => "@keyframes { to { color: red } }"
+						if !p.options.MangleSyntax || len(rules) > 0 {
+							blocks = append(blocks, css_ast.KeyframeBlock{
+								Selectors: selectors,
+								Rules:     rules,
+							})
+						}
 					}
 				}
 			}
@@ -420,7 +544,27 @@ func (p *parser) parseAtRule(context atRuleContext) css_ast.R {
 		// https://developer.mozilla.org/en-US/docs/Web/CSS/At-rule. Deprecated
 		// and Firefox-only at-rules have been removed.
 		if kind == atRuleUnknown {
-			p.log.AddRangeWarning(&p.source, atRange, fmt.Sprintf("%q is not a known rule name", "@"+atToken))
+			if atToken == "namespace" {
+				// CSS namespaces are a weird feature that appears to only really be
+				// useful for styling XML. And the world has moved on from XHTML to
+				// HTML5 so pretty much no one uses CSS namespaces anymore. They are
+				// also complicated to support in a bundler because CSS namespaces are
+				// file-scoped, which means:
+				//
+				// * Default namespaces can be different in different files, in which
+				//   case some default namespaces would have to be converted to prefixed
+				//   namespaces to avoid collisions.
+				//
+				// * Prefixed namespaces from different files can use the same name, in
+				//   which case some prefixed namespaces would need to be renamed to
+				//   avoid collisions.
+				//
+				// Instead of implementing all of that for an extremely obscure feature,
+				// CSS namespaces are just explicitly not supported.
+				p.log.AddRangeWarning(&p.source, atRange, "\"@namespace\" rules are not supported")
+			} else {
+				p.log.AddRangeWarning(&p.source, atRange, fmt.Sprintf("%q is not a known rule name", "@"+atToken))
+			}
 		}
 	}
 
@@ -484,49 +628,47 @@ prelude:
 	default:
 		// Otherwise, parse an unknown rule
 		p.parseBlock(css_lexer.TOpenBrace, css_lexer.TCloseBrace)
-		block := p.convertTokensWithImports(p.tokens[blockStart:p.index])
+		block, _ := p.convertTokensHelper(p.tokens[blockStart:p.index], css_lexer.TEndOfFile, convertTokensOpts{allowImports: true})
 		return &css_ast.RUnknownAt{AtToken: atToken, Prelude: prelude, Block: block}
 	}
 }
 
 func (p *parser) convertTokens(tokens []css_lexer.Token) []css_ast.Token {
-	result, _ := p.convertTokensHelper(tokens, css_lexer.TEndOfFile, false)
+	result, _ := p.convertTokensHelper(tokens, css_lexer.TEndOfFile, convertTokensOpts{})
 	return result
 }
 
-func (p *parser) convertTokensWithImports(tokens []css_lexer.Token) []css_ast.Token {
-	result, _ := p.convertTokensHelper(tokens, css_lexer.TEndOfFile, true)
-	return result
+type convertTokensOpts struct {
+	allowImports       bool
+	verbatimWhitespace bool
 }
 
-func (p *parser) convertTokensHelper(tokens []css_lexer.Token, close css_lexer.T, allowImports bool) ([]css_ast.Token, []css_lexer.Token) {
+func (p *parser) convertTokensHelper(tokens []css_lexer.Token, close css_lexer.T, opts convertTokensOpts) ([]css_ast.Token, []css_lexer.Token) {
 	var result []css_ast.Token
+	var nextWhitespace css_ast.WhitespaceFlags
+
 loop:
 	for len(tokens) > 0 {
 		t := tokens[0]
 		tokens = tokens[1:]
-		token := css_ast.Token{
-			Kind: t.Kind,
-			Text: t.DecodedText(p.source.Contents),
+		if t.Kind == close {
+			break loop
 		}
+		token := css_ast.Token{
+			Kind:       t.Kind,
+			Text:       t.DecodedText(p.source.Contents),
+			Whitespace: nextWhitespace,
+		}
+		nextWhitespace = 0
 
 		switch t.Kind {
 		case css_lexer.TWhitespace:
 			if last := len(result) - 1; last >= 0 {
-				result[last].HasWhitespaceAfter = true
+				result[last].Whitespace |= css_ast.WhitespaceAfter
 			}
+			nextWhitespace = css_ast.WhitespaceBefore
 			continue
 
-		case css_lexer.TComma:
-			// Assume that whitespace can always be removed before a comma
-			if last := len(result) - 1; last >= 0 {
-				result[last].HasWhitespaceAfter = false
-			}
-
-			// Assume whitespace can always be added after a comma (it will be
-			// automatically omitted by the printer if we're minifying)
-			token.HasWhitespaceAfter = true
-
 		case css_lexer.TNumber:
 			if p.options.MangleSyntax {
 				if text, ok := mangleNumber(token.Text); ok {
@@ -557,14 +699,19 @@ loop:
 				Kind:     ast.ImportURL,
 				Path:     logger.Path{Text: token.Text},
 				Range:    t.Range,
-				IsUnused: !allowImports,
+				IsUnused: !opts.allowImports,
 			})
 			token.Text = ""
 
 		case css_lexer.TFunction:
 			var nested []css_ast.Token
 			original := tokens
-			nested, tokens = p.convertTokensHelper(tokens, css_lexer.TCloseParen, allowImports)
+			nestedOpts := opts
+			if token.Text == "var" {
+				// CSS variables require verbatim whitespace for correctness
+				nestedOpts.verbatimWhitespace = true
+			}
+			nested, tokens = p.convertTokensHelper(tokens, css_lexer.TCloseParen, nestedOpts)
 			token.Children = &nested
 
 			// Treat a URL function call with a string just like a URL token
@@ -577,33 +724,83 @@ loop:
 					Kind:     ast.ImportURL,
 					Path:     logger.Path{Text: nested[0].Text},
 					Range:    original[0].Range,
-					IsUnused: !allowImports,
+					IsUnused: !opts.allowImports,
 				})
 			}
 
 		case css_lexer.TOpenParen:
 			var nested []css_ast.Token
-			nested, tokens = p.convertTokensHelper(tokens, css_lexer.TCloseParen, allowImports)
+			nested, tokens = p.convertTokensHelper(tokens, css_lexer.TCloseParen, opts)
 			token.Children = &nested
 
 		case css_lexer.TOpenBrace:
 			var nested []css_ast.Token
-			nested, tokens = p.convertTokensHelper(tokens, css_lexer.TCloseBrace, allowImports)
+			nested, tokens = p.convertTokensHelper(tokens, css_lexer.TCloseBrace, opts)
+
+			// Pretty-printing: insert leading and trailing whitespace when not minifying
+			if !opts.verbatimWhitespace && !p.options.RemoveWhitespace && len(nested) > 0 {
+				nested[0].Whitespace |= css_ast.WhitespaceBefore
+				nested[len(nested)-1].Whitespace |= css_ast.WhitespaceAfter
+			}
+
 			token.Children = &nested
 
 		case css_lexer.TOpenBracket:
 			var nested []css_ast.Token
-			nested, tokens = p.convertTokensHelper(tokens, css_lexer.TCloseBracket, allowImports)
+			nested, tokens = p.convertTokensHelper(tokens, css_lexer.TCloseBracket, opts)
 			token.Children = &nested
-
-		default:
-			if t.Kind == close {
-				break loop
-			}
 		}
 
 		result = append(result, token)
 	}
+
+	if !opts.verbatimWhitespace {
+		for i := range result {
+			token := &result[i]
+
+			// Always remove leading and trailing whitespace
+			if i == 0 {
+				token.Whitespace &= ^css_ast.WhitespaceBefore
+			}
+			if i+1 == len(result) {
+				token.Whitespace &= ^css_ast.WhitespaceAfter
+			}
+
+			switch token.Kind {
+			case css_lexer.TComma:
+				// Assume that whitespace can always be removed before a comma
+				token.Whitespace &= ^css_ast.WhitespaceBefore
+				if i > 0 {
+					result[i-1].Whitespace &= ^css_ast.WhitespaceAfter
+				}
+
+				// Assume whitespace can always be added after a comma
+				if p.options.RemoveWhitespace {
+					token.Whitespace &= ^css_ast.WhitespaceAfter
+					if i+1 < len(result) {
+						result[i+1].Whitespace &= ^css_ast.WhitespaceBefore
+					}
+				} else {
+					token.Whitespace |= css_ast.WhitespaceAfter
+					if i+1 < len(result) {
+						result[i+1].Whitespace |= css_ast.WhitespaceBefore
+					}
+				}
+			}
+		}
+	}
+
+	// Insert an explicit whitespace token if we're in verbatim mode and all
+	// tokens were whitespace. In this case there is no token to attach the
+	// whitespace before/after flags so this is the only way to represent this.
+	// This is the only case where this function generates an explicit whitespace
+	// token. It represents whitespace as flags in all other cases.
+	if opts.verbatimWhitespace && len(result) == 0 && nextWhitespace == css_ast.WhitespaceBefore {
+		result = append(result, css_ast.Token{
+			Kind: css_lexer.TWhitespace,
+		})
+	}
+
 	return result, tokens
 }
 
@@ -619,6 +816,9 @@ func mangleNumber(t string) (string, bool) {
 		// Remove the decimal point if it's unnecessary
 		if dot+1 == len(t) {
 			t = t[:dot]
+			if t == "" || t == "+" || t == "-" {
+				t += "0"
+			}
 		} else {
 			// Remove a leading zero
 			if len(t) >= 3 && t[0] == '0' && t[1] == '.' && t[2] >= '0' && t[2] <= '9' {
@@ -724,34 +924,49 @@ stop:
 		}
 	}
 
-	// Remove leading and trailing whitespace from the value
-	value := trimWhitespace(p.tokens[valueStart:p.index])
+	keyToken := p.tokens[keyStart]
+	keyText := keyToken.DecodedText(p.source.Contents)
+	value := p.tokens[valueStart:p.index]
+	verbatimWhitespace := strings.HasPrefix(keyText, "--")
 
 	// Remove trailing "!important"
 	important := false
-	if last := len(value) - 1; last >= 0 {
-		if t := value[last]; t.Kind == css_lexer.TIdent && strings.EqualFold(t.DecodedText(p.source.Contents), "important") {
-			i := len(value) - 2
-			if i >= 0 && value[i].Kind == css_lexer.TWhitespace {
-				i--
-			}
-			if i >= 0 && value[i].Kind == css_lexer.TDelimExclamation {
-				if i >= 1 && value[i-1].Kind == css_lexer.TWhitespace {
-					i--
-				}
-				value = value[:i]
-				important = true
-			}
+	i := len(value) - 1
+	if i >= 0 && value[i].Kind == css_lexer.TWhitespace {
+		i--
+	}
+	if i >= 0 && value[i].Kind == css_lexer.TIdent && strings.EqualFold(value[i].DecodedText(p.source.Contents), "important") {
+		i--
+		if i >= 0 && value[i].Kind == css_lexer.TWhitespace {
+			i--
+		}
+		if i >= 0 && value[i].Kind == css_lexer.TDelimExclamation {
+			value = value[:i]
+			important = true
+		}
+	}
+
+	result, _ := p.convertTokensHelper(value, css_lexer.TEndOfFile, convertTokensOpts{
+		allowImports: true,
+
+		// CSS variables require verbatim whitespace for correctness
+		verbatimWhitespace: verbatimWhitespace,
+	})
+
+	// Insert or remove whitespace before the first token
+	if !verbatimWhitespace && len(result) > 0 {
+		if p.options.RemoveWhitespace {
+			result[0].Whitespace &= ^css_ast.WhitespaceBefore
+		} else {
+			result[0].Whitespace |= css_ast.WhitespaceBefore
 		}
 	}
 
-	keyToken := p.tokens[keyStart]
-	keyText := keyToken.DecodedText(p.source.Contents)
 	return &css_ast.RDeclaration{
 		Key:       css_ast.KnownDeclarations[keyText],
 		KeyText:   keyText,
 		KeyRange:  keyToken.Range,
-		Value:     p.convertTokensWithImports(value),
+		Value:     result,
 		Important: important,
 	}
 }
@@ -790,13 +1005,3 @@ func (p *parser) parseBlock(open css_lexer.T, close css_lexer.T) {
 		}
 	}
 }
-
-func trimWhitespace(tokens []css_lexer.Token) []css_lexer.Token {
-	if len(tokens) > 0 && tokens[0].Kind == css_lexer.TWhitespace {
-		tokens = tokens[1:]
-	}
-	if i := len(tokens) - 1; i >= 0 && tokens[i].Kind == css_lexer.TWhitespace {
-		tokens = tokens[:i]
-	}
-	return tokens
-}
