@@ -39,6 +39,7 @@ func expectPrintedCommon(t *testing.T, name string, contents string, expected st
 		log := logger.NewDeferLog()
 		tree := Parse(log, test.SourceForTest(contents), Options{
 			MangleSyntax:           options.MangleSyntax,
+			RemoveWhitespace:       options.RemoveWhitespace,
 			UnsupportedCSSFeatures: options.UnsupportedCSSFeatures,
 		})
 		msgs := log.Done()
@@ -49,7 +50,9 @@ func expectPrintedCommon(t *testing.T, name string, contents string, expected st
 			}
 		}
 		assertEqual(t, text, "")
-		css := css_printer.Print(tree, css_printer.Options{})
+		css := css_printer.Print(tree, css_printer.Options{
+			RemoveWhitespace: options.RemoveWhitespace,
+		})
 		assertEqual(t, string(css), expected)
 	})
 }
@@ -73,6 +76,22 @@ func expectPrintedMangle(t *testing.T, contents string, expected string) {
 	})
 }
 
+func expectPrintedLowerMangle(t *testing.T, contents string, expected string) {
+	t.Helper()
+	expectPrintedCommon(t, contents+" [mangle]", contents, expected, config.Options{
+		UnsupportedCSSFeatures: ^compat.CSSFeature(0),
+		MangleSyntax:           true,
+	})
+}
+
+func expectPrintedMangleMinify(t *testing.T, contents string, expected string) {
+	t.Helper()
+	expectPrintedCommon(t, contents+" [mangle, minify]", contents, expected, config.Options{
+		MangleSyntax:     true,
+		RemoveWhitespace: true,
+	})
+}
+
 func TestEscapes(t *testing.T) {
 	// TIdent
 	expectPrinted(t, "a { value: id\\65nt }", "a {\n  value: ident;\n}\n")
@@ -255,18 +274,46 @@ func TestString(t *testing.T) {
 
 func TestNumber(t *testing.T) {
 	for _, ext := range []string{"", "%", "px+"} {
+		expectPrinted(t, "a { width: .0"+ext+"; }", "a {\n  width: .0"+ext+";\n}\n")
+		expectPrinted(t, "a { width: .00"+ext+"; }", "a {\n  width: .00"+ext+";\n}\n")
+		expectPrinted(t, "a { width: .10"+ext+"; }", "a {\n  width: .10"+ext+";\n}\n")
 		expectPrinted(t, "a { width: 0."+ext+"; }", "a {\n  width: 0."+ext+";\n}\n")
+		expectPrinted(t, "a { width: 0.0"+ext+"; }", "a {\n  width: 0.0"+ext+";\n}\n")
 		expectPrinted(t, "a { width: 0.1"+ext+"; }", "a {\n  width: 0.1"+ext+";\n}\n")
+
+		expectPrinted(t, "a { width: +.0"+ext+"; }", "a {\n  width: +.0"+ext+";\n}\n")
+		expectPrinted(t, "a { width: +.00"+ext+"; }", "a {\n  width: +.00"+ext+";\n}\n")
+		expectPrinted(t, "a { width: +.10"+ext+"; }", "a {\n  width: +.10"+ext+";\n}\n")
 		expectPrinted(t, "a { width: +0."+ext+"; }", "a {\n  width: +0."+ext+";\n}\n")
+		expectPrinted(t, "a { width: +0.0"+ext+"; }", "a {\n  width: +0.0"+ext+";\n}\n")
 		expectPrinted(t, "a { width: +0.1"+ext+"; }", "a {\n  width: +0.1"+ext+";\n}\n")
+
+		expectPrinted(t, "a { width: -.0"+ext+"; }", "a {\n  width: -.0"+ext+";\n}\n")
+		expectPrinted(t, "a { width: -.00"+ext+"; }", "a {\n  width: -.00"+ext+";\n}\n")
+		expectPrinted(t, "a { width: -.10"+ext+"; }", "a {\n  width: -.10"+ext+";\n}\n")
 		expectPrinted(t, "a { width: -0."+ext+"; }", "a {\n  width: -0."+ext+";\n}\n")
+		expectPrinted(t, "a { width: -0.0"+ext+"; }", "a {\n  width: -0.0"+ext+";\n}\n")
 		expectPrinted(t, "a { width: -0.1"+ext+"; }", "a {\n  width: -0.1"+ext+";\n}\n")
 
+		expectPrintedMangle(t, "a { width: .0"+ext+"; }", "a {\n  width: 0"+ext+";\n}\n")
+		expectPrintedMangle(t, "a { width: .00"+ext+"; }", "a {\n  width: 0"+ext+";\n}\n")
+		expectPrintedMangle(t, "a { width: .10"+ext+"; }", "a {\n  width: .1"+ext+";\n}\n")
 		expectPrintedMangle(t, "a { width: 0."+ext+"; }", "a {\n  width: 0"+ext+";\n}\n")
+		expectPrintedMangle(t, "a { width: 0.0"+ext+"; }", "a {\n  width: 0"+ext+";\n}\n")
 		expectPrintedMangle(t, "a { width: 0.1"+ext+"; }", "a {\n  width: .1"+ext+";\n}\n")
+
+		expectPrintedMangle(t, "a { width: +.0"+ext+"; }", "a {\n  width: +0"+ext+";\n}\n")
+		expectPrintedMangle(t, "a { width: +.00"+ext+"; }", "a {\n  width: +0"+ext+";\n}\n")
+		expectPrintedMangle(t, "a { width: +.10"+ext+"; }", "a {\n  width: +.1"+ext+";\n}\n")
 		expectPrintedMangle(t, "a { width: +0."+ext+"; }", "a {\n  width: +0"+ext+";\n}\n")
+		expectPrintedMangle(t, "a { width: +0.0"+ext+"; }", "a {\n  width: +0"+ext+";\n}\n")
 		expectPrintedMangle(t, "a { width: +0.1"+ext+"; }", "a {\n  width: +.1"+ext+";\n}\n")
+
+		expectPrintedMangle(t, "a { width: -.0"+ext+"; }", "a {\n  width: -0"+ext+";\n}\n")
+		expectPrintedMangle(t, "a { width: -.00"+ext+"; }", "a {\n  width: -0"+ext+";\n}\n")
+		expectPrintedMangle(t, "a { width: -.10"+ext+"; }", "a {\n  width: -.1"+ext+";\n}\n")
 		expectPrintedMangle(t, "a { width: -0."+ext+"; }", "a {\n  width: -0"+ext+";\n}\n")
+		expectPrintedMangle(t, "a { width: -0.0"+ext+"; }", "a {\n  width: -0"+ext+";\n}\n")
 		expectPrintedMangle(t, "a { width: -0.1"+ext+"; }", "a {\n  width: -.1"+ext+";\n}\n")
 	}
 }
@@ -429,6 +476,9 @@ func TestColorRGBA(t *testing.T) {
 	expectPrintedMangle(t, "a { color: rgba(1% 2% 3% / 50%) }", "a {\n  color: #0305087f;\n}\n")
 	expectPrintedMangle(t, "a { color: rgba(1%, 2%, 3%, 0.5) }", "a {\n  color: #03050880;\n}\n")
 	expectPrintedMangle(t, "a { color: rgba(1%, 2%, 3%, 50%) }", "a {\n  color: #0305087f;\n}\n")
+
+	expectPrintedLowerMangle(t, "a { color: rgb(1, 2, 3, 0.4) }", "a {\n  color: rgba(1, 2, 3, .4);\n}\n")
+	expectPrintedLowerMangle(t, "a { color: rgba(1, 2, 3, 40%) }", "a {\n  color: rgba(1, 2, 3, .4);\n}\n")
 }
 
 func TestColorHSLA(t *testing.T) {
@@ -443,6 +493,9 @@ func TestColorHSLA(t *testing.T) {
 
 	expectPrintedMangle(t, "a { color: hsl(30 25% 50% / 50%) }", "a {\n  color: #9f80607f;\n}\n")
 	expectPrintedMangle(t, "a { color: hsla(30 25% 50% / 50%) }", "a {\n  color: #9f80607f;\n}\n")
+
+	expectPrintedLowerMangle(t, "a { color: hsl(1, 2%, 3%, 0.4) }", "a {\n  color: rgba(8, 8, 7, .4);\n}\n")
+	expectPrintedLowerMangle(t, "a { color: hsla(1, 2%, 3%, 40%) }", "a {\n  color: rgba(8, 8, 7, .4);\n}\n")
 }
 
 func TestLowerColor(t *testing.T) {
@@ -593,6 +646,10 @@ func TestSelector(t *testing.T) {
 	expectParseError(t, "#-0 {}", "<stdin>: warning: Unexpected \"#-0\"\n")
 	expectParseError(t, "div#0 {}", "<stdin>: warning: Unexpected \"#0\"\n")
 	expectParseError(t, "div#-0 {}", "<stdin>: warning: Unexpected \"#-0\"\n")
+
+	// Make sure '-' and '\\' consume an ident-like token instead of a name
+	expectParseError(t, "_:-ms-lang(x) {}", "")
+	expectParseError(t, "_:\\ms-lang(x) {}", "")
 }
 
 func TestNestedSelector(t *testing.T) {
@@ -634,6 +691,90 @@ func TestAtRule(t *testing.T) {
 	expectParseError(t, "@", "<stdin>: warning: Unexpected \"@\"\n")
 	expectParseError(t, "@;", "<stdin>: warning: Unexpected \"@\"\n")
 	expectParseError(t, "@{}", "<stdin>: warning: Unexpected \"@\"\n")
+
+	expectPrinted(t, "@viewport { width: 100vw }", "@viewport {\n  width: 100vw;\n}\n")
+	expectPrinted(t, "@-ms-viewport { width: 100vw }", "@-ms-viewport {\n  width: 100vw;\n}\n")
+
+	expectPrinted(t, "@document url(\"https://www.example.com/\") { h1 { color: green } }",
+		"@document url(https://www.example.com/) {\n  h1 {\n    color: green;\n  }\n}\n")
+	expectPrinted(t, "@-moz-document url-prefix() { h1 { color: green } }",
+		"@-moz-document url-prefix() {\n  h1 {\n    color: green;\n  }\n}\n")
+
+	// https://www.w3.org/TR/css-page-3/#syntax-page-selector
+	expectPrinted(t, `
+		@page :first { margin: 0 }
+		@page {
+			@top-left-corner { content: 'tlc' }
+			@top-left { content: 'tl' }
+			@top-center { content: 'tc' }
+			@top-right { content: 'tr' }
+			@top-right-corner { content: 'trc' }
+			@bottom-left-corner { content: 'blc' }
+			@bottom-left { content: 'bl' }
+			@bottom-center { content: 'bc' }
+			@bottom-right { content: 'br' }
+			@bottom-right-corner { content: 'brc' }
+			@left-top { content: 'lt' }
+			@left-middle { content: 'lm' }
+			@left-bottom { content: 'lb' }
+			@right-top { content: 'rt' }
+			@right-middle { content: 'rm' }
+			@right-bottom { content: 'rb' }
+		}
+	`, `@page :first {
+  margin: 0;
+}
+@page {
+  @top-left-corner {
+    content: "tlc";
+  }
+  @top-left {
+    content: "tl";
+  }
+  @top-center {
+    content: "tc";
+  }
+  @top-right {
+    content: "tr";
+  }
+  @top-right-corner {
+    content: "trc";
+  }
+  @bottom-left-corner {
+    content: "blc";
+  }
+  @bottom-left {
+    content: "bl";
+  }
+  @bottom-center {
+    content: "bc";
+  }
+  @bottom-right {
+    content: "br";
+  }
+  @bottom-right-corner {
+    content: "brc";
+  }
+  @left-top {
+    content: "lt";
+  }
+  @left-middle {
+    content: "lm";
+  }
+  @left-bottom {
+    content: "lb";
+  }
+  @right-top {
+    content: "rt";
+  }
+  @right-middle {
+    content: "rm";
+  }
+  @right-bottom {
+    content: "rb";
+  }
+}
+`)
 }
 
 func TestAtCharset(t *testing.T) {
@@ -651,48 +792,6 @@ func TestAtCharset(t *testing.T) {
 	expectParseError(t, "@charset \"UTF-8\"{}", "<stdin>: warning: Expected \";\" but found \"{\"\n")
 }
 
-func TestAtNamespace(t *testing.T) {
-	expectPrinted(t, "@namespace\"http://www.com\";", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace \"http://www.com\";", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace \"http://www.com\" ;", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace url();", "@namespace \"\";\n")
-	expectPrinted(t, "@namespace url(http://www.com);", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace url(http://www.com) ;", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace url(\"http://www.com\");", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace url(\"http://www.com\") ;", "@namespace \"http://www.com\";\n")
-
-	expectPrinted(t, "@namespace ns\"http://www.com\";", "@namespace ns \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns \"http://www.com\";", "@namespace ns \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns \"http://www.com\" ;", "@namespace ns \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns url();", "@namespace ns \"\";\n")
-	expectPrinted(t, "@namespace ns url(http://www.com);", "@namespace ns \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns url(http://www.com) ;", "@namespace ns \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns url(\"http://www.com\");", "@namespace ns \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns url(\"http://www.com\") ;", "@namespace ns \"http://www.com\";\n")
-
-	expectParseError(t, "@namespace;", "<stdin>: warning: Expected URL token but found \";\"\n")
-	expectParseError(t, "@namespace \"http://www.com\"", "<stdin>: warning: Expected \";\" but found end of file\n")
-	expectParseError(t, "@namespace url(\"http://www.com\";", "<stdin>: warning: Expected \")\" but found \";\"\n")
-	expectParseError(t, "@namespace noturl(\"http://www.com\");", "<stdin>: warning: Expected URL token but found \"noturl(\"\n")
-	expectParseError(t, "@namespace url(", `<stdin>: warning: Expected URL token but found bad URL token
-<stdin>: error: Expected ")" to end URL token
-<stdin>: warning: Expected ";" but found end of file
-`)
-
-	expectParseError(t, "@namespace ns;", "<stdin>: warning: Expected URL token but found \";\"\n")
-	expectParseError(t, "@namespace ns \"http://www.com\"", "<stdin>: warning: Expected \";\" but found end of file\n")
-	expectParseError(t, "@namespace ns url(\"http://www.com\";", "<stdin>: warning: Expected \")\" but found \";\"\n")
-	expectParseError(t, "@namespace ns noturl(\"http://www.com\");", "<stdin>: warning: Expected URL token but found \"noturl(\"\n")
-	expectParseError(t, "@namespace ns url(", `<stdin>: warning: Expected URL token but found bad URL token
-<stdin>: error: Expected ")" to end URL token
-<stdin>: warning: Expected ";" but found end of file
-`)
-
-	expectParseError(t, "@namespace \"http://www.com\" {}", `<stdin>: warning: Expected ";"
-<stdin>: warning: Unexpected "{"
-`)
-}
-
 func TestAtImport(t *testing.T) {
 	expectPrinted(t, "@import\"foo.css\";", "@import \"foo.css\";\n")
 	expectPrinted(t, "@import \"foo.css\";", "@import \"foo.css\";\n")
@@ -702,6 +801,8 @@ func TestAtImport(t *testing.T) {
 	expectPrinted(t, "@import url(foo.css) ;", "@import \"foo.css\";\n")
 	expectPrinted(t, "@import url(\"foo.css\");", "@import \"foo.css\";\n")
 	expectPrinted(t, "@import url(\"foo.css\") ;", "@import \"foo.css\";\n")
+	expectPrinted(t, "@import url(\"foo.css\") print;", "@import \"foo.css\" print;\n")
+	expectPrinted(t, "@import url(\"foo.css\") screen and (orientation:landscape);", "@import \"foo.css\" screen and (orientation:landscape);\n")
 
 	expectParseError(t, "@import;", "<stdin>: warning: Expected URL token but found \";\"\n")
 	expectParseError(t, "@import ;", "<stdin>: warning: Expected URL token but found \";\"\n")
@@ -713,9 +814,7 @@ func TestAtImport(t *testing.T) {
 <stdin>: warning: Expected ";" but found end of file
 `)
 
-	expectParseError(t, "@import \"foo.css\" {}", `<stdin>: warning: Expected ";"
-<stdin>: warning: Unexpected "{"
-`)
+	expectParseError(t, "@import \"foo.css\" {}", "<stdin>: warning: Expected \";\" but found end of file\n")
 }
 
 func TestAtKeyframes(t *testing.T) {
@@ -752,3 +851,40 @@ func TestAtKeyframes(t *testing.T) {
 	expectParseError(t, "@keyframes name { 1 ,2 {} }", "<stdin>: warning: Expected percentage but found \"1\"\n<stdin>: warning: Expected percentage but found \"2\"\n")
 	expectParseError(t, "@keyframes name { 1%,,2% {} }", "<stdin>: warning: Expected percentage but found \",\"\n")
 }
+
+func TestAtRuleValidation(t *testing.T) {
+	expectParseError(t, "a {} @charset \"UTF-8\";",
+		"<stdin>: warning: \"@charset\" must be the first rule in the file\n"+
+			"<stdin>: note: This rule cannot come before a \"@charset\" rule\n")
+
+	expectParseError(t, "a {} @import \"foo\";",
+		"<stdin>: warning: All \"@import\" rules must come first\n"+
+			"<stdin>: note: This rule cannot come before an \"@import\" rule\n")
+}
+
+func TestEmptyRule(t *testing.T) {
+	expectPrinted(t, "div {}", "div {\n}\n")
+	expectPrinted(t, "@media screen {}", "@media screen {\n}\n")
+	expectPrinted(t, "@page { @top-left {} }", "@page {\n  @top-left {\n  }\n}\n")
+	expectPrinted(t, "@keyframes test { from {} to {} }", "@keyframes test {\n  from {\n  }\n  to {\n  }\n}\n")
+
+	expectPrintedMangle(t, "div {}", "")
+	expectPrintedMangle(t, "@media screen {}", "")
+	expectPrintedMangle(t, "@page { @top-left {} }", "")
+	expectPrintedMangle(t, "@keyframes test { from {} to {} }", "")
+
+	expectPrinted(t, "$invalid {}", "$invalid {\n}\n")
+	expectPrinted(t, "@page { color: red; @top-left {} }", "@page {\n  color: red;\n  @top-left {\n  }\n}\n")
+	expectPrinted(t, "@keyframes test { from {} to { color: red } }", "@keyframes test {\n  from {\n  }\n  to {\n    color: red;\n  }\n}\n")
+	expectPrinted(t, "@keyframes test { from { color: red } to {} }", "@keyframes test {\n  from {\n    color: red;\n  }\n  to {\n  }\n}\n")
+
+	expectPrintedMangle(t, "$invalid {}", "$invalid {\n}\n")
+	expectPrintedMangle(t, "@page { color: red; @top-left {} }", "@page {\n  color: red;\n}\n")
+	expectPrintedMangle(t, "@keyframes test { from {} to { color: red } }", "@keyframes test {\n  to {\n    color: red;\n  }\n}\n")
+	expectPrintedMangle(t, "@keyframes test { from { color: red } to {} }", "@keyframes test {\n  0% {\n    color: red;\n  }\n}\n")
+
+	expectPrintedMangleMinify(t, "$invalid {}", "$invalid{}")
+	expectPrintedMangleMinify(t, "@page { color: red; @top-left {} }", "@page{color:red}")
+	expectPrintedMangleMinify(t, "@keyframes test { from {} to { color: red } }", "@keyframes test{to{color:red}}")
+	expectPrintedMangleMinify(t, "@keyframes test { from { color: red } to {} }", "@keyframes test{0%{color:red}}")
+}
