@@ -13,7 +13,7 @@ import (
 const quoteForURL rune = -1
 
 type printer struct {
-	Options
+	options       Options
 	importRecords []ast.ImportRecord
 	sb            strings.Builder
 }
@@ -25,7 +25,7 @@ type Options struct {
 
 func Print(tree css_ast.AST, options Options) string {
 	p := printer{
-		Options:       options,
+		options:       options,
 		importRecords: tree.ImportRecords,
 	}
 	for _, rule := range tree.Rules {
@@ -34,8 +34,8 @@ func Print(tree css_ast.AST, options Options) string {
 	return p.sb.String()
 }
 
-func (p *printer) printRule(rule css_ast.R, indent int, omitTrailingSemicolon bool) {
-	if !p.RemoveWhitespace {
+func (p *printer) printRule(rule css_ast.R, indent int32, omitTrailingSemicolon bool) {
+	if !p.options.RemoveWhitespace {
 		p.printIndent(indent)
 	}
 
@@ -48,26 +48,14 @@ func (p *printer) printRule(rule css_ast.R, indent int, omitTrailingSemicolon bo
 		p.printQuotedWithQuote(r.Encoding, '"')
 		p.print(";")
 
-	case *css_ast.RAtNamespace:
-		if r.Prefix != "" {
-			p.print("@namespace ")
-			p.printIdent(r.Prefix, identNormal, canDiscardWhitespaceAfter)
-		} else {
-			p.print("@namespace")
-		}
-		if !p.RemoveWhitespace {
-			p.print(" ")
-		}
-		p.printQuoted(r.Path)
-		p.print(";")
-
 	case *css_ast.RAtImport:
-		if p.RemoveWhitespace {
+		if p.options.RemoveWhitespace {
 			p.print("@import")
 		} else {
 			p.print("@import ")
 		}
 		p.printQuoted(p.importRecords[r.ImportRecordIndex].Path.Text)
+		p.printTokens(r.ImportConditions, printTokensOpts{})
 		p.print(";")
 
 	case *css_ast.RAtKeyframes:
@@ -79,22 +67,22 @@ func (p *printer) printRule(rule css_ast.R, indent int, omitTrailingSemicolon bo
 		} else {
 			p.printIdent(r.Name, identNormal, canDiscardWhitespaceAfter)
 		}
-		if !p.RemoveWhitespace {
+		if !p.options.RemoveWhitespace {
 			p.print(" ")
 		}
-		if p.RemoveWhitespace {
+		if p.options.RemoveWhitespace {
 			p.print("{")
 		} else {
 			p.print("{\n")
 		}
 		indent++
 		for _, block := range r.Blocks {
-			if !p.RemoveWhitespace {
+			if !p.options.RemoveWhitespace {
 				p.printIndent(indent)
 			}
 			for i, sel := range block.Selectors {
 				if i > 0 {
-					if p.RemoveWhitespace {
+					if p.options.RemoveWhitespace {
 						p.print(",")
 					} else {
 						p.print(", ")
@@ -102,16 +90,16 @@ func (p *printer) printRule(rule css_ast.R, indent int, omitTrailingSemicolon bo
 				}
 				p.print(sel)
 			}
-			if !p.RemoveWhitespace {
+			if !p.options.RemoveWhitespace {
 				p.print(" ")
 			}
 			p.printRuleBlock(block.Rules, indent)
-			if !p.RemoveWhitespace {
+			if !p.options.RemoveWhitespace {
 				p.print("\n")
 			}
 		}
 		indent--
-		if !p.RemoveWhitespace {
+		if !p.options.RemoveWhitespace {
 			p.printIndent(indent)
 		}
 		p.print("}")
@@ -123,11 +111,11 @@ func (p *printer) printRule(rule css_ast.R, indent int, omitTrailingSemicolon bo
 			whitespace = canDiscardWhitespaceAfter
 		}
 		p.printIdent(r.AtToken, identNormal, whitespace)
-		if !p.RemoveWhitespace || len(r.Prelude) > 0 {
+		if !p.options.RemoveWhitespace || len(r.Prelude) > 0 {
 			p.print(" ")
 		}
-		p.printTokens(r.Prelude)
-		if !p.RemoveWhitespace && len(r.Prelude) > 0 {
+		p.printTokens(r.Prelude, printTokensOpts{})
+		if !p.options.RemoveWhitespace && len(r.Prelude) > 0 {
 			p.print(" ")
 		}
 		p.printRuleBlock(r.Rules, indent)
@@ -139,43 +127,42 @@ func (p *printer) printRule(rule css_ast.R, indent int, omitTrailingSemicolon bo
 			whitespace = canDiscardWhitespaceAfter
 		}
 		p.printIdent(r.AtToken, identNormal, whitespace)
-		if (!p.RemoveWhitespace && r.Block != nil) || len(r.Prelude) > 0 {
+		if (!p.options.RemoveWhitespace && r.Block != nil) || len(r.Prelude) > 0 {
 			p.print(" ")
 		}
-		p.printTokens(r.Prelude)
-		if !p.RemoveWhitespace && r.Block != nil && len(r.Prelude) > 0 {
+		p.printTokens(r.Prelude, printTokensOpts{})
+		if !p.options.RemoveWhitespace && r.Block != nil && len(r.Prelude) > 0 {
 			p.print(" ")
 		}
 		if r.Block == nil {
 			p.print(";")
 		} else {
-			p.printTokens(r.Block)
+			p.printTokens(r.Block, printTokensOpts{})
 		}
 
 	case *css_ast.RSelector:
 		p.printComplexSelectors(r.Selectors, indent)
-		if !p.RemoveWhitespace {
+		if !p.options.RemoveWhitespace {
 			p.print(" ")
 		}
 		p.printRuleBlock(r.Rules, indent)
 
 	case *css_ast.RQualified:
-		p.printTokens(r.Prelude)
-		if !p.RemoveWhitespace {
+		hasWhitespaceAfter := p.printTokens(r.Prelude, printTokensOpts{})
+		if !hasWhitespaceAfter && !p.options.RemoveWhitespace {
 			p.print(" ")
 		}
 		p.printRuleBlock(r.Rules, indent)
 
 	case *css_ast.RDeclaration:
 		p.printIdent(r.KeyText, identNormal, canDiscardWhitespaceAfter)
-		if p.RemoveWhitespace {
-			p.print(":")
-		} else {
-			p.print(": ")
-		}
-		p.printTokens(r.Value)
+		p.print(":")
+		hasWhitespaceAfter := p.printTokens(r.Value, printTokensOpts{
+			indent:        indent,
+			isDeclaration: true,
+		})
 		if r.Important {
-			if !p.RemoveWhitespace {
+			if !hasWhitespaceAfter && !p.options.RemoveWhitespace && len(r.Value) > 0 {
 				p.print(" ")
 			}
 			p.print("!important")
@@ -185,7 +172,7 @@ func (p *printer) printRule(rule css_ast.R, indent int, omitTrailingSemicolon bo
 		}
 
 	case *css_ast.RBadDeclaration:
-		p.printTokens(r.Tokens)
+		p.printTokens(r.Tokens, printTokensOpts{})
 		if !omitTrailingSemicolon {
 			p.print(";")
 		}
@@ -194,33 +181,33 @@ func (p *printer) printRule(rule css_ast.R, indent int, omitTrailingSemicolon bo
 		panic("Internal error")
 	}
 
-	if !p.RemoveWhitespace {
+	if !p.options.RemoveWhitespace {
 		p.print("\n")
 	}
 }
 
-func (p *printer) printRuleBlock(rules []css_ast.R, indent int) {
-	if p.RemoveWhitespace {
+func (p *printer) printRuleBlock(rules []css_ast.R, indent int32) {
+	if p.options.RemoveWhitespace {
 		p.print("{")
 	} else {
 		p.print("{\n")
 	}
 
 	for i, decl := range rules {
-		omitTrailingSemicolon := p.RemoveWhitespace && i+1 == len(rules)
+		omitTrailingSemicolon := p.options.RemoveWhitespace && i+1 == len(rules)
 		p.printRule(decl, indent+1, omitTrailingSemicolon)
 	}
 
-	if !p.RemoveWhitespace {
+	if !p.options.RemoveWhitespace {
 		p.printIndent(indent)
 	}
 	p.print("}")
 }
 
-func (p *printer) printComplexSelectors(selectors []css_ast.ComplexSelector, indent int) {
+func (p *printer) printComplexSelectors(selectors []css_ast.ComplexSelector, indent int32) {
 	for i, complex := range selectors {
 		if i > 0 {
-			if p.RemoveWhitespace {
+			if p.options.RemoveWhitespace {
 				p.print(",")
 			} else {
 				p.print(",\n")
@@ -240,11 +227,11 @@ func (p *printer) printCompoundSelector(sel css_ast.CompoundSelector, isFirst bo
 	}
 
 	if sel.Combinator != "" {
-		if !p.RemoveWhitespace {
+		if !p.options.RemoveWhitespace {
 			p.print(" ")
 		}
 		p.print(sel.Combinator)
-		if !p.RemoveWhitespace {
+		if !p.options.RemoveWhitespace {
 			p.print(" ")
 		}
 	} else if !isFirst {
@@ -360,7 +347,7 @@ func (p *printer) printPseudoClassSelector(pseudo css_ast.SSPseudoClass, whitesp
 	if len(pseudo.Args) > 0 {
 		p.printIdent(pseudo.Name, identNormal, canDiscardWhitespaceAfter)
 		p.print("(")
-		p.printTokens(pseudo.Args)
+		p.printTokens(pseudo.Args, printTokensOpts{})
 		p.print(")")
 	} else {
 		p.printIdent(pseudo.Name, identNormal, whitespace)
@@ -479,7 +466,7 @@ func (p *printer) printQuotedWithQuote(text string, quote rune) {
 			}
 
 		default:
-			if p.ASCIIOnly && c >= 0x80 || c == '\uFEFF' {
+			if p.options.ASCIIOnly && c >= 0x80 || c == '\uFEFF' {
 				escape = escapeHex
 			}
 		}
@@ -511,7 +498,7 @@ func (p *printer) printIdent(text string, mode identMode, whitespace trailingWhi
 	for i, c := range text {
 		escape := escapeNone
 
-		if p.ASCIIOnly && c >= 0x80 {
+		if p.options.ASCIIOnly && c >= 0x80 {
 			escape = escapeHex
 		} else if c == '\r' || c == '\n' || c == '\f' || c == '\uFEFF' {
 			// Use a hexadecimal escape for characters that would be invalid escapes
@@ -557,15 +544,48 @@ func (p *printer) printIdent(text string, mode identMode, whitespace trailingWhi
 	}
 }
 
-func (p *printer) printIndent(indent int) {
-	for i := 0; i < indent; i++ {
+func (p *printer) printIndent(indent int32) {
+	for i, n := 0, int(indent); i < n; i++ {
 		p.sb.WriteString("  ")
 	}
 }
 
-func (p *printer) printTokens(tokens []css_ast.Token) {
+type printTokensOpts struct {
+	indent        int32
+	isDeclaration bool
+}
+
+func (p *printer) printTokens(tokens []css_ast.Token, opts printTokensOpts) bool {
+	hasWhitespaceAfter := len(tokens) > 0 && (tokens[0].Whitespace&css_ast.WhitespaceBefore) != 0
+
+	// Pretty-print long comma-separated declarations of 3 or more items
+	isMultiLineValue := false
+	if !p.options.RemoveWhitespace && opts.isDeclaration {
+		commaCount := 0
+		for _, t := range tokens {
+			if t.Kind == css_lexer.TComma {
+				commaCount++
+			}
+		}
+		isMultiLineValue = commaCount >= 2
+	}
+
 	for i, t := range tokens {
-		hasWhitespaceAfter := t.HasWhitespaceAfter && i+1 != len(tokens)
+		if t.Kind == css_lexer.TWhitespace {
+			hasWhitespaceAfter = true
+			continue
+		}
+		if hasWhitespaceAfter {
+			if isMultiLineValue && (i == 0 || tokens[i-1].Kind == css_lexer.TComma) {
+				p.print("\n")
+				p.printIndent(opts.indent + 1)
+			} else {
+				p.print(" ")
+			}
+		}
+		hasWhitespaceAfter = (t.Whitespace&css_ast.WhitespaceAfter) != 0 ||
+			(i+1 < len(tokens) && (tokens[i+1].Whitespace&css_ast.WhitespaceBefore) != 0)
+
 		whitespace := mayNeedWhitespaceAfter
 		if !hasWhitespaceAfter {
 			whitespace = canDiscardWhitespaceAfter
@@ -605,13 +625,7 @@ func (p *printer) printTokens(tokens []css_ast.Token) {
 		}
 
 		if t.Children != nil {
-			children := *t.Children
-
-			if t.Kind == css_lexer.TOpenBrace && !p.RemoveWhitespace && len(children) > 0 {
-				p.print(" ")
-			}
-
-			p.printTokens(children)
+			p.printTokens(*t.Children, printTokensOpts{})
 
 			switch t.Kind {
 			case css_lexer.TFunction:
@@ -621,22 +635,15 @@ func (p *printer) printTokens(tokens []css_ast.Token) {
 				p.print(")")
 
 			case css_lexer.TOpenBrace:
-				if !p.RemoveWhitespace && len(children) > 0 {
-					p.print(" ")
-				}
 				p.print("}")
 
 			case css_lexer.TOpenBracket:
 				p.print("]")
 			}
 		}
-
-		if hasWhitespaceAfter {
-			if t.Kind == css_lexer.TComma && p.RemoveWhitespace {
-				// Assume that whitespace can always be removed after a comma
-			} else {
-				p.print(" ")
-			}
-		}
 	}
+	if hasWhitespaceAfter {
+		p.print(" ")
+	}
+	return hasWhitespaceAfter
 }
