@@ -20,7 +20,9 @@ func expectPrintedCommon(t *testing.T, name string, contents string, expected st
 	t.Run(name, func(t *testing.T) {
 		t.Helper()
 		log := logger.NewDeferLog()
-		tree := css_parser.Parse(log, test.SourceForTest(contents), css_parser.Options{})
+		tree := css_parser.Parse(log, test.SourceForTest(contents), css_parser.Options{
+			RemoveWhitespace: options.RemoveWhitespace,
+		})
 		msgs := log.Done()
 		text := ""
 		for _, msg := range msgs {
@@ -163,6 +165,109 @@ func TestDeclaration(t *testing.T) {
 	expectPrintedMinify(t, "* { unknown: x ( a + b ) }", "*{unknown:x (a + b)}")
 	expectPrintedMinify(t, "* { unknown: x ( a - b ) }", "*{unknown:x (a - b)}")
 	expectPrintedMinify(t, "* { unknown: x ( a , b ) }", "*{unknown:x (a,b)}")
+
+	// Pretty-print long lists in declarations
+	expectPrinted(t, "a { b: c, d }", "a {\n  b: c, d;\n}\n")
+	expectPrinted(t, "a { b: c, (d, e) }", "a {\n  b: c, (d, e);\n}\n")
+	expectPrinted(t, "a { b: c, d, e }", "a {\n  b:\n    c,\n    d,\n    e;\n}\n")
+	expectPrinted(t, "a { b: c, (d, e), f }", "a {\n  b:\n    c,\n    (d, e),\n    f;\n}\n")
+
+	expectPrintedMinify(t, "a { b: c, d }", "a{b:c,d}")
+	expectPrintedMinify(t, "a { b: c, (d, e) }", "a{b:c,(d,e)}")
+	expectPrintedMinify(t, "a { b: c, d, e }", "a{b:c,d,e}")
+	expectPrintedMinify(t, "a { b: c, (d, e), f }", "a{b:c,(d,e),f}")
+}
+
+func TestVerbatimWhitespace(t *testing.T) {
+	expectPrinted(t, "*{--x:}", "* {\n  --x:;\n}\n")
+	expectPrinted(t, "*{--x: }", "* {\n  --x: ;\n}\n")
+	expectPrinted(t, "* { --x:; }", "* {\n  --x:;\n}\n")
+	expectPrinted(t, "* { --x: ; }", "* {\n  --x: ;\n}\n")
+
+	expectPrintedMinify(t, "*{--x:}", "*{--x:}")
+	expectPrintedMinify(t, "*{--x: }", "*{--x: }")
+	expectPrintedMinify(t, "* { --x:; }", "*{--x:}")
+	expectPrintedMinify(t, "* { --x: ; }", "*{--x: }")
+
+	expectPrinted(t, "*{--x:!important}", "* {\n  --x:!important;\n}\n")
+	expectPrinted(t, "*{--x: !important}", "* {\n  --x: !important;\n}\n")
+	expectPrinted(t, "*{ --x:!important }", "* {\n  --x:!important;\n}\n")
+	expectPrinted(t, "*{ --x: !important }", "* {\n  --x: !important;\n}\n")
+	expectPrinted(t, "* { --x:!important; }", "* {\n  --x:!important;\n}\n")
+	expectPrinted(t, "* { --x: !important; }", "* {\n  --x: !important;\n}\n")
+	expectPrinted(t, "* { --x:! important ; }", "* {\n  --x:!important;\n}\n")
+	expectPrinted(t, "* { --x: ! important ; }", "* {\n  --x: !important;\n}\n")
+
+	expectPrintedMinify(t, "*{--x:!important}", "*{--x:!important}")
+	expectPrintedMinify(t, "*{--x: !important}", "*{--x: !important}")
+	expectPrintedMinify(t, "*{ --x:!important }", "*{--x:!important}")
+	expectPrintedMinify(t, "*{ --x: !important }", "*{--x: !important}")
+	expectPrintedMinify(t, "* { --x:!important; }", "*{--x:!important}")
+	expectPrintedMinify(t, "* { --x: !important; }", "*{--x: !important}")
+	expectPrintedMinify(t, "* { --x:! important ; }", "*{--x:!important}")
+	expectPrintedMinify(t, "* { --x: ! important ; }", "*{--x: !important}")
+
+	expectPrinted(t, "* { --x:y; }", "* {\n  --x:y;\n}\n")
+	expectPrinted(t, "* { --x: y; }", "* {\n  --x: y;\n}\n")
+	expectPrinted(t, "* { --x:y ; }", "* {\n  --x:y ;\n}\n")
+	expectPrinted(t, "* { --x:y, ; }", "* {\n  --x:y, ;\n}\n")
+	expectPrinted(t, "* { --x: var(y,); }", "* {\n  --x: var(y,);\n}\n")
+	expectPrinted(t, "* { --x: var(y, ); }", "* {\n  --x: var(y, );\n}\n")
+
+	expectPrintedMinify(t, "* { --x:y; }", "*{--x:y}")
+	expectPrintedMinify(t, "* { --x: y; }", "*{--x: y}")
+	expectPrintedMinify(t, "* { --x:y ; }", "*{--x:y }")
+	expectPrintedMinify(t, "* { --x:y, ; }", "*{--x:y, }")
+	expectPrintedMinify(t, "* { --x: var(y,); }", "*{--x: var(y,)}")
+	expectPrintedMinify(t, "* { --x: var(y, ); }", "*{--x: var(y, )}")
+
+	expectPrinted(t, "* { --x:(y); }", "* {\n  --x:(y);\n}\n")
+	expectPrinted(t, "* { --x:(y) ; }", "* {\n  --x:(y) ;\n}\n")
+	expectPrinted(t, "* { --x: (y); }", "* {\n  --x: (y);\n}\n")
+	expectPrinted(t, "* { --x:(y ); }", "* {\n  --x:(y );\n}\n")
+	expectPrinted(t, "* { --x:( y); }", "* {\n  --x:( y);\n}\n")
+
+	expectPrintedMinify(t, "* { --x:(y); }", "*{--x:(y)}")
+	expectPrintedMinify(t, "* { --x:(y) ; }", "*{--x:(y) }")
+	expectPrintedMinify(t, "* { --x: (y); }", "*{--x: (y)}")
+	expectPrintedMinify(t, "* { --x:(y ); }", "*{--x:(y )}")
+	expectPrintedMinify(t, "* { --x:( y); }", "*{--x:( y)}")
+
+	expectPrinted(t, "* { --x:f(y); }", "* {\n  --x:f(y);\n}\n")
+	expectPrinted(t, "* { --x:f(y) ; }", "* {\n  --x:f(y) ;\n}\n")
+	expectPrinted(t, "* { --x: f(y); }", "* {\n  --x: f(y);\n}\n")
+	expectPrinted(t, "* { --x:f(y ); }", "* {\n  --x:f(y );\n}\n")
+	expectPrinted(t, "* { --x:f( y); }", "* {\n  --x:f( y);\n}\n")
+
+	expectPrintedMinify(t, "* { --x:f(y); }", "*{--x:f(y)}")
+	expectPrintedMinify(t, "* { --x:f(y) ; }", "*{--x:f(y) }")
+	expectPrintedMinify(t, "* { --x: f(y); }", "*{--x: f(y)}")
+	expectPrintedMinify(t, "* { --x:f(y ); }", "*{--x:f(y )}")
+	expectPrintedMinify(t, "* { --x:f( y); }", "*{--x:f( y)}")
+
+	expectPrinted(t, "* { --x:[y]; }", "* {\n  --x:[y];\n}\n")
+	expectPrinted(t, "* { --x:[y] ; }", "* {\n  --x:[y] ;\n}\n")
+	expectPrinted(t, "* { --x: [y]; }", "* {\n  --x: [y];\n}\n")
+	expectPrinted(t, "* { --x:[y ]; }", "* {\n  --x:[y ];\n}\n")
+	expectPrinted(t, "* { --x:[ y]; }", "* {\n  --x:[ y];\n}\n")
+
+	expectPrintedMinify(t, "* { --x:[y]; }", "*{--x:[y]}")
+	expectPrintedMinify(t, "* { --x:[y] ; }", "*{--x:[y] }")
+	expectPrintedMinify(t, "* { --x: [y]; }", "*{--x: [y]}")
+	expectPrintedMinify(t, "* { --x:[y ]; }", "*{--x:[y ]}")
+	expectPrintedMinify(t, "* { --x:[ y]; }", "*{--x:[ y]}")
+
+	expectPrinted(t, "* { --x:{y}; }", "* {\n  --x:{y};\n}\n")
+	expectPrinted(t, "* { --x:{y} ; }", "* {\n  --x:{y} ;\n}\n")
+	expectPrinted(t, "* { --x: {y}; }", "* {\n  --x: {y};\n}\n")
+	expectPrinted(t, "* { --x:{y }; }", "* {\n  --x:{y };\n}\n")
+	expectPrinted(t, "* { --x:{ y}; }", "* {\n  --x:{ y};\n}\n")
+
+	expectPrintedMinify(t, "* { --x:{y}; }", "*{--x:{y}}")
+	expectPrintedMinify(t, "* { --x:{y} ; }", "*{--x:{y} }")
+	expectPrintedMinify(t, "* { --x: {y}; }", "*{--x: {y}}")
+	expectPrintedMinify(t, "* { --x:{y }; }", "*{--x:{y }}")
+	expectPrintedMinify(t, "* { --x:{ y}; }", "*{--x:{ y}}")
 }
 
 func TestAtRule(t *testing.T) {
@@ -189,36 +294,18 @@ func TestAtCharset(t *testing.T) {
 	expectPrintedMinify(t, "@charset \"UTF-8\";", "@charset \"UTF-8\";")
 }
 
-func TestAtNamespace(t *testing.T) {
-	expectPrinted(t, "@namespace\"http://www.com\";", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace \"http://www.com\";", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace url(http://www.com);", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace url(\"http://www.com\");", "@namespace \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns\"http://www.com\";", "@namespace ns \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns \"http://www.com\";", "@namespace ns \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns url(http://www.com);", "@namespace ns \"http://www.com\";\n")
-	expectPrinted(t, "@namespace ns url(\"http://www.com\");", "@namespace ns \"http://www.com\";\n")
-
-	expectPrintedMinify(t, "@namespace\"http://www.com\";", "@namespace\"http://www.com\";")
-	expectPrintedMinify(t, "@namespace \"http://www.com\";", "@namespace\"http://www.com\";")
-	expectPrintedMinify(t, "@namespace url(http://www.com);", "@namespace\"http://www.com\";")
-	expectPrintedMinify(t, "@namespace url(\"http://www.com\");", "@namespace\"http://www.com\";")
-	expectPrintedMinify(t, "@namespace ns\"http://www.com\";", "@namespace ns\"http://www.com\";")
-	expectPrintedMinify(t, "@namespace ns \"http://www.com\";", "@namespace ns\"http://www.com\";")
-	expectPrintedMinify(t, "@namespace ns url(http://www.com);", "@namespace ns\"http://www.com\";")
-	expectPrintedMinify(t, "@namespace ns url(\"http://www.com\");", "@namespace ns\"http://www.com\";")
-}
-
 func TestAtImport(t *testing.T) {
 	expectPrinted(t, "@import\"foo.css\";", "@import \"foo.css\";\n")
 	expectPrinted(t, "@import \"foo.css\";", "@import \"foo.css\";\n")
 	expectPrinted(t, "@import url(foo.css);", "@import \"foo.css\";\n")
 	expectPrinted(t, "@import url(\"foo.css\");", "@import \"foo.css\";\n")
+	expectPrinted(t, "@import url(\"foo.css\") print;", "@import \"foo.css\" print;\n")
 
 	expectPrintedMinify(t, "@import\"foo.css\";", "@import\"foo.css\";")
 	expectPrintedMinify(t, "@import \"foo.css\";", "@import\"foo.css\";")
 	expectPrintedMinify(t, "@import url(foo.css);", "@import\"foo.css\";")
 	expectPrintedMinify(t, "@import url(\"foo.css\");", "@import\"foo.css\";")
+	expectPrintedMinify(t, "@import url(\"foo.css\") print;", "@import\"foo.css\"print;")
 }
 
 func TestAtKeyframes(t *testing.T) {
