@@ -1,6 +1,7 @@
 package js_ast
 
 import (
+	"math"
 	"sort"
 
 	"github.com/evanw/esbuild/internal/ast"
@@ -319,6 +320,9 @@ type Fn struct {
 	IsGenerator bool
 	HasRestArg  bool
 	HasIfScope  bool
+
+	// This is true if the function is a method
+	IsUniqueFormalParameters bool
 }
 
 type FnBody struct {
@@ -327,6 +331,7 @@ type FnBody struct {
 }
 
 type Class struct {
+	ClassKeyword logger.Range
 	TSDecorators []Expr
 	Name         *LocRef
 	Extends      *Expr
@@ -378,8 +383,10 @@ type Expr struct {
 type E interface{ isExpr() }
 
 type EArray struct {
-	Items        []Expr
-	IsSingleLine bool
+	Items            []Expr
+	CommaAfterSpread logger.Loc
+	IsSingleLine     bool
+	IsParenthesized  bool
 }
 
 type EUnary struct {
@@ -501,6 +508,11 @@ type EClass struct{ Class Class }
 type EIdentifier struct {
 	Ref Ref
 
+	// If we're inside a "with" statement, this identifier may be a property
+	// access. In that case it would be incorrect to remove this identifier since
+	// the property access may be a getter or setter with side effects.
+	MustKeepDueToWithStmt bool
+
 	// If true, this identifier is known to not have a side effect (i.e. to not
 	// throw an exception) when referenced. If false, this identifier may or may
 	// not have side effects when referenced. This is used to allow the removal
@@ -534,6 +546,11 @@ type EIdentifier struct {
 // doing this instead of opt-out.
 type EImportIdentifier struct {
 	Ref Ref
+
+	// If true, this was originally an identifier expression such as "foo". If
+	// false, this could potentially have been a member access expression such
+	// as "ns.foo" off of an imported namespace object.
+	WasOriginallyIdentifier bool
 }
 
 // This is similar to EIdentifier but it represents class-private fields and
@@ -556,14 +573,19 @@ type ENumber struct{ Value float64 }
 type EBigInt struct{ Value string }
 
 type EObject struct {
-	Properties   []Property
-	IsSingleLine bool
+	Properties       []Property
+	CommaAfterSpread logger.Loc
+	IsSingleLine     bool
+	IsParenthesized  bool
 }
 
 type ESpread struct{ Value Expr }
 
+// This is used for both strings and no-substitution template literals to reduce
+// the number of cases that need to be checked for string optimization code
 type EString struct {
 	Value          []uint16
+	LegacyOctalLoc logger.Loc
 	PreferTemplate bool
 }
 
@@ -575,10 +597,11 @@ type TemplatePart struct {
 }
 
 type ETemplate struct {
-	Tag     *Expr
-	Head    []uint16
-	HeadRaw string // This is only filled out for tagged template literals
-	Parts   []TemplatePart
+	Tag            *Expr
+	Head           []uint16
+	HeadRaw        string // This is only filled out for tagged template literals
+	Parts          []TemplatePart
+	LegacyOctalLoc logger.Loc
 }
 
 type ERegExp struct{ Value string }
@@ -608,7 +631,7 @@ type ERequireResolve struct {
 
 type EImport struct {
 	Expr              Expr
-	ImportRecordIndex *uint32
+	ImportRecordIndex ast.Index32
 
 	// Comments inside "import()" expressions have special meaning for Webpack.
 	// Preserving comments inside these expressions makes it possible to use
@@ -656,6 +679,18 @@ func (*ERequire) isExpr()           {}
 func (*ERequireResolve) isExpr()    {}
 func (*EImport) isExpr()            {}
 
+func IsOptionalChain(value Expr) bool {
+	switch e := value.Data.(type) {
+	case *EDot:
+		return e.OptionalChain != OptionalChainNone
+	case *EIndex:
+		return e.OptionalChain != OptionalChainNone
+	case *ECall:
+		return e.OptionalChain != OptionalChainNone
+	}
+	return false
+}
+
 func Assign(a Expr, b Expr) Expr {
 	return Expr{Loc: a.Loc, Data: &EBinary{Op: BinOpAssign, Left: a, Right: b}}
 }
@@ -682,6 +717,24 @@ func Not(expr Expr) Expr {
 // that is undesired.
 func MaybeSimplifyNot(expr Expr) (Expr, bool) {
 	switch e := expr.Data.(type) {
+	case *ENull, *EUndefined:
+		return Expr{Loc: expr.Loc, Data: &EBoolean{Value: true}}, true
+
+	case *EBoolean:
+		return Expr{Loc: expr.Loc, Data: &EBoolean{Value: !e.Value}}, true
+
+	case *ENumber:
+		return Expr{Loc: expr.Loc, Data: &EBoolean{Value: e.Value == 0 || math.IsNaN(e.Value)}}, true
+
+	case *EBigInt:
+		return Expr{Loc: expr.Loc, Data: &EBoolean{Value: e.Value == "0"}}, true
+
+	case *EString:
+		return Expr{Loc: expr.Loc, Data: &EBoolean{Value: len(e.Value) == 0}}, true
+
+	case *EFunction, *EArrow, *ERegExp:
+		return Expr{Loc: expr.Loc, Data: &EBoolean{Value: false}}, true
+
 	case *EUnary:
 		// "!!!a" => "!a"
 		if e.Op == UnOpNot && IsBooleanValue(e.Value) {
@@ -844,15 +897,20 @@ func JoinWithLeftAssociativeOp(op OpCode, a Expr, b Expr) Expr {
 }
 
 func JoinWithComma(a Expr, b Expr) Expr {
+	if a.Data == nil {
+		return b
+	}
+	if b.Data == nil {
+		return a
+	}
 	return Expr{Loc: a.Loc, Data: &EBinary{Op: BinOpComma, Left: a, Right: b}}
 }
 
-func JoinAllWithComma(all []Expr) Expr {
-	result := all[0]
-	for _, value := range all[1:] {
+func JoinAllWithComma(all []Expr) (result Expr) {
+	for _, value := range all {
 		result = JoinWithComma(result, value)
 	}
-	return result
+	return
 }
 
 type ExprOrStmt struct {
@@ -879,13 +937,15 @@ type SEmpty struct{}
 type STypeScript struct{}
 
 type SComment struct {
-	Text string
+	Text           string
+	IsLegalComment bool
 }
 
 type SDebugger struct{}
 
 type SDirective struct {
-	Value []uint16
+	Value          []uint16
+	LegacyOctalLoc logger.Loc
 }
 
 type SExportClause struct {
@@ -1030,6 +1090,7 @@ type Finally struct {
 }
 
 type STry struct {
+	BodyLoc logger.Loc
 	Body    []Stmt
 	Catch   *Catch
 	Finally *Finally
@@ -1254,9 +1315,6 @@ const (
 
 	// This annotates all other symbols that don't have special behavior.
 	SymbolOther
-
-	// This symbol causes a compile error when referenced
-	SymbolError
 )
 
 func (kind SymbolKind) IsPrivate() bool {
@@ -1290,6 +1348,10 @@ func (kind SymbolKind) IsHoistedOrFunction() bool {
 	return kind.IsHoisted() || kind == SymbolGeneratorOrAsyncFunction
 }
 
+func (kind SymbolKind) IsFunction() bool {
+	return kind == SymbolHoistedFunction || kind == SymbolGeneratorOrAsyncFunction
+}
+
 var InvalidRef Ref = Ref{^uint32(0), ^uint32(0)}
 
 // Files are parsed in parallel for speed. We want to allow each parser to
@@ -1297,15 +1359,15 @@ var InvalidRef Ref = Ref{^uint32(0), ^uint32(0)}
 // able to quickly merge symbol tables from all files into one giant symbol
 // table.
 //
-// We can accomplish both goals by giving each symbol ID two parts: an outer
+// We can accomplish both goals by giving each symbol ID two parts: a source
 // index that is unique to the parser goroutine, and an inner index that
 // increments as the parser generates new symbol IDs. Then a symbol map can
-// be an array of arrays indexed first by outer index, then by inner index.
+// be an array of arrays indexed first by source index, then by inner index.
 // The maps can be merged quickly by creating a single outer array containing
 // all inner arrays from all parsed files.
 type Ref struct {
-	OuterIndex uint32
-	InnerIndex uint32
+	SourceIndex uint32
+	InnerIndex  uint32
 }
 
 type ImportItemStatus uint8
@@ -1354,8 +1416,7 @@ type Symbol struct {
 	UseCountEstimate uint32
 
 	// This is for generating cross-chunk imports and exports for code splitting.
-	// It's stored as one's complement so the zero value is invalid.
-	ChunkIndex uint32
+	ChunkIndex ast.Index32
 
 	// This is used for minification. Symbols that are declared in sibling scopes
 	// can share a name. A good heuristic (from Google Closure Compiler) is to
@@ -1367,9 +1428,8 @@ type Symbol struct {
 	//   function y(a, b, c) { ... }
 	//
 	// The parser fills this in for symbols inside nested scopes. There are three
-	// slot namespaces: regular symbols, label symbols, and private symbols. This
-	// is stored as one's complement so the zero value is invalid.
-	NestedScopeSlot uint32
+	// slot namespaces: regular symbols, label symbols, and private symbols.
+	NestedScopeSlot ast.Index32
 
 	Kind SymbolKind
 
@@ -1396,6 +1456,59 @@ type Symbol struct {
 	// avoid this. We also need to be able to replace such import items with
 	// undefined, which this status is also used for.
 	ImportItemStatus ImportItemStatus
+
+	// Sometimes we lower private symbols even if they are supported. For example,
+	// consider the following TypeScript code:
+	//
+	//   class Foo {
+	//     #foo = 123
+	//     bar = this.#foo
+	//   }
+	//
+	// If "useDefineForClassFields: false" is set in "tsconfig.json", then "bar"
+	// must use assignment semantics instead of define semantics. We can compile
+	// that to this code:
+	//
+	//   class Foo {
+	//     constructor() {
+	//       this.#foo = 123;
+	//       this.bar = this.#foo;
+	//     }
+	//     #foo;
+	//   }
+	//
+	// However, we can't do the same for static fields:
+	//
+	//   class Foo {
+	//     static #foo = 123
+	//     static bar = this.#foo
+	//   }
+	//
+	// Compiling these static fields to something like this would be invalid:
+	//
+	//   class Foo {
+	//     static #foo;
+	//   }
+	//   Foo.#foo = 123;
+	//   Foo.bar = Foo.#foo;
+	//
+	// Thus "#foo" must be lowered even though it's supported. Another case is
+	// when we're converting top-level class declarations to class expressions
+	// to avoid the TDZ and the class shadowing symbol is referenced within the
+	// class body:
+	//
+	//   class Foo {
+	//     static #foo = Foo
+	//   }
+	//
+	// This cannot be converted into something like this:
+	//
+	//   var Foo = class {
+	//     static #foo;
+	//   };
+	//   Foo.#foo = Foo;
+	//
+	PrivateSymbolMustBeLowered bool
 }
 
 type SlotNamespace uint8
@@ -1476,6 +1589,31 @@ type Scope struct {
 	// inside that scope can be renamed. We conservatively assume that the
 	// evaluated code might reference anything that it has access to.
 	ContainsDirectEval bool
+
+	// This is to help forbid "arguments" inside class body scopes
+	ForbidArguments bool
+
+	StrictMode StrictModeKind
+}
+
+type StrictModeKind uint8
+
+const (
+	SloppyMode StrictModeKind = iota
+	ExplicitStrictMode
+	ImplicitStrictModeImport
+	ImplicitStrictModeExport
+	ImplicitStrictModeTopLevelAwait
+	ImplicitStrictModeClass
+)
+
+func (s *Scope) RecursiveSetStrictMode(kind StrictModeKind) {
+	if s.StrictMode == SloppyMode {
+		s.StrictMode = kind
+		for _, child := range s.Children {
+			child.RecursiveSetStrictMode(kind)
+		}
+	}
 }
 
 type SymbolMap struct {
@@ -1486,7 +1624,7 @@ type SymbolMap struct {
 	// single inner array, so you can join the maps together by just make a
 	// single outer array containing all of the inner arrays. See the comment on
 	// "Ref" for more detail.
-	Outer [][]Symbol
+	SymbolsForSource [][]Symbol
 }
 
 func NewSymbolMap(sourceCount int) SymbolMap {
@@ -1494,7 +1632,41 @@ func NewSymbolMap(sourceCount int) SymbolMap {
 }
 
 func (sm SymbolMap) Get(ref Ref) *Symbol {
-	return &sm.Outer[ref.OuterIndex][ref.InnerIndex]
+	return &sm.SymbolsForSource[ref.SourceIndex][ref.InnerIndex]
+}
+
+type ExportsKind uint8
+
+const (
+	// This file doesn't have any kind of export, so it's impossible to say what
+	// kind of file this is. An empty file is in this category, for example.
+	ExportsNone ExportsKind = iota
+
+	// The exports are stored on "module" and/or "exports". Calling "require()"
+	// on this module returns "module.exports". All imports to this module are
+	// allowed but may return undefined.
+	ExportsCommonJS
+
+	// All export names are known explicitly. Calling "require()" on this module
+	// generates an exports object (stored in "exports") with getters for the
+	// export names. Named imports to this module are only allowed if they are
+	// in the set of export names.
+	ExportsESM
+
+	// Some export names are known explicitly, but others fall back to a dynamic
+	// run-time object. This is necessary when using the "export * from" syntax
+	// with either a CommonJS module or an external module (i.e. a module whose
+	// export names are not known at compile-time).
+	//
+	// Calling "require()" on this module generates an exports object (stored in
+	// "exports") with getters for the export names. All named imports to this
+	// module are allowed. Direct named imports reference the corresponding export
+	// directly. Other imports go through property accesses on "exports".
+	ExportsESMWithDynamicFallback
+)
+
+func (kind ExportsKind) IsDynamic() bool {
+	return kind == ExportsCommonJS || kind == ExportsESMWithDynamicFallback
 }
 
 type AST struct {
@@ -1508,10 +1680,13 @@ type AST struct {
 	HasTopLevelReturn bool
 	UsesExportsRef    bool
 	UsesModuleRef     bool
+	ExportsKind       ExportsKind
 
-	// This is a list of ES6 features
-	HasES6Imports bool
-	HasES6Exports bool
+	// This is a list of ES6 features. They are ranges instead of booleans so
+	// that they can be used in log messages. Check to see if "Len > 0".
+	ImportKeyword        logger.Range // Does not include TypeScript-specific syntax or "import()"
+	ExportKeyword        logger.Range // Does not include TypeScript-specific syntax
+	TopLevelAwaitKeyword logger.Range
 
 	Hashbang    string
 	Directive   string
@@ -1637,18 +1812,6 @@ func (minifier *NameMinifier) NumberToMinifiedName(i int) string {
 	return name
 }
 
-func (ast *AST) HasCommonJSFeatures() bool {
-	return ast.UsesCommonJSExports() || ast.HasTopLevelReturn
-}
-
-func (ast *AST) UsesCommonJSExports() bool {
-	return ast.UsesExportsRef || ast.UsesModuleRef
-}
-
-func (ast *AST) HasES6Syntax() bool {
-	return ast.HasES6Imports || ast.HasES6Exports
-}
-
 type NamedImport struct {
 	// Parts within this file that use this import
 	LocalPartsWithUses []uint32
@@ -1658,6 +1821,12 @@ type NamedImport struct {
 	NamespaceRef      Ref
 	ImportRecordIndex uint32
 
+	// If true, the alias refers to the entire export namespace object of a
+	// module. This is no longer represented as an alias called "*" because of
+	// the upcoming "Arbitrary module namespace identifier names" feature:
+	// https://github.com/tc39/ecma262/pull/2154
+	AliasIsStar bool
+
 	// It's useful to flag exported imports because if they are in a TypeScript
 	// file, we can't tell if they are a type or a value.
 	IsExported bool
@@ -1691,22 +1860,26 @@ type Part struct {
 
 	// The indices of the other parts in this file that are needed if this part
 	// is needed.
-	LocalDependencies map[uint32]bool
+	Dependencies []Dependency
 
 	// If true, this part can be removed if none of the declared symbols are
 	// used. If the file containing this part is imported, then all parts that
 	// don't have this flag enabled must be included.
 	CanBeRemovedIfUnused bool
 
-	// If true, this is the automatically-generated part for this file's ES6
-	// exports. It may hold the "var exports = {};" statement and also the
-	// "__export(exports, { ... })" call to initialize the getters.
-	IsNamespaceExport bool
-
 	// This is used for generated parts that we don't want to be present if they
 	// aren't needed. This enables tree shaking for these parts even if global
 	// tree shaking isn't enabled.
 	ForceTreeShaking bool
+
+	// This is true if this file has been marked as live by the tree shaking
+	// algorithm.
+	IsLive bool
+}
+
+type Dependency struct {
+	SourceIndex uint32
+	PartIndex   uint32
 }
 
 type DeclaredSymbol struct {
@@ -1716,7 +1889,6 @@ type DeclaredSymbol struct {
 
 type SymbolUse struct {
 	CountEstimate uint32
-	IsAssigned    bool
 }
 
 // Returns the canonical ref that represents the ref for the provided symbol.
@@ -1743,7 +1915,7 @@ func FollowSymbols(symbols SymbolMap, ref Ref) Ref {
 // but reading from a map is. Calling "FollowAllSymbols" first ensures that
 // all mutation is done up front.
 func FollowAllSymbols(symbols SymbolMap) {
-	for sourceIndex, inner := range symbols.Outer {
+	for sourceIndex, inner := range symbols.SymbolsForSource {
 		for symbolIndex := range inner {
 			FollowSymbols(symbols, Ref{uint32(sourceIndex), uint32(symbolIndex)})
 		}
@@ -1773,6 +1945,7 @@ func MergeSymbols(symbols SymbolMap, old Ref, new Ref) Ref {
 	oldSymbol.Link = new
 	newSymbol.UseCountEstimate += oldSymbol.UseCountEstimate
 	if oldSymbol.MustNotBeRenamed {
+		newSymbol.OriginalName = oldSymbol.OriginalName
 		newSymbol.MustNotBeRenamed = true
 	}
 	return new
@@ -1831,3 +2004,58 @@ func EnsureValidIdentifier(base string) string {
 	}
 	return string(bytes)
 }
+
+func ConvertBindingToExpr(binding Binding, wrapIdentifier func(logger.Loc, Ref) Expr) Expr {
+	loc := binding.Loc
+
+	switch b := binding.Data.(type) {
+	case *BMissing:
+		return Expr{Loc: loc, Data: &EMissing{}}
+
+	case *BIdentifier:
+		if wrapIdentifier != nil {
+			return wrapIdentifier(loc, b.Ref)
+		}
+		return Expr{Loc: loc, Data: &EIdentifier{Ref: b.Ref}}
+
+	case *BArray:
+		exprs := make([]Expr, len(b.Items))
+		for i, item := range b.Items {
+			expr := ConvertBindingToExpr(item.Binding, wrapIdentifier)
+			if b.HasSpread && i+1 == len(b.Items) {
+				expr = Expr{Loc: expr.Loc, Data: &ESpread{Value: expr}}
+			} else if item.DefaultValue != nil {
+				expr = Assign(expr, *item.DefaultValue)
+			}
+			exprs[i] = expr
+		}
+		return Expr{Loc: loc, Data: &EArray{
+			Items:        exprs,
+			IsSingleLine: b.IsSingleLine,
+		}}
+
+	case *BObject:
+		properties := make([]Property, len(b.Properties))
+		for i, property := range b.Properties {
+			value := ConvertBindingToExpr(property.Value, wrapIdentifier)
+			kind := PropertyNormal
+			if property.IsSpread {
+				kind = PropertySpread
+			}
+			properties[i] = Property{
+				Kind:        kind,
+				IsComputed:  property.IsComputed,
+				Key:         property.Key,
+				Value:       &value,
+				Initializer: property.DefaultValue,
+			}
+		}
+		return Expr{Loc: loc, Data: &EObject{
+			Properties:   properties,
+			IsSingleLine: b.IsSingleLine,
+		}}
+
+	default:
+		panic("Internal error")
+	}
+}
