@@ -38,7 +38,7 @@ func expectPrintedCommon(t *testing.T, name string, contents string, expected st
 			t.Fatal("Parse error")
 		}
 		symbols := js_ast.NewSymbolMap(1)
-		symbols.Outer[0] = tree.Symbols
+		symbols.SymbolsForSource[0] = tree.Symbols
 		r := renamer.NewNoOpRenamer(symbols)
 		js := Print(tree, symbols, r, options).JS
 		assertEqual(t, string(js), expected)
@@ -374,13 +374,13 @@ func TestString(t *testing.T) {
 }
 
 func TestTemplate(t *testing.T) {
-	expectPrinted(t, "let x = `\\1`", "let x = `\x01`;\n")
+	expectPrinted(t, "let x = `\\0`", "let x = `\\0`;\n")
 	expectPrinted(t, "let x = `\\x01`", "let x = `\x01`;\n")
-	expectPrinted(t, "let x = `\\1${0}`", "let x = `\x01${0}`;\n")
+	expectPrinted(t, "let x = `\\0${0}`", "let x = `\\0${0}`;\n")
 	expectPrinted(t, "let x = `\\x01${0}`", "let x = `\x01${0}`;\n")
-	expectPrinted(t, "let x = `${0}\\1`", "let x = `${0}\x01`;\n")
+	expectPrinted(t, "let x = `${0}\\0`", "let x = `${0}\\0`;\n")
 	expectPrinted(t, "let x = `${0}\\x01`", "let x = `${0}\x01`;\n")
-	expectPrinted(t, "let x = `${0}\\1${1}`", "let x = `${0}\x01${1}`;\n")
+	expectPrinted(t, "let x = `${0}\\0${1}`", "let x = `${0}\\0${1}`;\n")
 	expectPrinted(t, "let x = `${0}\\x01${1}`", "let x = `${0}\x01${1}`;\n")
 
 	expectPrinted(t, "let x = String.raw`\\1`", "let x = String.raw`\\1`;\n")
@@ -596,6 +596,11 @@ func TestClass(t *testing.T) {
 	expectPrinted(t, "class Foo { static set foo(x) {} }", "class Foo {\n  static set foo(x) {\n  }\n}\n")
 }
 
+func TestPrivateIdentifiers(t *testing.T) {
+	expectPrinted(t, "class Foo { #foo; foo() { return #foo in this } }", "class Foo {\n  #foo;\n  foo() {\n    return #foo in this;\n  }\n}\n")
+	expectPrintedMinify(t, "class Foo { #foo; foo() { return #foo in this } }", "class Foo{#foo;foo(){return#foo in this}}")
+}
+
 func TestImport(t *testing.T) {
 	expectPrinted(t, "import('path');", "import(\"path\");\n") // The semicolon must not be a separate statement
 
@@ -721,9 +726,14 @@ func TestMinify(t *testing.T) {
 	expectPrintedMinify(t, "import a from 'path'", "import a from\"path\";")
 	expectPrintedMinify(t, "import * as ns from 'path'", "import*as ns from\"path\";")
 	expectPrintedMinify(t, "import {a, b as c} from 'path'", "import{a,b as c}from\"path\";")
+	expectPrintedMinify(t, "import {a, ' ' as c} from 'path'", "import{a,\" \"as c}from\"path\";")
 
 	expectPrintedMinify(t, "export * as ns from 'path'", "export*as ns from\"path\";")
+	expectPrintedMinify(t, "export * as ' ' from 'path'", "export*as\" \"from\"path\";")
 	expectPrintedMinify(t, "export {a, b as c} from 'path'", "export{a,b as c}from\"path\";")
+	expectPrintedMinify(t, "export {' ', '-' as ';'} from 'path'", "export{\" \",\"-\"as\";\"}from\"path\";")
+	expectPrintedMinify(t, "let a, b; export {a, b as c}", "let a,b;export{a,b as c};")
+	expectPrintedMinify(t, "let a, b; export {a, b as ' '}", "let a,b;export{a,b as\" \"};")
 
 	// Print some strings using template literals when minifying
 	expectPrinted(t, "x = '\\n'", "x = \"\\n\";\n")
