@@ -13,20 +13,31 @@ import (
 	"strings"
 	"sync"
 	"time"
+	"unicode/utf8"
 )
 
 const defaultTerminalWidth = 80
 
 type Log struct {
+	Level LogLevel
+
 	AddMsg    func(Msg)
 	HasErrors func() bool
-	Done      func() []Msg
+
+	// This is called after the build has finished but before writing to stdout.
+	// It exists to ensure that deferred warning messages end up in the terminal
+	// before the data written to stdout.
+	AlmostDone func()
+
+	Done func() []Msg
 }
 
 type LogLevel int8
 
 const (
 	LevelNone LogLevel = iota
+	LevelVerbose
+	LevelDebug
 	LevelInfo
 	LevelWarning
 	LevelError
@@ -38,7 +49,10 @@ type MsgKind uint8
 const (
 	Error MsgKind = iota
 	Warning
+	Info
 	Note
+	Debug
+	Verbose
 )
 
 func (kind MsgKind) String() string {
@@ -47,17 +61,24 @@ func (kind MsgKind) String() string {
 		return "error"
 	case Warning:
 		return "warning"
+	case Info:
+		return "info"
 	case Note:
 		return "note"
+	case Debug:
+		return "debug"
+	case Verbose:
+		return "verbose"
 	default:
 		panic("Internal error")
 	}
 }
 
 type Msg struct {
-	Kind  MsgKind
-	Data  MsgData
-	Notes []MsgData
+	PluginName string
+	Kind       MsgKind
+	Data       MsgData
+	Notes      []MsgData
 }
 
 type MsgData struct {
@@ -69,12 +90,13 @@ type MsgData struct {
 }
 
 type MsgLocation struct {
-	File      string
-	Namespace string
-	Line      int // 1-based
-	Column    int // 0-based, in bytes
-	Length    int // in bytes
-	LineText  string
+	File       string
+	Namespace  string
+	Line       int // 1-based
+	Column     int // 0-based, in bytes
+	Length     int // in bytes
+	LineText   string
+	Suggestion string
 }
 
 type Loc struct {
@@ -127,10 +149,32 @@ func (a SortableMsgs) Less(i int, j int) bool {
 type Path struct {
 	Text      string
 	Namespace string
+
+	// This feature was added to support ancient CSS libraries that append things
+	// like "?#iefix" and "#icons" to some of their import paths as a hack for IE6.
+	// The intent is for these suffix parts to be ignored but passed through to
+	// the output. This is supported by other bundlers, so we also support this.
+	IgnoredSuffix string
+
+	Flags PathFlags
+}
+
+type PathFlags uint8
+
+const (
+	// This corresponds to a value of "false' in the "browser" package.json field
+	PathDisabled PathFlags = 1 << iota
+)
+
+func (p Path) IsDisabled() bool {
+	return (p.Flags & PathDisabled) != 0
 }
 
 func (a Path) ComesBeforeInSortedOrder(b Path) bool {
-	return a.Namespace > b.Namespace || (a.Namespace == b.Namespace && a.Text < b.Text)
+	return a.Namespace > b.Namespace ||
+		(a.Namespace == b.Namespace && (a.Text < b.Text ||
+			(a.Text == b.Text && (a.Flags < b.Flags ||
+				(a.Flags == b.Flags && a.IgnoredSuffix < b.IgnoredSuffix)))))
 }
 
 // This has a custom implementation instead of using "filepath.Dir/Base/Ext"
@@ -260,30 +304,68 @@ func (s *Source) RangeOfNumber(loc Loc) (r Range) {
 	return
 }
 
-func plural(prefix string, count int) string {
-	if count == 1 {
-		return fmt.Sprintf("%d %s", count, prefix)
+func (s *Source) RangeOfLegacyOctalEscape(loc Loc) (r Range) {
+	text := s.Contents[loc.Start:]
+	r = Range{Loc: loc, Len: 0}
+
+	if len(text) >= 2 && text[0] == '\\' {
+		r.Len = 2
+		for r.Len < 4 && int(r.Len) < len(text) {
+			c := text[r.Len]
+			if c < '0' || c > '9' {
+				break
+			}
+			r.Len++
+		}
 	}
-	return fmt.Sprintf("%d %ss", count, prefix)
+	return
 }
 
-func errorAndWarningSummary(errors int, warnings int) string {
+func plural(prefix string, count int, shown int, someAreMissing bool) string {
+	var text string
+	if count == 1 {
+		text = fmt.Sprintf("%d %s", count, prefix)
+	} else {
+		text = fmt.Sprintf("%d %ss", count, prefix)
+	}
+	if shown < count {
+		text = fmt.Sprintf("%d of %s", shown, text)
+	} else if someAreMissing && count > 1 {
+		text = "all " + text
+	}
+	return text
+}
+
+func errorAndWarningSummary(errors int, warnings int, shownErrors int, shownWarnings int) string {
+	someAreMissing := shownWarnings < warnings || shownErrors < errors
 	switch {
 	case errors == 0:
-		return plural("warning", warnings)
+		return plural("warning", warnings, shownWarnings, someAreMissing)
 	case warnings == 0:
-		return plural("error", errors)
+		return plural("error", errors, shownErrors, someAreMissing)
 	default:
 		return fmt.Sprintf("%s and %s",
-			plural("warning", warnings),
-			plural("error", errors))
+			plural("warning", warnings, shownWarnings, someAreMissing),
+			plural("error", errors, shownErrors, someAreMissing))
 	}
 }
 
+type APIKind uint8
+
+const (
+	GoAPI APIKind = iota
+	CLIAPI
+	JSAPI
+)
+
+// This can be used to customize error messages for the current API kind
+var API APIKind
+
 type TerminalInfo struct {
 	IsTTY           bool
 	UseColorEscapes bool
 	Width           int
+	Height          int
 }
 
 func NewStderrLog(options OutputOptions) Log {
@@ -292,7 +374,39 @@ func NewStderrLog(options OutputOptions) Log {
 	terminalInfo := GetTerminalInfo(os.Stderr)
 	errors := 0
 	warnings := 0
-	errorLimitWasHit := false
+	shownErrors := 0
+	shownWarnings := 0
+	hasErrors := false
+	remainingMessagesBeforeLimit := options.MessageLimit
+	if remainingMessagesBeforeLimit == 0 {
+		remainingMessagesBeforeLimit = 0x7FFFFFFF
+	}
+	var deferredWarnings []Msg
+	didFinalizeLog := false
+
+	finalizeLog := func() {
+		if didFinalizeLog {
+			return
+		}
+		didFinalizeLog = true
+
+		// Print the deferred warning now if there was no error after all
+		for remainingMessagesBeforeLimit > 0 && len(deferredWarnings) > 0 {
+			shownWarnings++
+			writeStringWithColor(os.Stderr, deferredWarnings[0].String(options, terminalInfo))
+			deferredWarnings = deferredWarnings[1:]
+			remainingMessagesBeforeLimit--
+		}
+
+		// Print out a summary
+		if options.MessageLimit > 0 && errors+warnings > options.MessageLimit {
+			writeStringWithColor(os.Stderr, fmt.Sprintf("%s shown (disable the message limit with --log-limit=0)\n",
+				errorAndWarningSummary(errors, warnings, shownErrors, shownWarnings)))
+		} else if options.LogLevel <= LevelInfo && (warnings != 0 || errors != 0) {
+			writeStringWithColor(os.Stderr, fmt.Sprintf("%s\n",
+				errorAndWarningSummary(errors, warnings, shownErrors, shownWarnings)))
+		}
+	}
 
 	switch options.Color {
 	case ColorNever:
@@ -302,56 +416,89 @@ func NewStderrLog(options OutputOptions) Log {
 	}
 
 	return Log{
+		Level: options.LogLevel,
+
 		AddMsg: func(msg Msg) {
 			mutex.Lock()
 			defer mutex.Unlock()
 			msgs = append(msgs, msg)
 
+			switch msg.Kind {
+			case Verbose:
+				if options.LogLevel <= LevelVerbose {
+					writeStringWithColor(os.Stderr, msg.String(options, terminalInfo))
+				}
+
+			case Debug:
+				if options.LogLevel <= LevelDebug {
+					writeStringWithColor(os.Stderr, msg.String(options, terminalInfo))
+				}
+
+			case Info:
+				if options.LogLevel <= LevelInfo {
+					writeStringWithColor(os.Stderr, msg.String(options, terminalInfo))
+				}
+
+			case Error:
+				hasErrors = true
+				if options.LogLevel <= LevelError {
+					errors++
+				}
+
+			case Warning:
+				if options.LogLevel <= LevelWarning {
+					warnings++
+				}
+			}
+
 			// Be silent if we're past the limit so we don't flood the terminal
-			if errorLimitWasHit {
+			if remainingMessagesBeforeLimit == 0 {
 				return
 			}
 
 			switch msg.Kind {
 			case Error:
-				errors++
 				if options.LogLevel <= LevelError {
-					options.MaybeWriteFirstNewline(os.Stderr)
+					shownErrors++
 					writeStringWithColor(os.Stderr, msg.String(options, terminalInfo))
+					remainingMessagesBeforeLimit--
 				}
-			case Warning:
-				warnings++
-				if options.LogLevel <= LevelWarning {
-					options.MaybeWriteFirstNewline(os.Stderr)
-					writeStringWithColor(os.Stderr, msg.String(options, terminalInfo))
-				}
-			}
 
-			// Silence further output if we reached the error limit
-			if options.ErrorLimit != 0 && errors >= options.ErrorLimit {
-				errorLimitWasHit = true
-				if options.LogLevel <= LevelError {
-					options.MaybeWriteFirstNewline(os.Stderr)
-					writeStringWithColor(os.Stderr, fmt.Sprintf(
-						"%s reached (disable error limit with --error-limit=0)\n", errorAndWarningSummary(errors, warnings)))
+			case Warning:
+				if options.LogLevel <= LevelWarning {
+					if remainingMessagesBeforeLimit > (options.MessageLimit+1)/2 {
+						shownWarnings++
+						writeStringWithColor(os.Stderr, msg.String(options, terminalInfo))
+						remainingMessagesBeforeLimit--
+					} else {
+						// If we have less than half of the slots left, wait for potential
+						// future errors instead of using up all of the slots with warnings.
+						// We want the log for a failed build to always have at least one
+						// error in it.
+						deferredWarnings = append(deferredWarnings, msg)
+					}
 				}
 			}
 		},
+
 		HasErrors: func() bool {
 			mutex.Lock()
 			defer mutex.Unlock()
-			return errors > 0
+			return hasErrors
 		},
+
+		AlmostDone: func() {
+			mutex.Lock()
+			defer mutex.Unlock()
+
+			finalizeLog()
+		},
+
 		Done: func() []Msg {
 			mutex.Lock()
 			defer mutex.Unlock()
 
-			// Print out a summary if the error limit wasn't hit
-			if !errorLimitWasHit && options.LogLevel <= LevelInfo && (warnings != 0 || errors != 0) {
-				options.MaybeWriteFirstNewline(os.Stderr)
-				writeStringWithColor(os.Stderr, fmt.Sprintf("%s\n", errorAndWarningSummary(errors, warnings)))
-			}
-
+			finalizeLog()
 			sort.Stable(msgs)
 			return msgs
 		},
@@ -394,9 +541,10 @@ func PrintMessageToStderr(osArgs []string, msg Msg) {
 }
 
 type Colors struct {
-	Default string
-	Bold    string
-	Dim     string
+	Reset     string
+	Bold      string
+	Dim       string
+	Underline string
 
 	Red   string
 	Green string
@@ -405,8 +553,21 @@ type Colors struct {
 	Cyan    string
 	Magenta string
 	Yellow  string
+}
 
-	Underline string
+var TerminalColors = Colors{
+	Reset:     "\033[0m",
+	Bold:      "\033[1m",
+	Dim:       "\033[37m",
+	Underline: "\033[4m",
+
+	Red:   "\033[31m",
+	Green: "\033[32m",
+	Blue:  "\033[34m",
+
+	Cyan:    "\033[36m",
+	Magenta: "\033[35m",
+	Yellow:  "\033[33m",
 }
 
 func PrintText(file *os.File, level LogLevel, osArgs []string, callback func(Colors) string) {
@@ -417,8 +578,12 @@ func PrintText(file *os.File, level LogLevel, osArgs []string, callback func(Col
 		return
 	}
 
+	PrintTextWithColor(file, options.Color, callback)
+}
+
+func PrintTextWithColor(file *os.File, useColor UseColor, callback func(Colors) string) {
 	var useColorEscapes bool
-	switch options.Color {
+	switch useColor {
 	case ColorNever:
 		useColorEscapes = false
 	case ColorAlways:
@@ -429,19 +594,7 @@ func PrintText(file *os.File, level LogLevel, osArgs []string, callback func(Col
 
 	var colors Colors
 	if useColorEscapes {
-		colors.Default = colorReset
-		colors.Bold = colorResetBold
-		colors.Dim = colorResetDim
-
-		colors.Red = colorRed
-		colors.Green = colorGreen
-		colors.Blue = colorBlue
-
-		colors.Cyan = colorCyan
-		colors.Magenta = colorMagenta
-		colors.Yellow = colorYellow
-
-		colors.Underline = colorResetUnderline
+		colors = TerminalColors
 	}
 	writeStringWithColor(file, callback(colors))
 }
@@ -480,14 +633,6 @@ func (t SummaryTable) Less(i int, j int) bool {
 		return false
 	}
 
-	// Sort subdirectories first
-	if strings.HasPrefix(ti.Dir, tj.Dir) {
-		return true
-	}
-	if strings.HasPrefix(tj.Dir, ti.Dir) {
-		return false
-	}
-
 	// Sort alphabetically by directory first
 	if ti.Dir < tj.Dir {
 		return true
@@ -503,8 +648,8 @@ func (t SummaryTable) Less(i int, j int) bool {
 // Show a warning icon next to output files that are 1mb or larger
 const sizeWarningThreshold = 1024 * 1024
 
-func PrintSummary(osArgs []string, table SummaryTable, start time.Time) {
-	PrintText(os.Stderr, LevelInfo, osArgs, func(colors Colors) string {
+func PrintSummary(useColor UseColor, table SummaryTable, start *time.Time) {
+	PrintTextWithColor(os.Stderr, useColor, func(colors Colors) string {
 		isProbablyWindowsCommandPrompt := false
 		sb := strings.Builder{}
 
@@ -523,6 +668,21 @@ func PrintSummary(osArgs []string, table SummaryTable, start time.Time) {
 		}
 
 		if len(table) > 0 {
+			info := GetTerminalInfo(os.Stderr)
+
+			// Truncate the table in case it's really long
+			maxLength := info.Height / 2
+			if info.Height == 0 {
+				maxLength = 20
+			} else if maxLength < 5 {
+				maxLength = 5
+			}
+			length := len(table)
+			sort.Sort(table)
+			if length > maxLength {
+				table = table[:maxLength]
+			}
+
 			// Compute the maximum width of the size column
 			spacingBetweenColumns := 2
 			hasSizeWarning := false
@@ -543,7 +703,7 @@ func PrintSummary(osArgs []string, table SummaryTable, start time.Time) {
 			}
 
 			margin := "  "
-			layoutWidth := GetTerminalInfo(os.Stderr).Width
+			layoutWidth := info.Width
 			if layoutWidth < 1 {
 				layoutWidth = defaultTerminalWidth
 			}
@@ -555,11 +715,9 @@ func PrintSummary(osArgs []string, table SummaryTable, start time.Time) {
 			if layoutWidth > maxPath+maxSize {
 				layoutWidth = maxPath + maxSize
 			}
-			sort.Sort(table)
-			sb.WriteString("\n")
+			sb.WriteByte('\n')
 
-			wasSourceMap := false
-			for i, entry := range table {
+			for _, entry := range table {
 				dir, base := entry.Dir, entry.Base
 				pathWidth := layoutWidth - maxSize
 
@@ -589,13 +747,6 @@ func PrintSummary(osArgs []string, table SummaryTable, start time.Time) {
 					spacer = 0
 				}
 
-				// Print a boundary in between normal files and source map files if
-				// there was more than one normal file. This improves scannability.
-				if !wasSourceMap && entry.IsSourceMap && i > 1 {
-					sb.WriteString("\n")
-					wasSourceMap = true
-				}
-
 				// Put a warning next to the size if it's above a certain threshold
 				sizeColor := colors.Cyan
 				sizeWarning := ""
@@ -608,21 +759,32 @@ func PrintSummary(osArgs []string, table SummaryTable, start time.Time) {
 					}
 				}
 
-				sb.WriteString(fmt.Sprintf("%s%s%s%s%s%s%s%s%s%s%s\n",
+				sb.WriteString(fmt.Sprintf("%s%s%s%s%s%s%s%s%s%s%s%s\n",
 					margin,
 					colors.Dim,
 					dir,
+					colors.Reset,
 					colors.Bold,
 					base,
-					colors.Default,
+					colors.Reset,
 					strings.Repeat(" ", spacer),
 					sizeColor,
 					entry.Size,
 					sizeWarning,
-					colors.Default,
+					colors.Reset,
 				))
 			}
+
+			// Say how many remaining files are not shown
+			if length > maxLength {
+				plural := "s"
+				if length == maxLength+1 {
+					plural = ""
+				}
+				sb.WriteString(fmt.Sprintf("%s%s...and %d more output file%s...%s\n", margin, colors.Dim, length-maxLength, plural, colors.Reset))
+			}
 		}
+		sb.WriteByte('\n')
 
 		lightningSymbol := "⚡ "
 
@@ -631,12 +793,16 @@ func PrintSummary(osArgs []string, table SummaryTable, start time.Time) {
 			lightningSymbol = ""
 		}
 
-		sb.WriteString(fmt.Sprintf("\n%s%sDone in %dms%s\n\n",
-			lightningSymbol,
-			colors.Green,
-			time.Since(start).Milliseconds(),
-			colors.Default,
-		))
+		// Printing the time taken is optional
+		if start != nil {
+			sb.WriteString(fmt.Sprintf("%s%sDone in %dms%s\n",
+				lightningSymbol,
+				colors.Green,
+				time.Since(*start).Milliseconds(),
+				colors.Reset,
+			))
+		}
+
 		return sb.String()
 	})
 }
@@ -647,7 +813,13 @@ func NewDeferLog() Log {
 	var hasErrors bool
 
 	return Log{
+		Level: LevelInfo,
+
 		AddMsg: func(msg Msg) {
+			if msg.Kind == Verbose || msg.Kind == Debug {
+				// Ignore these when not writing to stderr
+				return
+			}
 			mutex.Lock()
 			defer mutex.Unlock()
 			if msg.Kind == Error {
@@ -655,11 +827,16 @@ func NewDeferLog() Log {
 			}
 			msgs = append(msgs, msg)
 		},
+
 		HasErrors: func() bool {
 			mutex.Lock()
 			defer mutex.Unlock()
 			return hasErrors
 		},
+
+		AlmostDone: func() {
+		},
+
 		Done: func() []Msg {
 			mutex.Lock()
 			defer mutex.Unlock()
@@ -669,21 +846,6 @@ func NewDeferLog() Log {
 	}
 }
 
-const colorReset = "\033[0m"
-
-const colorRed = "\033[31m"
-const colorGreen = "\033[32m"
-const colorBlue = "\033[34m"
-
-const colorCyan = "\033[36m"
-const colorMagenta = "\033[35m"
-const colorYellow = "\033[33m"
-
-const colorResetDim = "\033[0;37m"
-const colorBold = "\033[1m"
-const colorResetBold = "\033[0;1m"
-const colorResetUnderline = "\033[0;4m"
-
 type UseColor uint8
 
 const (
@@ -694,19 +856,9 @@ const (
 
 type OutputOptions struct {
 	IncludeSource bool
-	ErrorLimit    int
+	MessageLimit  int
 	Color         UseColor
 	LogLevel      LogLevel
-
-	// This is used with watch mode
-	NewlineBeforeFirstMessage bool
-}
-
-func (opts *OutputOptions) MaybeWriteFirstNewline(file *os.File) {
-	if opts.NewlineBeforeFirstMessage {
-		opts.NewlineBeforeFirstMessage = false
-		file.WriteString("\n")
-	}
 }
 
 func (msg Msg) String(options OutputOptions, terminalInfo TerminalInfo) string {
@@ -726,10 +878,19 @@ func (msg Msg) String(options OutputOptions, terminalInfo TerminalInfo) string {
 		}
 	}
 
-	// Format the messages
-	text := msgString(options, terminalInfo, msg.Kind, msg.Data, maxMargin)
+	// Format the message
+	text := msgString(options.IncludeSource, terminalInfo, msg.Kind, msg.Data, maxMargin, msg.PluginName)
+
+	// Put a blank line between the message and the notes if the message has a stack trace
+	gap := ""
+	if loc := msg.Data.Location; loc != nil && strings.ContainsRune(loc.LineText, '\n') {
+		gap = "\n"
+	}
+
+	// Format the notes
 	for _, note := range msg.Notes {
-		text += msgString(options, terminalInfo, Note, note, maxMargin)
+		text += gap
+		text += msgString(options.IncludeSource, terminalInfo, Note, note, maxMargin, "")
 	}
 
 	// Add extra spacing between messages if source code is present
@@ -747,32 +908,50 @@ func marginWithLineText(maxMargin int, line int) string {
 	return fmt.Sprintf("    %s%s │ ", strings.Repeat(" ", maxMargin-len(number)), number)
 }
 
-func emptyMarginText(maxMargin int) string {
-	return fmt.Sprintf("    %s ╵ ", strings.Repeat(" ", maxMargin))
+func emptyMarginText(maxMargin int, isLast bool) string {
+	space := strings.Repeat(" ", maxMargin)
+	if isLast {
+		return fmt.Sprintf("    %s ╵ ", space)
+	}
+	return fmt.Sprintf("    %s │ ", space)
 }
 
-func msgString(options OutputOptions, terminalInfo TerminalInfo, kind MsgKind, data MsgData, maxMargin int) string {
+func msgString(includeSource bool, terminalInfo TerminalInfo, kind MsgKind, data MsgData, maxMargin int, pluginName string) string {
+	var colors Colors
+	if terminalInfo.UseColorEscapes {
+		colors = TerminalColors
+	}
+
 	var kindColor string
-	textColor := colorBold
-	textResetColor := colorResetBold
+	prefixColor := colors.Bold
+	messageColor := colors.Bold
 	textIndent := ""
 
-	if options.IncludeSource {
+	if includeSource {
 		textIndent = " > "
 	}
 
 	switch kind {
+	case Verbose:
+		kindColor = colors.Cyan
+
+	case Debug:
+		kindColor = colors.Blue
+
+	case Info:
+		kindColor = colors.Green
+
 	case Error:
-		kindColor = colorRed
+		kindColor = colors.Red
 
 	case Warning:
-		kindColor = colorMagenta
+		kindColor = colors.Magenta
 
 	case Note:
-		textColor = colorReset
-		kindColor = colorResetBold
-		textResetColor = colorReset
-		if options.IncludeSource {
+		prefixColor = colors.Reset
+		kindColor = colors.Bold
+		messageColor = ""
+		if includeSource {
 			textIndent = "   "
 		}
 
@@ -780,46 +959,44 @@ func msgString(options OutputOptions, terminalInfo TerminalInfo, kind MsgKind, d
 		panic("Internal error")
 	}
 
-	if data.Location == nil {
-		if terminalInfo.UseColorEscapes {
-			return fmt.Sprintf("%s%s%s%s: %s%s%s\n",
-				textColor, textIndent, kindColor, kind.String(),
-				textResetColor, data.Text,
-				colorReset)
-		}
-
-		return fmt.Sprintf("%s%s: %s\n", textIndent, kind.String(), data.Text)
+	var pluginText string
+	if pluginName != "" {
+		pluginText = fmt.Sprintf("%s[plugin: %s] ", colors.Yellow, pluginName)
 	}
 
-	if !options.IncludeSource {
-		if terminalInfo.UseColorEscapes {
-			return fmt.Sprintf("%s%s%s: %s%s: %s%s%s\n",
-				textColor, textIndent, data.Location.File,
-				kindColor, kind.String(),
-				textResetColor, data.Text,
-				colorReset)
-		}
+	if data.Location == nil {
+		return fmt.Sprintf("%s%s%s%s: %s%s%s%s\n%s",
+			prefixColor, textIndent, kindColor, kind.String(),
+			pluginText, colors.Reset, messageColor, data.Text,
+			colors.Reset)
+	}
 
-		return fmt.Sprintf("%s%s: %s: %s\n",
-			textIndent, data.Location.File, kind.String(), data.Text)
+	if !includeSource {
+		return fmt.Sprintf("%s%s%s: %s%s: %s%s%s%s\n%s",
+			prefixColor, textIndent, data.Location.File,
+			kindColor, kind.String(),
+			pluginText, colors.Reset, messageColor, data.Text,
+			colors.Reset)
 	}
 
 	d := detailStruct(data, terminalInfo, maxMargin)
 
-	if terminalInfo.UseColorEscapes {
-		return fmt.Sprintf("%s%s%s: %s%s: %s%s\n%s%s%s%s%s%s\n%s%s%s%s%s%s\n",
-			textColor, textIndent, d.Path,
-			kindColor, kind.String(),
-			textResetColor, d.Message,
-			colorResetDim, d.SourceBefore, colorGreen, d.SourceMarked, colorResetDim, d.SourceAfter,
-			d.Indent, colorGreen, d.Marker,
-			colorResetDim, d.ContentAfter, colorReset)
+	callout := d.Marker
+	calloutPrefix := ""
+
+	if d.Suggestion != "" {
+		callout = d.Suggestion
+		calloutPrefix = fmt.Sprintf("%s%s%s%s%s\n",
+			emptyMarginText(maxMargin, false), d.Indent, colors.Green, d.Marker, colors.Dim)
 	}
 
-	return fmt.Sprintf("%s%s: %s: %s\n%s%s%s\n%s%s%s\n",
-		textIndent, d.Path, kind.String(), d.Message,
-		d.SourceBefore, d.SourceMarked, d.SourceAfter,
-		d.Indent, d.Marker, d.ContentAfter)
+	return fmt.Sprintf("%s%s%s:%d:%d: %s%s: %s%s%s%s\n%s%s%s%s%s%s%s\n%s%s%s%s%s%s%s\n%s",
+		prefixColor, textIndent, d.Path, d.Line, d.Column,
+		kindColor, kind.String(),
+		pluginText, colors.Reset, messageColor, d.Message,
+		colors.Reset, colors.Dim, d.SourceBefore, colors.Green, d.SourceMarked, colors.Dim, d.SourceAfter,
+		calloutPrefix, emptyMarginText(maxMargin, true), d.Indent, colors.Green, callout, colors.Dim, d.ContentAfter,
+		colors.Reset)
 }
 
 type MsgDetail struct {
@@ -832,8 +1009,9 @@ type MsgDetail struct {
 	SourceMarked string
 	SourceAfter  string
 
-	Indent string
-	Marker string
+	Indent     string
+	Marker     string
+	Suggestion string
 
 	ContentAfter string
 }
@@ -926,10 +1104,11 @@ func detailStruct(data MsgData, terminalInfo TerminalInfo, maxMargin int) MsgDet
 
 	spacesPerTab := 2
 	lineText := renderTabStops(firstLine, spacesPerTab)
-	indent := strings.Repeat(" ", len(renderTabStops(firstLine[:loc.Column], spacesPerTab)))
+	textUpToLoc := renderTabStops(firstLine[:loc.Column], spacesPerTab)
+	markerStart := len(textUpToLoc)
+	markerEnd := markerStart
+	indent := strings.Repeat(" ", estimateWidthInTerminal(textUpToLoc))
 	marker := "^"
-	markerStart := len(indent)
-	markerEnd := len(indent)
 
 	// Extend markers to cover the full range of the error
 	if loc.Length > 0 {
@@ -1005,13 +1184,13 @@ func detailStruct(data MsgData, terminalInfo TerminalInfo, maxMargin int) MsgDet
 		}
 
 		// Now we can compute the indent
-		indent = strings.Repeat(" ", markerStart)
 		lineText = slicedLine
+		indent = strings.Repeat(" ", estimateWidthInTerminal(lineText[:markerStart]))
 	}
 
 	// If marker is still multi-character after clipping, make the marker wider
 	if markerEnd-markerStart > 1 {
-		marker = strings.Repeat("~", markerEnd-markerStart)
+		marker = strings.Repeat("~", estimateWidthInTerminal(lineText[markerStart:markerEnd]))
 	}
 
 	// Put a margin before the marker indent
@@ -1027,13 +1206,31 @@ func detailStruct(data MsgData, terminalInfo TerminalInfo, maxMargin int) MsgDet
 		SourceMarked: lineText[markerStart:markerEnd],
 		SourceAfter:  lineText[markerEnd:],
 
-		Indent: emptyMarginText(maxMargin) + indent,
-		Marker: marker,
+		Indent:     indent,
+		Marker:     marker,
+		Suggestion: loc.Suggestion,
 
 		ContentAfter: afterFirstLine,
 	}
 }
 
+// Estimate the number of columns this string will take when printed
+func estimateWidthInTerminal(text string) int {
+	// For now just assume each code point is one column. This is wrong but is
+	// less wrong than assuming each code unit is one column.
+	width := 0
+	for text != "" {
+		c, size := utf8.DecodeRuneInString(text)
+		text = text[size:]
+
+		// Ignore the Zero Width No-Break Space character (UTF-8 BOM)
+		if c != 0xFEFF {
+			width++
+		}
+	}
+	return width
+}
+
 func renderTabStops(withTabs string, spacesPerTab int) string {
 	if !strings.ContainsRune(withTabs, '\t') {
 		return withTabs
@@ -1080,6 +1277,36 @@ func (log Log) AddWarning(source *Source, loc Loc, text string) {
 	})
 }
 
+func (log Log) AddDebug(source *Source, loc Loc, text string) {
+	log.AddMsg(Msg{
+		Kind: Debug,
+		Data: RangeData(source, Range{Loc: loc}, text),
+	})
+}
+
+func (log Log) AddDebugWithNotes(source *Source, loc Loc, text string, notes []MsgData) {
+	log.AddMsg(Msg{
+		Kind:  Debug,
+		Data:  RangeData(source, Range{Loc: loc}, text),
+		Notes: notes,
+	})
+}
+
+func (log Log) AddVerbose(source *Source, loc Loc, text string) {
+	log.AddMsg(Msg{
+		Kind: Verbose,
+		Data: RangeData(source, Range{Loc: loc}, text),
+	})
+}
+
+func (log Log) AddVerboseWithNotes(source *Source, loc Loc, text string, notes []MsgData) {
+	log.AddMsg(Msg{
+		Kind:  Verbose,
+		Data:  RangeData(source, Range{Loc: loc}, text),
+		Notes: notes,
+	})
+}
+
 func (log Log) AddRangeError(source *Source, r Range, text string) {
 	log.AddMsg(Msg{
 		Kind: Error,
@@ -1094,6 +1321,13 @@ func (log Log) AddRangeWarning(source *Source, r Range, text string) {
 	})
 }
 
+func (log Log) AddRangeDebug(source *Source, r Range, text string) {
+	log.AddMsg(Msg{
+		Kind: Debug,
+		Data: RangeData(source, r, text),
+	})
+}
+
 func (log Log) AddRangeErrorWithNotes(source *Source, r Range, text string, notes []MsgData) {
 	log.AddMsg(Msg{
 		Kind:  Error,
