@@ -5,6 +5,7 @@ import (
 	"strconv"
 	"sync"
 
+	"github.com/evanw/esbuild/internal/ast"
 	"github.com/evanw/esbuild/internal/js_ast"
 	"github.com/evanw/esbuild/internal/js_lexer"
 )
@@ -74,7 +75,7 @@ type MinifyRenamer struct {
 	sortedBuffer  StableRefArray
 	reservedNames map[string]uint32
 	slots         [3][]symbolSlot
-	symbolToSlot  map[js_ast.Ref]uint32
+	symbolToSlot  map[js_ast.Ref]ast.Index32
 }
 
 func NewMinifyRenamer(symbols js_ast.SymbolMap, firstTopLevelSlots js_ast.SlotCounts, reservedNames map[string]uint32) *MinifyRenamer {
@@ -86,7 +87,7 @@ func NewMinifyRenamer(symbols js_ast.SymbolMap, firstTopLevelSlots js_ast.SlotCo
 			make([]symbolSlot, firstTopLevelSlots[1]),
 			make([]symbolSlot, firstTopLevelSlots[2]),
 		},
-		symbolToSlot: make(map[js_ast.Ref]uint32),
+		symbolToSlot: make(map[js_ast.Ref]ast.Index32),
 	}
 }
 
@@ -102,10 +103,10 @@ func (r *MinifyRenamer) NameForSymbol(ref js_ast.Ref) string {
 	}
 
 	// Check if it's a nested scope symbol
-	i := ^symbol.NestedScopeSlot
+	i := symbol.NestedScopeSlot
 
 	// If it's not (i.e. it's in a top-level scope), look up the slot
-	if i == ^uint32(0) {
+	if !i.IsValid() {
 		var ok bool
 		i, ok = r.symbolToSlot[ref]
 		if !ok {
@@ -118,7 +119,7 @@ func (r *MinifyRenamer) NameForSymbol(ref js_ast.Ref) string {
 		}
 	}
 
-	return r.slots[ns][i].name
+	return r.slots[ns][i.GetIndex()].name
 }
 
 func (r *MinifyRenamer) AccumulateSymbolUseCounts(symbolUses map[js_ast.Ref]js_ast.SymbolUse, stableSourceIndices []uint32) {
@@ -126,8 +127,8 @@ func (r *MinifyRenamer) AccumulateSymbolUseCounts(symbolUses map[js_ast.Ref]js_a
 	r.sortedBuffer = r.sortedBuffer[:0]
 	for ref := range symbolUses {
 		r.sortedBuffer = append(r.sortedBuffer, StableRef{
-			StableOuterIndex: stableSourceIndices[ref.OuterIndex],
-			Ref:              ref,
+			StableSourceIndex: stableSourceIndices[ref.SourceIndex],
+			Ref:               ref,
 		})
 	}
 	sort.Sort(r.sortedBuffer)
@@ -155,20 +156,20 @@ func (r *MinifyRenamer) AccumulateSymbolCount(ref js_ast.Ref, count uint32) {
 
 	// Check if it's a nested scope symbol
 	slots := &r.slots[ns]
-	i := ^symbol.NestedScopeSlot
+	i := symbol.NestedScopeSlot
 
 	// If it's not (i.e. it's in a top-level scope), allocate a slot for it
-	if i == ^uint32(0) {
+	if !i.IsValid() {
 		var ok bool
 		i, ok = r.symbolToSlot[ref]
 		if !ok {
-			i = uint32(len(*slots))
+			i = ast.MakeIndex32(uint32(len(*slots)))
 			*slots = append(*slots, symbolSlot{})
 			r.symbolToSlot[ref] = i
 		}
 	}
 
-	(*slots)[i].count += count
+	(*slots)[i.GetIndex()].count += count
 }
 
 func (r *MinifyRenamer) AssignNamesByFrequency(minifier *js_ast.NameMinifier) {
@@ -217,16 +218,17 @@ func (r *MinifyRenamer) AssignNamesByFrequency(minifier *js_ast.NameMinifier) {
 
 // Returns the number of nested slots
 func AssignNestedScopeSlots(moduleScope *js_ast.Scope, symbols []js_ast.Symbol) (slotCounts js_ast.SlotCounts) {
-	// Temporarily set the nested scope slots of top-level symbols to non-zero so
+	// Temporarily set the nested scope slots of top-level symbols to valid so
 	// they aren't renamed in nested scopes. This prevents us from accidentally
 	// assigning nested scope slots to variables declared using "var" in a nested
 	// scope that are actually hoisted up to the module scope to become a top-
 	// level symbol.
+	validSlot := ast.MakeIndex32(1)
 	for _, member := range moduleScope.Members {
-		symbols[member.Ref.InnerIndex].NestedScopeSlot = 1
+		symbols[member.Ref.InnerIndex].NestedScopeSlot = validSlot
 	}
 	for _, ref := range moduleScope.Generated {
-		symbols[ref.InnerIndex].NestedScopeSlot = 1
+		symbols[ref.InnerIndex].NestedScopeSlot = validSlot
 	}
 
 	// Assign nested scope slots independently for each nested scope
@@ -237,10 +239,10 @@ func AssignNestedScopeSlots(moduleScope *js_ast.Scope, symbols []js_ast.Symbol)
 	// Then set the nested scope slots of top-level symbols back to zero. Top-
 	// level symbols are not supposed to have nested scope slots.
 	for _, member := range moduleScope.Members {
-		symbols[member.Ref.InnerIndex].NestedScopeSlot = 0
+		symbols[member.Ref.InnerIndex].NestedScopeSlot = ast.Index32{}
 	}
 	for _, ref := range moduleScope.Generated {
-		symbols[ref.InnerIndex].NestedScopeSlot = 0
+		symbols[ref.InnerIndex].NestedScopeSlot = ast.Index32{}
 	}
 	return
 }
@@ -253,20 +255,20 @@ func assignNestedScopeSlotsHelper(scope *js_ast.Scope, symbols []js_ast.Symbol,
 	}
 	sort.Ints(sortedMembers)
 
-	// Assign slots for this scope's symbols. Only do this if the slot is 0 (i.e.
-	// not already assigned). Nested scopes have copies of symbols from parent
+	// Assign slots for this scope's symbols. Only do this if the slot is
+	// not already assigned. Nested scopes have copies of symbols from parent
 	// scopes and we want to use the slot from the parent scope, not child scopes.
 	for _, innerIndex := range sortedMembers {
 		symbol := &symbols[innerIndex]
-		if ns := symbol.SlotNamespace(); ns != js_ast.SlotMustNotBeRenamed && symbol.NestedScopeSlot == 0 {
-			symbol.NestedScopeSlot = ^slot[ns]
+		if ns := symbol.SlotNamespace(); ns != js_ast.SlotMustNotBeRenamed && !symbol.NestedScopeSlot.IsValid() {
+			symbol.NestedScopeSlot = ast.MakeIndex32(slot[ns])
 			slot[ns]++
 		}
 	}
 	for _, ref := range scope.Generated {
 		symbol := &symbols[ref.InnerIndex]
-		if ns := symbol.SlotNamespace(); ns != js_ast.SlotMustNotBeRenamed && symbol.NestedScopeSlot == 0 {
-			symbol.NestedScopeSlot = ^slot[ns]
+		if ns := symbol.SlotNamespace(); ns != js_ast.SlotMustNotBeRenamed && !symbol.NestedScopeSlot.IsValid() {
+			symbol.NestedScopeSlot = ast.MakeIndex32(slot[ns])
 			slot[ns]++
 		}
 	}
@@ -274,7 +276,7 @@ func assignNestedScopeSlotsHelper(scope *js_ast.Scope, symbols []js_ast.Symbol,
 	// Labels are always declared in a nested scope, so we don't need to check.
 	if scope.LabelRef != js_ast.InvalidRef {
 		symbol := &symbols[scope.LabelRef.InnerIndex]
-		symbol.NestedScopeSlot = ^slot[js_ast.SlotLabel]
+		symbol.NestedScopeSlot = ast.MakeIndex32(slot[js_ast.SlotLabel])
 		slot[js_ast.SlotLabel]++
 	}
 
@@ -301,9 +303,16 @@ func (a slotAndCountArray) Less(i int, j int) bool {
 	return ai.count > aj.count || (ai.count == aj.count && ai.slot < aj.slot)
 }
 
+// The sort order here is arbitrary but needs to be consistent between builds.
+// The InnerIndex should be stable because the parser for a single file is
+// single-threaded and deterministically assigns out InnerIndex values
+// sequentially. But the SourceIndex should be unstable because the main thread
+// assigns out source index values sequentially to newly-discovered dependencies
+// in a multi-threaded producer/consumer relationship. So instead we use the
+// index of the source in the DFS order over all entry points for stability.
 type StableRef struct {
-	StableOuterIndex uint32
-	Ref              js_ast.Ref
+	StableSourceIndex uint32
+	Ref               js_ast.Ref
 }
 
 // This type is just so we can use Go's native sort function
@@ -313,7 +322,7 @@ func (a StableRefArray) Len() int          { return len(a) }
 func (a StableRefArray) Swap(i int, j int) { a[i], a[j] = a[j], a[i] }
 func (a StableRefArray) Less(i int, j int) bool {
 	ai, aj := a[i], a[j]
-	return ai.StableOuterIndex > aj.StableOuterIndex || (ai.StableOuterIndex == aj.StableOuterIndex && ai.Ref.InnerIndex < aj.Ref.InnerIndex)
+	return ai.StableSourceIndex < aj.StableSourceIndex || (ai.StableSourceIndex == aj.StableSourceIndex && ai.Ref.InnerIndex < aj.Ref.InnerIndex)
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -328,14 +337,14 @@ type NumberRenamer struct {
 func NewNumberRenamer(symbols js_ast.SymbolMap, reservedNames map[string]uint32) *NumberRenamer {
 	return &NumberRenamer{
 		symbols: symbols,
-		names:   make([][]string, len(symbols.Outer)),
+		names:   make([][]string, len(symbols.SymbolsForSource)),
 		root:    numberScope{nameCounts: reservedNames},
 	}
 }
 
 func (r *NumberRenamer) NameForSymbol(ref js_ast.Ref) string {
 	ref = js_ast.FollowSymbols(r.symbols, ref)
-	if inner := r.names[ref.OuterIndex]; inner != nil {
+	if inner := r.names[ref.SourceIndex]; inner != nil {
 		if name := inner[ref.InnerIndex]; name != "" {
 			return name
 		}
@@ -351,7 +360,7 @@ func (r *NumberRenamer) assignName(scope *numberScope, ref js_ast.Ref) {
 	ref = js_ast.FollowSymbols(r.symbols, ref)
 
 	// Don't rename the same symbol more than once
-	inner := r.names[ref.OuterIndex]
+	inner := r.names[ref.SourceIndex]
 	if inner != nil && inner[ref.InnerIndex] != "" {
 		return
 	}
@@ -376,8 +385,8 @@ func (r *NumberRenamer) assignName(scope *numberScope, ref js_ast.Ref) {
 		// "names" array ahead of time, that will waste a lot more memory for
 		// builds that make heavy use of code splitting and have many chunks. Doing
 		// things lazily like this means we use less memory but still stay safe.
-		inner = make([]string, len(r.symbols.Outer[ref.OuterIndex]))
-		r.names[ref.OuterIndex] = inner
+		inner = make([]string, len(r.symbols.SymbolsForSource[ref.SourceIndex]))
+		r.names[ref.SourceIndex] = inner
 	}
 	inner[ref.InnerIndex] = name
 }
@@ -394,7 +403,7 @@ func (r *NumberRenamer) assignNamesRecursive(scope *js_ast.Scope, sourceIndex ui
 
 	// Rename all symbols in this scope
 	for _, innerIndex := range *sorted {
-		r.assignName(s, js_ast.Ref{OuterIndex: sourceIndex, InnerIndex: uint32(innerIndex)})
+		r.assignName(s, js_ast.Ref{SourceIndex: sourceIndex, InnerIndex: uint32(innerIndex)})
 	}
 	for _, ref := range scope.Generated {
 		r.assignName(s, ref)
