@@ -94,6 +94,17 @@ const (
 	SourcesContentExclude
 )
 
+type LegalComments uint8
+
+const (
+	LegalCommentsDefault LegalComments = iota
+	LegalCommentsNone
+	LegalCommentsInline
+	LegalCommentsEndOfFile
+	LegalCommentsLinked
+	LegalCommentsExternal
+)
+
 type Target uint8
 
 const (
@@ -159,23 +170,31 @@ type Engine struct {
 }
 
 type Location struct {
-	File      string
-	Namespace string
-	Line      int // 1-based
-	Column    int // 0-based, in bytes
-	Length    int // in bytes
-	LineText  string
+	File       string
+	Namespace  string
+	Line       int // 1-based
+	Column     int // 0-based, in bytes
+	Length     int // in bytes
+	LineText   string
+	Suggestion string
 }
 
 type Message struct {
-	Text     string
-	Location *Location
+	PluginName string
+	Text       string
+	Location   *Location
+	Notes      []Note
 
 	// Optional user-specified data that is passed through unmodified. You can
 	// use this to stash the original error, for example.
 	Detail interface{}
 }
 
+type Note struct {
+	Text     string
+	Location *Location
+}
+
 type StderrColor uint8
 
 const (
@@ -188,6 +207,8 @@ type LogLevel uint8
 
 const (
 	LogLevelSilent LogLevel = iota
+	LogLevelVerbose
+	LogLevelDebug
 	LogLevelInfo
 	LogLevelWarning
 	LogLevelError
@@ -212,11 +233,12 @@ const (
 // Build API
 
 type BuildOptions struct {
-	Color      StderrColor
-	ErrorLimit int
-	LogLevel   LogLevel
+	Color    StderrColor
+	LogLimit int
+	LogLevel LogLevel
 
 	Sourcemap      SourceMap
+	SourceRoot     string
 	SourcesContent SourcesContent
 
 	Target  Target
@@ -227,20 +249,21 @@ type BuildOptions struct {
 	MinifySyntax      bool
 	Charset           Charset
 	TreeShaking       TreeShaking
+	LegalComments     LegalComments
 
 	JSXFactory  string
 	JSXFragment string
 
 	Define    map[string]string
 	Pure      []string
-	AvoidTDZ  bool
 	KeepNames bool
 
 	GlobalName        string
 	Bundle            bool
+	PreserveSymlinks  bool
 	Splitting         bool
 	Outfile           string
-	Metafile          string
+	Metafile          bool
 	Outdir            string
 	Outbase           string
 	AbsWorkingDir     string
@@ -248,28 +271,40 @@ type BuildOptions struct {
 	Format            Format
 	External          []string
 	MainFields        []string
+	Conditions        []string // For the "exports" field in "package.json"
 	Loader            map[string]Loader
 	ResolveExtensions []string
 	Tsconfig          string
 	OutExtensions     map[string]string
 	PublicPath        string
 	Inject            []string
-	Banner            string
-	Footer            string
+	Banner            map[string]string
+	Footer            map[string]string
+	NodePaths         []string // The "NODE_PATH" variable from Node.js
 
-	EntryPoints []string
-	Stdin       *StdinOptions
-	Write       bool
-	Incremental bool
-	Plugins     []Plugin
+	EntryNames string
+	ChunkNames string
+	AssetNames string
+
+	EntryPoints         []string
+	EntryPointsAdvanced []EntryPoint
+
+	Stdin          *StdinOptions
+	Write          bool
+	AllowOverwrite bool
+	Incremental    bool
+	Plugins        []Plugin
 
 	Watch *WatchMode
 }
 
+type EntryPoint struct {
+	InputPath  string
+	OutputPath string
+}
+
 type WatchMode struct {
-	SpinnerBusy string
-	SpinnerIdle []string
-	OnRebuild   func(BuildResult)
+	OnRebuild func(BuildResult)
 }
 
 type StdinOptions struct {
@@ -284,6 +319,7 @@ type BuildResult struct {
 	Warnings []Message
 
 	OutputFiles []OutputFile
+	Metafile    string
 
 	Rebuild func() BuildResult // Only when "Incremental: true"
 	Stop    func()             // Only when "Watch: true"
@@ -302,11 +338,12 @@ func Build(options BuildOptions) BuildResult {
 // Transform API
 
 type TransformOptions struct {
-	Color      StderrColor
-	ErrorLimit int
-	LogLevel   LogLevel
+	Color    StderrColor
+	LogLimit int
+	LogLevel LogLevel
 
 	Sourcemap      SourceMap
+	SourceRoot     string
 	SourcesContent SourcesContent
 
 	Target     Target
@@ -319,6 +356,7 @@ type TransformOptions struct {
 	MinifySyntax      bool
 	Charset           Charset
 	TreeShaking       TreeShaking
+	LegalComments     LegalComments
 
 	JSXFactory  string
 	JSXFragment string
@@ -328,7 +366,6 @@ type TransformOptions struct {
 
 	Define    map[string]string
 	Pure      []string
-	AvoidTDZ  bool
 	KeepNames bool
 
 	Sourcefile string
@@ -353,6 +390,7 @@ func Transform(input string, options TransformOptions) TransformResult {
 type ServeOptions struct {
 	Port      uint16
 	Host      string
+	Servedir  string
 	OnRequest func(ServeOnRequestArgs)
 }
 
@@ -383,9 +421,17 @@ type Plugin struct {
 	Setup func(PluginBuild)
 }
 
-type PluginBuild interface {
-	OnResolve(options OnResolveOptions, callback func(OnResolveArgs) (OnResolveResult, error))
-	OnLoad(options OnLoadOptions, callback func(OnLoadArgs) (OnLoadResult, error))
+type PluginBuild struct {
+	InitialOptions *BuildOptions
+	OnStart        func(callback func() (OnStartResult, error))
+	OnEnd          func(callback func(result *BuildResult))
+	OnResolve      func(options OnResolveOptions, callback func(OnResolveArgs) (OnResolveResult, error))
+	OnLoad         func(options OnLoadOptions, callback func(OnLoadArgs) (OnLoadResult, error))
+}
+
+type OnStartResult struct {
+	Errors   []Message
+	Warnings []Message
 }
 
 type OnResolveOptions struct {
@@ -398,6 +444,7 @@ type OnResolveArgs struct {
 	Importer   string
 	Namespace  string
 	ResolveDir string
+	Kind       ResolveKind
 	PluginData interface{}
 }
 
@@ -411,6 +458,9 @@ type OnResolveResult struct {
 	External   bool
 	Namespace  string
 	PluginData interface{}
+
+	WatchFiles []string
+	WatchDirs  []string
 }
 
 type OnLoadOptions struct {
@@ -434,4 +484,39 @@ type OnLoadResult struct {
 	ResolveDir string
 	Loader     Loader
 	PluginData interface{}
+
+	WatchFiles []string
+	WatchDirs  []string
+}
+
+type ResolveKind uint8
+
+const (
+	ResolveEntryPoint ResolveKind = iota
+	ResolveJSImportStatement
+	ResolveJSRequireCall
+	ResolveJSDynamicImport
+	ResolveJSRequireResolve
+	ResolveCSSImportRule
+	ResolveCSSURLToken
+)
+
+////////////////////////////////////////////////////////////////////////////////
+// FormatMessages API
+
+type MessageKind uint8
+
+const (
+	ErrorMessage MessageKind = iota
+	WarningMessage
+)
+
+type FormatMessagesOptions struct {
+	TerminalWidth int
+	Kind          MessageKind
+	Color         bool
+}
+
+func FormatMessages(msgs []Message, opts FormatMessagesOptions) []string {
+	return formatMsgsImpl(msgs, opts)
 }
