@@ -8,7 +8,6 @@ import (
 	"sort"
 	"strconv"
 	"strings"
-	"time"
 
 	"github.com/evanw/esbuild/internal/cli_helpers"
 	"github.com/evanw/esbuild/internal/fs"
@@ -20,6 +19,8 @@ func newBuildOptions() api.BuildOptions {
 	return api.BuildOptions{
 		Loader: make(map[string]api.Loader),
 		Define: make(map[string]string),
+		Banner: make(map[string]string),
+		Footer: make(map[string]string),
 	}
 }
 
@@ -29,7 +30,19 @@ func newTransformOptions() api.TransformOptions {
 	}
 }
 
-func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpts *api.TransformOptions) error {
+type parseOptionsKind uint8
+
+const (
+	kindInternal parseOptionsKind = iota
+	kindExternal
+)
+
+func parseOptionsImpl(
+	osArgs []string,
+	buildOpts *api.BuildOptions,
+	transformOpts *api.TransformOptions,
+	kind parseOptionsKind,
+) (err error, metafile *string) {
 	hasBareSourceMapFlag := false
 
 	// Parse the arguments now that we know what we're parsing
@@ -38,9 +51,15 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 		case arg == "--bundle" && buildOpts != nil:
 			buildOpts.Bundle = true
 
+		case arg == "--preserve-symlinks" && buildOpts != nil:
+			buildOpts.PreserveSymlinks = true
+
 		case arg == "--splitting" && buildOpts != nil:
 			buildOpts.Splitting = true
 
+		case arg == "--allow-overwrite" && buildOpts != nil:
+			buildOpts.AllowOverwrite = true
+
 		case arg == "--watch" && buildOpts != nil:
 			buildOpts.Watch = &api.WatchMode{}
 
@@ -76,6 +95,29 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 				transformOpts.MinifyIdentifiers = true
 			}
 
+		case strings.HasPrefix(arg, "--legal-comments="):
+			value := arg[len("--legal-comments="):]
+			var legalComments api.LegalComments
+			switch value {
+			case "none":
+				legalComments = api.LegalCommentsNone
+			case "inline":
+				legalComments = api.LegalCommentsInline
+			case "eof":
+				legalComments = api.LegalCommentsEndOfFile
+			case "linked":
+				legalComments = api.LegalCommentsLinked
+			case "external":
+				legalComments = api.LegalCommentsExternal
+			default:
+				return fmt.Errorf("Invalid legal comments value: %q (valid: none, inline, eof, linked, external)", value), nil
+			}
+			if buildOpts != nil {
+				buildOpts.LegalComments = legalComments
+			} else {
+				transformOpts.LegalComments = legalComments
+			}
+
 		case strings.HasPrefix(arg, "--charset="):
 			var value *api.Charset
 			if buildOpts != nil {
@@ -90,7 +132,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			case "utf8":
 				*value = api.CharsetUTF8
 			default:
-				return fmt.Errorf("Invalid charset value: %q (valid: ascii, utf8)", name)
+				return fmt.Errorf("Invalid charset value: %q (valid: ascii, utf8)", name), nil
 			}
 
 		case strings.HasPrefix(arg, "--tree-shaking="):
@@ -105,14 +147,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			case "ignore-annotations":
 				*value = api.TreeShakingIgnoreAnnotations
 			default:
-				return fmt.Errorf("Invalid tree shaking value: %q (valid: ignore-annotations)", name)
-			}
-
-		case arg == "--avoid-tdz":
-			if buildOpts != nil {
-				buildOpts.AvoidTDZ = true
-			} else {
-				transformOpts.AvoidTDZ = true
+				return fmt.Errorf("Invalid tree shaking value: %q (valid: ignore-annotations)", name), nil
 			}
 
 		case arg == "--keep-names":
@@ -141,7 +176,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			case "both":
 				sourcemap = api.SourceMapInlineAndExternal
 			default:
-				return fmt.Errorf("Invalid sourcemap: %q (valid: inline, external, both)", value)
+				return fmt.Errorf("Invalid sourcemap: %q (valid: inline, external, both)", value), nil
 			}
 			if buildOpts != nil {
 				buildOpts.Sourcemap = sourcemap
@@ -150,6 +185,14 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			}
 			hasBareSourceMapFlag = false
 
+		case strings.HasPrefix(arg, "--source-root="):
+			sourceRoot := arg[len("--source-root="):]
+			if buildOpts != nil {
+				buildOpts.SourceRoot = sourceRoot
+			} else {
+				transformOpts.SourceRoot = sourceRoot
+			}
+
 		case strings.HasPrefix(arg, "--sources-content="):
 			value := arg[len("--sources-content="):]
 			var sourcesContent api.SourcesContent
@@ -159,7 +202,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			case "true":
 				sourcesContent = api.SourcesContentInclude
 			default:
-				return fmt.Errorf("Invalid sources content: %q (valid: false, true)", value)
+				return fmt.Errorf("Invalid sources content: %q (valid: false, true)", value), nil
 			}
 			if buildOpts != nil {
 				buildOpts.SourcesContent = sourcesContent
@@ -183,6 +226,9 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 		case strings.HasPrefix(arg, "--main-fields=") && buildOpts != nil:
 			buildOpts.MainFields = strings.Split(arg[len("--main-fields="):], ",")
 
+		case strings.HasPrefix(arg, "--conditions=") && buildOpts != nil:
+			buildOpts.Conditions = strings.Split(arg[len("--conditions="):], ",")
+
 		case strings.HasPrefix(arg, "--public-path=") && buildOpts != nil:
 			buildOpts.PublicPath = arg[len("--public-path="):]
 
@@ -193,8 +239,13 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 				transformOpts.GlobalName = arg[len("--global-name="):]
 			}
 
-		case strings.HasPrefix(arg, "--metafile=") && buildOpts != nil:
-			buildOpts.Metafile = arg[len("--metafile="):]
+		case strings.HasPrefix(arg, "--metafile") && buildOpts != nil && kind == kindExternal:
+			buildOpts.Metafile = true
+
+		case strings.HasPrefix(arg, "--metafile=") && buildOpts != nil && kind == kindInternal:
+			metafilePath := arg[len("--metafile="):]
+			buildOpts.Metafile = true
+			metafile = &metafilePath
 
 		case strings.HasPrefix(arg, "--outfile=") && buildOpts != nil:
 			buildOpts.Outfile = arg[len("--outfile="):]
@@ -211,11 +262,20 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 		case strings.HasPrefix(arg, "--tsconfig-raw=") && transformOpts != nil:
 			transformOpts.TsconfigRaw = arg[len("--tsconfig-raw="):]
 
+		case strings.HasPrefix(arg, "--entry-names=") && buildOpts != nil:
+			buildOpts.EntryNames = arg[len("--entry-names="):]
+
+		case strings.HasPrefix(arg, "--chunk-names=") && buildOpts != nil:
+			buildOpts.ChunkNames = arg[len("--chunk-names="):]
+
+		case strings.HasPrefix(arg, "--asset-names=") && buildOpts != nil:
+			buildOpts.AssetNames = arg[len("--asset-names="):]
+
 		case strings.HasPrefix(arg, "--define:"):
 			value := arg[len("--define:"):]
 			equals := strings.IndexByte(value, '=')
 			if equals == -1 {
-				return fmt.Errorf("Missing \"=\": %q", value)
+				return fmt.Errorf("Missing \"=\": %q", value), nil
 			}
 			if buildOpts != nil {
 				buildOpts.Define[value[:equals]] = value[equals+1:]
@@ -235,12 +295,12 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			value := arg[len("--loader:"):]
 			equals := strings.IndexByte(value, '=')
 			if equals == -1 {
-				return fmt.Errorf("Missing \"=\": %q", value)
+				return fmt.Errorf("Missing \"=\": %q", value), nil
 			}
 			ext, text := value[:equals], value[equals+1:]
 			loader, err := cli_helpers.ParseLoader(text)
 			if err != nil {
-				return err
+				return err, nil
 			}
 			buildOpts.Loader[ext] = loader
 
@@ -248,10 +308,10 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			value := arg[len("--loader="):]
 			loader, err := cli_helpers.ParseLoader(value)
 			if err != nil {
-				return err
+				return err, nil
 			}
 			if loader == api.LoaderFile {
-				return fmt.Errorf("Cannot transform using the \"file\" loader")
+				return fmt.Errorf("Cannot transform using the \"file\" loader"), nil
 			}
 			if buildOpts != nil {
 				if buildOpts.Stdin == nil {
@@ -265,7 +325,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 		case strings.HasPrefix(arg, "--target="):
 			target, engines, err := parseTargets(strings.Split(arg[len("--target="):], ","))
 			if err != nil {
-				return err
+				return err, nil
 			}
 			if buildOpts != nil {
 				buildOpts.Target = target
@@ -279,7 +339,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			value := arg[len("--out-extension:"):]
 			equals := strings.IndexByte(value, '=')
 			if equals == -1 {
-				return fmt.Errorf("Missing \"=\": %q", value)
+				return fmt.Errorf("Missing \"=\": %q", value), nil
 			}
 			if buildOpts.OutExtensions == nil {
 				buildOpts.OutExtensions = make(map[string]string)
@@ -296,7 +356,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			case "neutral":
 				buildOpts.Platform = api.PlatformNeutral
 			default:
-				return fmt.Errorf("Invalid platform: %q (valid: browser, node, neutral)", value)
+				return fmt.Errorf("Invalid platform: %q (valid: browser, node, neutral)", value), nil
 			}
 
 		case strings.HasPrefix(arg, "--format="):
@@ -321,7 +381,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 					transformOpts.Format = api.FormatESModule
 				}
 			default:
-				return fmt.Errorf("Invalid format: %q (valid: iife, cjs, esm)", value)
+				return fmt.Errorf("Invalid format: %q (valid: iife, cjs, esm)", value), nil
 			}
 
 		case strings.HasPrefix(arg, "--external:") && buildOpts != nil:
@@ -346,32 +406,38 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 				transformOpts.JSXFragment = value
 			}
 
-		case strings.HasPrefix(arg, "--banner="):
-			value := arg[len("--banner="):]
-			if buildOpts != nil {
-				buildOpts.Banner = value
-			} else {
-				transformOpts.Banner = value
-			}
+		case strings.HasPrefix(arg, "--banner=") && transformOpts != nil:
+			transformOpts.Banner = arg[len("--banner="):]
 
-		case strings.HasPrefix(arg, "--footer="):
-			value := arg[len("--footer="):]
-			if buildOpts != nil {
-				buildOpts.Footer = value
-			} else {
-				transformOpts.Footer = value
-			}
+		case strings.HasPrefix(arg, "--footer=") && transformOpts != nil:
+			transformOpts.Footer = arg[len("--footer="):]
 
-		case strings.HasPrefix(arg, "--error-limit="):
-			value := arg[len("--error-limit="):]
+		case strings.HasPrefix(arg, "--banner:") && buildOpts != nil:
+			value := arg[len("--banner:"):]
+			equals := strings.IndexByte(value, '=')
+			if equals == -1 {
+				return fmt.Errorf("Missing \"=\": %q", value), nil
+			}
+			buildOpts.Banner[value[:equals]] = value[equals+1:]
+
+		case strings.HasPrefix(arg, "--footer:") && buildOpts != nil:
+			value := arg[len("--footer:"):]
+			equals := strings.IndexByte(value, '=')
+			if equals == -1 {
+				return fmt.Errorf("Missing \"=\": %q", value), nil
+			}
+			buildOpts.Footer[value[:equals]] = value[equals+1:]
+
+		case strings.HasPrefix(arg, "--log-limit="):
+			value := arg[len("--log-limit="):]
 			limit, err := strconv.Atoi(value)
 			if err != nil || limit < 0 {
-				return fmt.Errorf("Invalid error limit: %q", value)
+				return fmt.Errorf("Invalid log limit: %q", value), nil
 			}
 			if buildOpts != nil {
-				buildOpts.ErrorLimit = limit
+				buildOpts.LogLimit = limit
 			} else {
-				transformOpts.ErrorLimit = limit
+				transformOpts.LogLimit = limit
 			}
 
 			// Make sure this stays in sync with "PrintErrorToStderr"
@@ -384,7 +450,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			case "true":
 				color = api.ColorAlways
 			default:
-				return fmt.Errorf("Invalid color: %q (valid: false, true)", value)
+				return fmt.Errorf("Invalid color: %q (valid: false, true)", value), nil
 			}
 			if buildOpts != nil {
 				buildOpts.Color = color
@@ -397,6 +463,10 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			value := arg[len("--log-level="):]
 			var logLevel api.LogLevel
 			switch value {
+			case "verbose":
+				logLevel = api.LogLevelVerbose
+			case "debug":
+				logLevel = api.LogLevelDebug
 			case "info":
 				logLevel = api.LogLevelInfo
 			case "warning":
@@ -406,7 +476,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			case "silent":
 				logLevel = api.LogLevelSilent
 			default:
-				return fmt.Errorf("Invalid log level: %q (valid: info, warning, error, silent)", arg)
+				return fmt.Errorf("Invalid log level: %q (valid: verbose, debug, info, warning, error, silent)", arg), nil
 			}
 			if buildOpts != nil {
 				buildOpts.LogLevel = logLevel
@@ -415,16 +485,23 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 			}
 
 		case strings.HasPrefix(arg, "'--"):
-			return fmt.Errorf("Unexpected single quote character before flag (use \\\" to escape double quotes): %s", arg)
+			return fmt.Errorf("Unexpected single quote character before flag (use \\\" to escape double quotes): %s", arg), nil
 
 		case !strings.HasPrefix(arg, "-") && buildOpts != nil:
-			buildOpts.EntryPoints = append(buildOpts.EntryPoints, arg)
+			if equals := strings.IndexByte(arg, '='); equals != -1 {
+				buildOpts.EntryPointsAdvanced = append(buildOpts.EntryPointsAdvanced, api.EntryPoint{
+					OutputPath: arg[:equals],
+					InputPath:  arg[equals+1:],
+				})
+			} else {
+				buildOpts.EntryPoints = append(buildOpts.EntryPoints, arg)
+			}
 
 		default:
 			if buildOpts != nil {
-				return fmt.Errorf("Invalid build flag: %q", arg)
+				return fmt.Errorf("Invalid build flag: %q", arg), nil
 			} else {
-				return fmt.Errorf("Invalid transform flag: %q", arg)
+				return fmt.Errorf("Invalid transform flag: %q", arg), nil
 			}
 		}
 	}
@@ -436,7 +513,7 @@ func parseOptionsImpl(osArgs []string, buildOpts *api.BuildOptions, transformOpt
 		buildOpts.Sourcemap = api.SourceMapInline
 	}
 
-	return nil
+	return
 }
 
 func parseTargets(targets []string) (target api.Target, engines []api.Engine, err error) {
@@ -491,22 +568,22 @@ outer:
 }
 
 // This returns either BuildOptions, TransformOptions, or an error
-func parseOptionsForRun(osArgs []string) (*api.BuildOptions, *api.TransformOptions, error) {
+func parseOptionsForRun(osArgs []string) (*api.BuildOptions, *string, *api.TransformOptions, error) {
 	// If there's an entry point or we're bundling, then we're building
 	for _, arg := range osArgs {
 		if !strings.HasPrefix(arg, "-") || arg == "--bundle" {
 			options := newBuildOptions()
 
 			// Apply defaults appropriate for the CLI
-			options.ErrorLimit = 10
+			options.LogLimit = 10
 			options.LogLevel = api.LogLevelInfo
 			options.Write = true
 
-			err := parseOptionsImpl(osArgs, &options, nil)
+			err, metafile := parseOptionsImpl(osArgs, &options, nil, kindInternal)
 			if err != nil {
-				return nil, nil, err
+				return nil, nil, nil, err
 			}
-			return &options, nil, nil
+			return &options, metafile, nil, nil
 		}
 	}
 
@@ -514,69 +591,59 @@ func parseOptionsForRun(osArgs []string) (*api.BuildOptions, *api.TransformOptio
 	options := newTransformOptions()
 
 	// Apply defaults appropriate for the CLI
-	options.ErrorLimit = 10
+	options.LogLimit = 10
 	options.LogLevel = api.LogLevelInfo
 
-	err := parseOptionsImpl(osArgs, nil, &options)
+	err, _ := parseOptionsImpl(osArgs, nil, &options, kindInternal)
 	if err != nil {
-		return nil, nil, err
+		return nil, nil, nil, err
 	}
 	if options.Sourcemap != api.SourceMapNone && options.Sourcemap != api.SourceMapInline {
-		return nil, nil, fmt.Errorf("Must use \"inline\" source map when transforming stdin")
+		return nil, nil, nil, fmt.Errorf("Must use \"inline\" source map when transforming stdin")
 	}
-	return nil, &options, nil
+	return nil, nil, &options, nil
 }
 
 func runImpl(osArgs []string) int {
-	shouldPrintSummary := false
-	start := time.Now()
 	end := 0
 
-	for i, arg := range osArgs {
+	for _, arg := range osArgs {
 		// Special-case running a server
-		if arg == "--serve" {
-			arg = "--serve=0"
-		}
-		if strings.HasPrefix(arg, "--serve=") {
-			serve := arg[len("--serve="):]
-			osArgs = append(append([]string{}, osArgs[:i]...), osArgs[i+1:]...)
-			if err := serveImpl(serve, osArgs); err != nil {
+		if arg == "--serve" || strings.HasPrefix(arg, "--serve=") || strings.HasPrefix(arg, "--servedir=") {
+			if err := serveImpl(osArgs); err != nil {
 				logger.PrintErrorToStderr(osArgs, err.Error())
 				return 1
 			}
 			return 0
 		}
 
-		// Filter out the "--summary" flag
-		if arg == "--summary" {
-			shouldPrintSummary = true
-			continue
-		}
-
 		osArgs[end] = arg
 		end++
 	}
 	osArgs = osArgs[:end]
 
-	buildOptions, transformOptions, err := parseOptionsForRun(osArgs)
+	buildOptions, metafile, transformOptions, err := parseOptionsForRun(osArgs)
 
 	switch {
 	case buildOptions != nil:
-		// These characters were chosen because they work in Windows Command Prompt
-		if buildOptions.Watch != nil {
-			buildOptions.Watch.SpinnerBusy = "··· "
-			buildOptions.Watch.SpinnerIdle = []string{
-				"▫▫▫ ",
-				"▪▫▫ ",
-				"▪▪▫ ",
-				"▪▪▪ ",
-				"▫▪▪ ",
-				"▫▫▪ ",
+		// Read the "NODE_PATH" from the environment. This is part of node's
+		// module resolution algorithm. Documentation for this can be found here:
+		// https://nodejs.org/api/modules.html#modules_loading_from_the_global_folders
+		for _, key := range os.Environ() {
+			if strings.HasPrefix(key, "NODE_PATH=") {
+				value := key[len("NODE_PATH="):]
+				separator := ":"
+				if fs.CheckIfWindows() {
+					// On Windows, NODE_PATH is delimited by semicolons instead of colons
+					separator = ";"
+				}
+				buildOptions.NodePaths = strings.Split(value, separator)
+				break
 			}
 		}
 
 		// Read from stdin when there are no entry points
-		if len(buildOptions.EntryPoints) == 0 {
+		if len(buildOptions.EntryPoints)+len(buildOptions.EntryPointsAdvanced) == 0 {
 			if buildOptions.Stdin == nil {
 				buildOptions.Stdin = &api.StdinOptions{}
 			}
@@ -599,20 +666,60 @@ func runImpl(osArgs []string) int {
 			return 1
 		}
 
-		// Run the build and stop if there were errors
-		result := api.Build(*buildOptions)
-		if len(result.Errors) > 0 {
-			return 1
+		// Validate the metafile absolute path and directory ahead of time so we
+		// don't write any output files if it's incorrect. That makes this API
+		// option consistent with how we handle all other API options.
+		var metafileAbsPath string
+		var metafileAbsDir string
+		if metafile != nil {
+			if buildOptions.Outfile == "" && buildOptions.Outdir == "" {
+				// Cannot use "metafile" when writing to stdout
+				logger.PrintErrorToStderr(osArgs, "Cannot use \"metafile\" without an output path")
+				return 1
+			}
+			if realFS, err := fs.RealFS(fs.RealFSOptions{AbsWorkingDir: buildOptions.AbsWorkingDir}); err == nil {
+				absPath, ok := realFS.Abs(*metafile)
+				if !ok {
+					logger.PrintErrorToStderr(osArgs, fmt.Sprintf("Invalid metafile path: %s", *metafile))
+					return 1
+				}
+				metafileAbsPath = absPath
+				metafileAbsDir = realFS.Dir(absPath)
+			} else {
+				// Don't fail in this case since the error will be reported by "api.Build"
+			}
 		}
 
+		// Run the build
+		result := api.Build(*buildOptions)
+
 		// Do not exit if we're in watch mode
 		if buildOptions.Watch != nil {
 			<-make(chan bool)
 		}
 
-		// Print a summary to stderr
-		if shouldPrintSummary {
-			printSummary(osArgs, result.OutputFiles, start)
+		// Stop if there were errors
+		if len(result.Errors) > 0 {
+			return 1
+		}
+
+		// Write the metafile to the file system
+		if metafile != nil {
+			if err != nil {
+				// This should already have been checked above
+				panic(err.Error())
+			}
+			fs.BeforeFileOpen()
+			defer fs.AfterFileClose()
+			if err := os.MkdirAll(metafileAbsDir, 0755); err != nil {
+				logger.PrintErrorToStderr(osArgs, fmt.Sprintf(
+					"Failed to create output directory: %s", err.Error()))
+			} else {
+				if err := ioutil.WriteFile(metafileAbsPath, []byte(result.Metafile), 0644); err != nil {
+					logger.PrintErrorToStderr(osArgs, fmt.Sprintf(
+						"Failed to write to output file: %s", err.Error()))
+				}
+			}
 		}
 
 	case transformOptions != nil:
@@ -633,11 +740,6 @@ func runImpl(osArgs []string) int {
 		// Write the output to stdout
 		os.Stdout.Write(result.Code)
 
-		// Print a summary to stderr
-		if shouldPrintSummary {
-			printSummary(osArgs, nil, start)
-		}
-
 	case err != nil:
 		logger.PrintErrorToStderr(osArgs, err.Error())
 		return 1
@@ -646,91 +748,79 @@ func runImpl(osArgs []string) int {
 	return 0
 }
 
-func printSummary(osArgs []string, outputFiles []api.OutputFile, start time.Time) {
-	var table logger.SummaryTable = make([]logger.SummaryTableEntry, len(outputFiles))
+func parseServeOptionsImpl(osArgs []string) (api.ServeOptions, []string, error) {
+	host := ""
+	portText := "0"
+	servedir := ""
 
-	if len(outputFiles) > 0 {
-		if cwd, err := os.Getwd(); err == nil {
-			if realFS, err := fs.RealFS(fs.RealFSOptions{AbsWorkingDir: cwd}); err == nil {
-				for i, file := range outputFiles {
-					path, ok := realFS.Rel(realFS.Cwd(), file.Path)
-					if !ok {
-						path = file.Path
-					}
-					base := realFS.Base(path)
-					n := len(file.Contents)
-					var size string
-					if n < 1024 {
-						size = fmt.Sprintf("%db ", n)
-					} else if n < 1024*1024 {
-						size = fmt.Sprintf("%.1fkb", float64(n)/(1024))
-					} else if n < 1024*1024*1024 {
-						size = fmt.Sprintf("%.1fmb", float64(n)/(1024*1024))
-					} else {
-						size = fmt.Sprintf("%.1fgb", float64(n)/(1024*1024*1024))
-					}
-					table[i] = logger.SummaryTableEntry{
-						Dir:         path[:len(path)-len(base)],
-						Base:        base,
-						Size:        size,
-						Bytes:       n,
-						IsSourceMap: strings.HasSuffix(base, ".map"),
-					}
-				}
-			}
+	// Filter out server-specific flags
+	filteredArgs := make([]string, 0, len(osArgs))
+	for _, arg := range osArgs {
+		if arg == "--serve" {
+			// Just ignore this flag
+		} else if strings.HasPrefix(arg, "--serve=") {
+			portText = arg[len("--serve="):]
+		} else if strings.HasPrefix(arg, "--servedir=") {
+			servedir = arg[len("--servedir="):]
+		} else {
+			filteredArgs = append(filteredArgs, arg)
 		}
 	}
 
-	logger.PrintSummary(osArgs, table, start)
-}
-
-func serveImpl(serveText string, osArgs []string) error {
-	host := ""
-	portText := serveText
-
 	// Specifying the host is optional
-	if strings.ContainsRune(serveText, ':') {
+	if strings.ContainsRune(portText, ':') {
 		var err error
-		host, portText, err = net.SplitHostPort(serveText)
+		host, portText, err = net.SplitHostPort(portText)
 		if err != nil {
-			return err
+			return api.ServeOptions{}, nil, err
 		}
 	}
 
 	// Parse the port
 	port, err := strconv.ParseInt(portText, 10, 32)
 	if err != nil {
-		return err
+		return api.ServeOptions{}, nil, err
 	}
 	if port < 0 || port > 0xFFFF {
-		return fmt.Errorf("Invalid port number: %s", portText)
+		return api.ServeOptions{}, nil, fmt.Errorf("Invalid port number: %s", portText)
+	}
+
+	return api.ServeOptions{
+		Port:     uint16(port),
+		Host:     host,
+		Servedir: servedir,
+	}, filteredArgs, nil
+}
+
+func serveImpl(osArgs []string) error {
+	serveOptions, filteredArgs, err := parseServeOptionsImpl(osArgs)
+	if err != nil {
+		return err
 	}
 
 	options := newBuildOptions()
 
 	// Apply defaults appropriate for the CLI
-	options.ErrorLimit = 5
+	options.LogLimit = 5
 	options.LogLevel = api.LogLevelInfo
 
-	if err := parseOptionsImpl(osArgs, &options, nil); err != nil {
-		logger.PrintErrorToStderr(osArgs, err.Error())
+	if err, _ := parseOptionsImpl(filteredArgs, &options, nil, kindInternal); err != nil {
+		logger.PrintErrorToStderr(filteredArgs, err.Error())
 		return err
 	}
 
-	serveOptions := api.ServeOptions{
-		Port: uint16(port),
-		Host: host,
-		OnRequest: func(args api.ServeOnRequestArgs) {
-			logger.PrintText(os.Stderr, logger.LevelInfo, osArgs, func(colors logger.Colors) string {
-				statusColor := colors.Red
-				if args.Status == 200 {
-					statusColor = colors.Green
-				}
-				return fmt.Sprintf("%s%s - %q %s%d%s [%dms]%s\n",
-					colors.Dim, args.RemoteAddress, args.Method+" "+args.Path,
-					statusColor, args.Status, colors.Dim, args.TimeInMS, colors.Default)
-			})
-		},
+	serveOptions.OnRequest = func(args api.ServeOnRequestArgs) {
+		logger.PrintText(os.Stderr, logger.LevelInfo, filteredArgs, func(colors logger.Colors) string {
+			statusColor := colors.Red
+			if args.Status >= 200 && args.Status <= 299 {
+				statusColor = colors.Green
+			} else if args.Status >= 300 && args.Status <= 399 {
+				statusColor = colors.Yellow
+			}
+			return fmt.Sprintf("%s%s - %q %s%d%s [%dms]%s\n",
+				colors.Dim, args.RemoteAddress, args.Method+" "+args.Path,
+				statusColor, args.Status, colors.Dim, args.TimeInMS, colors.Reset)
+		})
 	}
 
 	result, err := api.Serve(serveOptions, options)
@@ -739,9 +829,50 @@ func serveImpl(serveText string, osArgs []string) error {
 	}
 
 	// Show what actually got bound if the port was 0
-	logger.PrintText(os.Stderr, logger.LevelInfo, osArgs, func(colors logger.Colors) string {
-		return fmt.Sprintf("%s\n > %shttp://%s:%d/%s\n\n",
-			colors.Default, colors.Underline, result.Host, result.Port, colors.Default)
+	logger.PrintText(os.Stderr, logger.LevelInfo, filteredArgs, func(colors logger.Colors) string {
+		var hosts []string
+		sb := strings.Builder{}
+		sb.WriteString(colors.Reset)
+
+		// If this is "0.0.0.0" or "::", list all relevant IP addresses
+		if ip := net.ParseIP(result.Host); ip != nil && ip.IsUnspecified() {
+			if addrs, err := net.InterfaceAddrs(); err == nil {
+				for _, addr := range addrs {
+					if addr, ok := addr.(*net.IPNet); ok && (addr.IP.To4() != nil) == (ip.To4() != nil) && !addr.IP.IsLinkLocalUnicast() {
+						hosts = append(hosts, addr.IP.String())
+					}
+				}
+			}
+		}
+
+		// Otherwise, just list the one IP address
+		if len(hosts) == 0 {
+			hosts = append(hosts, result.Host)
+		}
+
+		// Determine the host kinds
+		kinds := make([]string, len(hosts))
+		maxLen := 0
+		for i, host := range hosts {
+			kind := "Network"
+			if ip := net.ParseIP(host); ip != nil && ip.IsLoopback() {
+				kind = "Local"
+			}
+			kinds[i] = kind
+			if len(kind) > maxLen {
+				maxLen = len(kind)
+			}
+		}
+
+		// Pretty-print the host list
+		for i, kind := range kinds {
+			sb.WriteString(fmt.Sprintf("\n > %s:%s %shttp://%s/%s",
+				kind, strings.Repeat(" ", maxLen-len(kind)), colors.Underline,
+				net.JoinHostPort(hosts[i], fmt.Sprintf("%d", result.Port)), colors.Reset))
+		}
+
+		sb.WriteString("\n\n")
+		return sb.String()
 	})
 	return result.Wait()
 }
