@@ -37,8 +37,7 @@ function formatErrors(content, error) {
 }
 
 async function main() {
-  const { startService } = installForTests();
-  const service = await startService();
+  const esbuild = installForTests();
   const files = findFiles();
   let runCount = 0;
   let shouldHavePassed = 0;
@@ -49,15 +48,15 @@ async function main() {
 
   async function esbuildFile(input, options) {
     try {
-      const { js } = await service.transform(input, options);
-      return { success: true, output: js };
+      const { code } = await esbuild.transform(input, options);
+      return { success: true, output: code };
     } catch (error) {
       return { success: false, error };
     }
   }
 
   async function processFile(file) {
-    const content = fs.readFileSync(file, 'utf8');
+    let content = fs.readFileSync(file, 'utf8');
     const start = content.indexOf('/*---');
     const end = content.indexOf('---*/');
 
@@ -76,6 +75,10 @@ async function main() {
       if (yaml.features.includes('regexp-unicode-property-escapes')) return
     }
 
+    if (yaml.flags) {
+      if (yaml.flags.includes('onlyStrict')) content = '"use strict";\n' + content
+    }
+
     const result = await esbuildFile(content, { minify: false });
 
     if (result.success !== shouldParse) {
@@ -141,9 +144,6 @@ async function main() {
   console.log(`reparse failures: ${reparseCount}`);
   console.log(`reprint failures: ${reprintCount}`);
   console.log(`minify failures: ${minifyCount}`);
-
-  // Clean up after all tests finish
-  service.stop();
 }
 
 main().catch(e => setTimeout(() => {
