@@ -16,13 +16,11 @@ const tests = {
     import esbuild = require('esbuild')
     esbuild.buildSync({})
     esbuild.build({})
-    esbuild.startService().then(service => service.build({}))
   `,
   emptyBuildImport: `
     import * as esbuild from 'esbuild'
     esbuild.buildSync({})
     esbuild.build({})
-    esbuild.startService().then(service => service.build({}))
   `,
   emptyTransformRequire: `
     export {}
@@ -31,8 +29,6 @@ const tests = {
     esbuild.transformSync('', {})
     esbuild.transform('')
     esbuild.transform('', {})
-    esbuild.startService().then(service => service.transform(''))
-    esbuild.startService().then(service => service.transform('', {}))
   `,
   emptyTransformImport: `
     import * as esbuild from 'esbuild'
@@ -40,14 +36,11 @@ const tests = {
     esbuild.transformSync('', {})
     esbuild.transform('')
     esbuild.transform('', {})
-    esbuild.startService().then(service => service.transform(''))
-    esbuild.startService().then(service => service.transform('', {}))
   `,
   writeFalseOutputFiles: `
     import * as esbuild from 'esbuild'
     esbuild.buildSync({ write: false }).outputFiles[0]
     esbuild.build({ write: false }).then(result => result.outputFiles[0])
-    esbuild.startService().then(service => service.build({ write: false }).then(result => result.outputFiles[0]))
   `,
   incrementalTrueRebuild: `
     import * as esbuild from 'esbuild'
@@ -59,16 +52,6 @@ const tests = {
       })
       result.rebuild.dispose()
     })
-    esbuild.startService().then(service => {
-      service.build({ incremental: true }).then(result => {
-        result.rebuild().then(result => {
-          result.rebuild().then(() => {
-            result.rebuild.dispose()
-          })
-        })
-        result.rebuild.dispose()
-      })
-    })
     async function a() {
       let result = await esbuild.build({ incremental: true })
       let result2 = await result.rebuild()
@@ -76,15 +59,6 @@ const tests = {
       result2.rebuild.dispose()
       result.rebuild.dispose()
     }
-    async function b() {
-      let service = await esbuild.startService()
-      let result = await service.build({ incremental: true })
-      let result2 = await result.rebuild()
-      await result2.rebuild()
-      result2.rebuild.dispose()
-      result.rebuild.dispose()
-      service.stop()
-    }
   `,
   ifRebuild: `
     import * as esbuild from 'esbuild'
@@ -101,20 +75,6 @@ const tests = {
         })
       }
     })
-    esbuild.startService().then(service => {
-      service.build(options).then(result => {
-        if (result.rebuild) {
-          result.rebuild().then(result => {
-            if (result.rebuild) {
-              result.rebuild().then(result => {
-                result.rebuild.dispose()
-              })
-            }
-            result.rebuild.dispose()
-          })
-        }
-      })
-    })
     async function a() {
       let result = await esbuild.build(options)
       if (result.rebuild) {
@@ -126,19 +86,6 @@ const tests = {
         result.rebuild.dispose()
       }
     }
-    async function b() {
-      let service = await esbuild.startService()
-      let result = await service.build(options)
-      if (result.rebuild) {
-        let result2 = await result.rebuild()
-        if (result2.rebuild) {
-          await result2.rebuild()
-          result2.rebuild.dispose()
-        }
-        result.rebuild.dispose()
-      }
-      service.stop()
-    }
   `,
   allOptionsTransform: `
     export {}
@@ -157,10 +104,9 @@ const tests = {
       jsxFragment: '',
       define: { 'x': 'y' },
       pure: ['x'],
-      avoidTDZ: true,
       color: true,
       logLevel: 'info',
-      errorLimit: 0,
+      logLimit: 0,
       tsconfigRaw: {
         compilerOptions: {
           jsxFactory: '',
@@ -204,14 +150,13 @@ const tests = {
       jsxFragment: '',
       define: { 'x': 'y' },
       pure: ['x'],
-      avoidTDZ: true,
       color: true,
       logLevel: 'info',
-      errorLimit: 0,
+      logLimit: 0,
       bundle: true,
       splitting: true,
       outfile: '',
-      metafile: '',
+      metafile: true,
       outdir: '',
       outbase: '',
       platform: 'node',
@@ -322,7 +267,7 @@ async function main() {
   fs.mkdirSync(testDir, { recursive: true })
   fs.writeFileSync(path.join(testDir, 'tsconfig.json'), JSON.stringify(tsconfigJson))
 
-  const types = fs.readFileSync(path.join(__dirname, '..', 'lib', 'types.ts'), 'utf8')
+  const types = fs.readFileSync(path.join(__dirname, '..', 'lib', 'shared', 'types.ts'), 'utf8')
   const esbuild_d_ts = path.join(testDir, 'node_modules', 'esbuild', 'index.d.ts')
   fs.mkdirSync(path.dirname(esbuild_d_ts), { recursive: true })
   fs.writeFileSync(esbuild_d_ts, `
@@ -353,4 +298,8 @@ async function main() {
   }
 }
 
-main()
+main().catch(e => {
+  setTimeout(() => {
+    throw e
+  })
+})
